package anl.aida;

import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.GregorianCalendar;

import javax.measure.unit.Unit;

import ucar.ma2.Array;
import ucar.ma2.ArrayFloat;
import ucar.ma2.InvalidRangeException;
import anl.aida.data.AbstractDataFrame;
import anl.aida.data.AbstractDataset;
import anl.aida.data.Axes;
import anl.aida.data.AxisData;
import anl.aida.data.AxisType;
import anl.aida.data.DataFrame;
import anl.aida.data.DataFrameAxis;
import anl.aida.data.DataFrameBuilder;
import anl.aida.data.DataFrameIndex;
import anl.aida.data.Dataset;
import anl.aida.data.DefaultStringCoordAxis;
import anl.aida.data.DefaultVariable;
import anl.aida.data.Range;
import anl.aida.data.RegularTimeAxis;
import anl.aida.data.Slice;
import anl.aida.data.Units;
import anl.aida.data.Variable;

/**
 * Creates a Dataset from an "AIDA Format" data file.
 * 
 */
public class AIDADatasetProducer {
  
  private AxisType catAxisType;
  private String catAxisDescription;
  private Unit<?> timeAxisUnit;
  
  

  /**
   * @param catAxisType
   * @param catAxisDescription
   * @param timeAxisUnit
   */
  public AIDADatasetProducer(AxisType catAxisType, String catAxisDescription, Unit<?> timeAxisUnit) {
    this.catAxisType = catAxisType;
    this.catAxisDescription = catAxisDescription;
    this.timeAxisUnit = timeAxisUnit;
  }

  public Dataset createDataset(String file, String name) throws IOException {
    AFDataset dataset = new AFDataset(name);
    Calendar cal = GregorianCalendar.getInstance();

    RandomAccessFile in = new RandomAccessFile(new File(file), "r");
    int count = in.readInt();
    // System.out.println(count);
    for (int i = 0; i < count; i++) {
      String varName = in.readUTF();
      // System.out.println(varName);
      DefaultVariable var = new DefaultVariable(varName, varName, Unit.ONE, dataset);
      dataset.addVar(var);
      int catCount = in.readInt();
      // System.out.println(catCount);
      String[] categories = new String[catCount];
      for (int j = 0; j < catCount; j++) {
	String category = in.readUTF();
	categories[j] = category;
	// System.out.println(category);
      }

      DataFrameBuilder.AxesBuilder builder = new DataFrameBuilder.AxesBuilder();

      AxisData axisData = new AxisData();
      axisData.setUnit(Units.createUnit(catAxisDescription));
      axisData.setDescription(catAxisDescription);
      axisData.setName(catAxisDescription);
      axisData.setType(catAxisType);
      axisData.setRange(new Range(0, categories.length));
      DefaultStringCoordAxis axis = new DefaultStringCoordAxis(categories, axisData);
      builder.addAxis(DataFrameAxis.createDataFrameAxis(axis, 1));

      int min = in.readInt();
      int max = in.readInt();

      // System.out.println(min);
      // System.out.println(max);

      cal.set(min, 11, 31, 12, 0, 0);
      axisData = new AxisData();
      axisData.setUnit(timeAxisUnit);
      axisData.setDescription("Date");
      axisData.setName("Date");
      axisData.setRange(new Range(0, max - min + 1));
      axisData.setType(AxisType.TIME);
      RegularTimeAxis tAxis = new RegularTimeAxis(cal.getTime(), 1, axisData);
      builder.addAxis(DataFrameAxis.createDataFrameAxis(tAxis, 0));

      AFDataFrame frame = new AFDataFrame(var, dataset, builder.buildAxes());
      dataset.addDataFrame(frame);

      frame.file = new File(file);
      frame.dataCount = in.readInt();
      frame.offset = in.readLong();
    }

    in.close();

    return dataset;
  }

  private static class AFDataFrame extends AbstractDataFrame {

    private File file;
    private int dataCount;
    private long offset;

    /*
     * protected Array array; protected Axes<DataFrameAxis> axes; protected
     * List<Dataset> datasets; protected Variable variable
     */

    public AFDataFrame(Variable var, Dataset data, Axes<DataFrameAxis> axes) {
      this.variable = var;
      datasets = new ArrayList<Dataset>();
      datasets.add(data);
      this.axes = axes;
    }

    private void init() {
      array = new ArrayFloat.D2(axes.getTimeAxis().getExtent(), axes.getNonTimeAxis().getExtent());
      for (int i = 0; i < array.getShape()[0]; i++) {
	for (int j = 0; j < array.getShape()[1]; j++) {
	  ((ArrayFloat.D2) array).set(i, j, Float.NaN);
	}
      }

      try {
	RandomAccessFile in = new RandomAccessFile(file, "r");
	in.seek(offset);
	for (int i = 0; i < dataCount; i++) {
	  int tIndex = in.readInt();
	  int cIndex = in.readInt();
	  float value = in.readFloat();
	  // System.out.println(tIndex + ", " + cIndex + ": " + value);
	  ((ArrayFloat.D2) array).set(tIndex, cIndex, value);
	}
	in.close();
      } catch (IOException ex) {
	ex.printStackTrace();
      }

      ArrayFloat.D2 farray = ((ArrayFloat.D2) array);
      for (int j = 0; j < array.getShape()[1]; j++) {
	for (int i = 0; i < array.getShape()[0]; i++) {
	  float val = farray.get(i, j);
	  if (Float.isNaN(val)) {
	    // first time step so find and set to first good value
	    if (i == 0)
	      i = skipToFirstGoodValue(farray, j);
	    // interpolate missing values
	    else
	      i = interpolate(farray, j, i);
	  }
	}
      }
    }
    
    private int interpolate(ArrayFloat.D2 farray, int category, int start) {
      float startVal = farray.get(start - 1, category);
      float endVal = Float.NaN;
      int endIndex = array.getShape()[0];
      // know start is bad so skip it.
      for (int i = start + 1; i < array.getShape()[0]; i++) {
	float val = farray.get(i, category);
	if (!Float.isNaN(val)) {
	  endVal = val;
	  endIndex = i;
	  break;
	}
      }
      
      if (Float.isNaN(endVal)) {
	// never found a good val after a NaN
	// so set the rest to the last known good val.
	endIndex = array.getShape()[0];
	for (int i = start; i < array.getShape()[0]; i++) {
	  farray.set(i, category, startVal);
	}
      } else {
	float interval = endVal - startVal;
	float increment = interval / (endIndex - (start - 1));
	for (int i = start; i < endIndex; i++) {
	  farray.set(i, category, startVal + (increment * (i - start)));
	}
      }
      
      return endIndex;
      
    }

    private int skipToFirstGoodValue(ArrayFloat.D2 farray, int category) {
      // find the first non NaN value.
      int index = 0;
      float val = 0;
      for (int i = 0; i < array.getShape()[0]; i++) {
	val = farray.get(i, category);
	if (!Float.isNaN(val)) {
	  index = i;
	  break;
	}
      }

      /*
      for (int i = 0; i < index; i++) {
	farray.set(i, category, val);
      }
      */

      return index;
    }

    /*
     * (non-Javadoc)
     * 
     * @see anl.aida.data.AbstractDataFrame#getArray()
     */
    @Override
    public Array getArray() {
      if (array == null)
	init();
      return super.getArray();
    }

    /*
     * (non-Javadoc)
     * 
     * @see anl.aida.data.AbstractDataFrame#getArrayType()
     */
    @Override
    public Class<?> getArrayType() {
      if (array == null)
	init();
      return super.getArrayType();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * anl.aida.data.AbstractDataFrame#getDouble(anl.aida.data.DataFrameIndex)
     */
    @Override
    public double getDouble(DataFrameIndex index) {
      if (array == null)
	init();
      return super.getDouble(index);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * anl.aida.data.AbstractDataFrame#getFloat(anl.aida.data.DataFrameIndex)
     */
    @Override
    public float getFloat(DataFrameIndex index) {
      if (array == null)
	init();
      return super.getFloat(index);
    }

    /*
     * (non-Javadoc)
     * 
     * @see anl.aida.data.AbstractDataFrame#getIndex()
     */
    @Override
    public DataFrameIndex getIndex() {
      if (array == null)
	init();
      return super.getIndex();
    }

    /*
     * (non-Javadoc)
     * 
     * @see anl.aida.data.AbstractDataFrame#getInt(anl.aida.data.DataFrameIndex)
     */
    @Override
    public int getInt(DataFrameIndex index) {
      if (array == null)
	init();
      return super.getInt(index);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * anl.aida.data.AbstractDataFrame#getLong(anl.aida.data.DataFrameIndex)
     */
    @Override
    public long getLong(DataFrameIndex index) {
      if (array == null)
	init();
      return super.getLong(index);
    }

    /*
     * (non-Javadoc)
     * 
     * @see anl.aida.data.AbstractDataFrame#getShape()
     */
    @Override
    public int[] getShape() {
      if (array == null)
	init();
      return super.getShape();
    }

    /*
     * (non-Javadoc)
     * 
     * @see anl.aida.data.AbstractDataFrame#getSize()
     */
    @Override
    public long getSize() {
      if (array == null)
	init();
      return super.getSize();
    }

    /*
     * (non-Javadoc)
     * 
     * @see anl.aida.data.AbstractDataFrame#slice(anl.aida.data.Slice)
     */
    @Override
    public DataFrame slice(Slice slice) throws InvalidRangeException {
      if (array == null)
	init();
      return super.slice(slice);
    }
  }

  private static class AFDataset extends AbstractDataset {

    private String name;

    AFDataset(String name) {
      this.name = name;
    }

    private void addVar(Variable var) {
      this.addVariable(var);
    }

    private void addDataFrame(DataFrame frame) {
      this.frames.put(frame.getVariable(), frame);
    }

    /*
     * (non-Javadoc)
     * 
     * @see anl.aida.data.Dataset#getName()
     */
    @Override
    public String getName() {
      return name;
    }
  }

}
