package anl.aida.algo;

import anl.aida.util.VectorUtilities;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Builder for clusters.
 *
 */
public class ClustersBuilder<T> {

  private List<SimilarityCluster<T>> clusters = new ArrayList<SimilarityCluster<T>>();
  private Map<T, float[]> itemScores = new HashMap<T, float[]>();

  public void addCluster(float[] vals, String... labels) {
    clusters.add(new SimilarityCluster<T>(vals, labels));
  }

  /**
   * Adds an item to the specified cluster. The items similarity with that cluster
   * will be calculated using the itemVector and the cluster's centroid.
   *
   * @param item         the item to add
   * @param clusterIndex the index of the cluster to add the item to
   * @param itemVector   the item's vector
   */
  public void addItemToCluster(T item, int clusterIndex, double[] itemVector) {
    Cluster cluster = clusters.get(clusterIndex);
    float score = VectorUtilities.scalarProduct(cluster.getCentroid(), itemVector);
    addItemToCluster(item, clusterIndex, score);
  }

  /**
   * Adds an item to the specified cluster. The items similarity with that cluster
   * will be calculated using the itemVector and the cluster's centroid.
   *
   * @param item         the item to add
   * @param clusterIndex the index of the cluster to add the item to
   * @param score        the item's score
   */
  public void addItemToCluster(T item, int clusterIndex, float score) {
    clusters.get(clusterIndex).addItem(item, score);
    addItemScore(item, clusterIndex, score);
  }

  public int clusterCount() {
    return clusters.size();
  }

  /**
   * Calculates the specified items similarity score for the specified cluster.
   *
   * @param item         the item to add
   * @param clusterIndex the index of the cluster to add the item to
   * @param itemVector   the item's vector
   */
  public void itemScore(T item, int clusterIndex, double[] itemVector) {
    Cluster cluster = clusters.get(clusterIndex);
    float score = VectorUtilities.scalarProduct(cluster.getCentroid(), itemVector);
    addItemScore(item, clusterIndex, score);
  }

  private void addItemScore(T item, int cluster, float score) {
    float[] scores = itemScores.get(item);
    if (scores == null) {
      scores = new float[clusters.size()];
      itemScores.put(item, scores);
    }
    scores[cluster] = score;
  }

  /**
   * Builds a Clusters instance out of the data that has been added
   * between any calls to buildCluster.
   *
   * @return the built Clusters instance
   */
  public Clusters<T> buildClusters() {
    DefClusters<T> clusters = new DefClusters<T>();
    clusters.clusters.addAll(this.clusters);
    clusters.itemScores = itemScores;
    this.itemScores = null;
    this.clusters = null;

    return clusters;

  }


  private static class DefClusters<T> implements Clusters<T> {

    private List<SimilarityCluster<T>> clusters = new ArrayList<SimilarityCluster<T>>();
    private Map<T, float[]> itemScores = new HashMap<T, float[]>();


    /**
     * Gets the similarity score of the specified item
     * with the specified cluster centriod.
     *
     * @param item         the item whose score we want
     * @param clusterIndex the index of the cluster containing the relevant centroid
     * @return the similarity score of the specified item
     *         with the specified cluster centriod.
     */
    public float getScore(T item, int clusterIndex) {
      float[] scores = itemScores.get(item);
      return scores[clusterIndex];
    }

    /**
     * Gets the number of clusters in this collection.
     *
     * @return the number of clusters in this collection.
     */
    public int size() {
      return clusters.size();
    }

    /**
     * Gets the total number of clustered items.
     *
     * @return the total number of clustered items.
     */
    public int itemCount() {
      return itemScores.size();
    }

    /**
     * Gets an iterable over all the clustered items in this collection.
     * Each item is contained by one or more clusters.
     *
     * @return an iterable over all the clustered items in this collection.
     */
    public Iterable<T> items() {
      return itemScores.keySet();
    }

    /**
     * Gets an iterable over all the Clusters contained by this collection.
     *
     * @return an iterable over all the Clusters contained by this collection.
     */
    public Iterable<? extends Cluster<T>> clusters() {
      return clusters;
    }

    /**
     * Gets the cluster at the specified index.
     *
     * @param index the index of the cluster
     * @return the cluster at the specified index.
     */
    public Cluster<T> getCluster(int index) {
      return clusters.get(index);
    }
  }
}
