package anl.aida.algo;

import anl.aida.util.ObjScore;

import java.util.LinkedHashMap;
import java.util.Map;

/**
 * Encapsulates a centroid vector and a collection
 * of items of type T and their similarity (0 to 1) with that centroid
 * vector.
 *
 */
public class SimilarityCluster<T> implements Cluster<T> {

  private float[] vector;
  private String labelString;
  private String[] labels;

  private Map<T, ObjScore<T>> items = new LinkedHashMap<T, ObjScore<T>>();

  /**
   * Creates a SimilarityCluster with the specified vector and label.
   *
   * @param vector the centroids vector
   * @param labels the centroids labels
   */
  public SimilarityCluster(float[] vector, String... labels) {
    this.vector = vector;
    this.labels = labels;
    StringBuilder b = new StringBuilder();
    boolean addSpace = false;
    for (String label : labels) {
      if (addSpace) b.append(" ");
      b.append(label);
      addSpace = true;
    }
    labelString = b.toString();

  }

  protected void addItem(T item, float similarity) {
    items.put(item, new ObjScore<T>(similarity, item));
  }

  /**
   * Gets the label for this cluster. This can be a list of keywords
   * or a descriptive label, etc.
   *
   * @return the label for this cluster
   */
  public String getCentroidLabel() {
    return labelString;
  }

  /**
   * The number of items in this cluster.
   *
   * @return the number of items in this cluster
   */
  public int getItemCount() {
    return items.size();
  }

  /**
   * Gets the vector for the cluster centroid.
   *
   * @return the vector for the cluster centroid
   */
  public float[] getCentroid() {
    return vector;
  }

  /**
   * Gets the similarity score for the specified item
   * with respect to the cluster centroid.
   *
   * @param item the item whose score we want.
   * @return the similarity score for the specified item
   *         with respect to the cluster centroid, or -1 if the
   *         item is not found.
   */
  public float getSimilarity(T item) {
    ObjScore score = items.get(item);
    if (score == null) {
      System.out.printf("%s not in cluster %s%n", item.toString(), labelString);
      return -1;
    } else return score.getScore();
  }

  public Iterable<ObjScore<T>> items() {
    return items.values();
  }


}
