package anl.aida.bubble;

import java.awt.Font;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.List;

import edu.umd.cs.piccolo.PCanvas;
import edu.umd.cs.piccolo.PLayer;
import edu.umd.cs.piccolo.nodes.PText;
import edu.umd.cs.piccolox.nodes.PComposite;

@SuppressWarnings("serial")
public abstract class AbstractAxis extends PComposite {

  protected DecimalFormat highlightFormatter = new DecimalFormat("###,###,###,###.##");
  protected DecimalFormat labelFormatter = new DecimalFormat();

  protected NumericRange range, origRange;
  protected String label;
  protected PCanvas canvas;
  protected int insets = 20;

  protected float pixelExtent;
  protected List<Float> tickLocations = new ArrayList<Float>();
  protected AxisHighlight highlight = new AxisHighlight();
  protected PText axisLabel = new PText("");

  public AbstractAxis(NumericRange range, String label) {
    this.label = label;
    this.range = new NumericRange(range);
    this.origRange = new NumericRange(this.range);
    highlightFormatter.setGroupingUsed(true);

    addPropertyChangeListener("parent", new PropertyChangeListener() {
      @Override
      public void propertyChange(PropertyChangeEvent evt) {
	if (canvas == null) {
	  PLayer layer = (PLayer) evt.getNewValue();
	  canvas = (PCanvas) layer.getCamera(0).getComponent();
	  layout();
	}
      }
    });

    axisLabel.setFont(axisLabel.getFont().deriveFont(Font.BOLD));
  }

  protected void initRange(double extra) {
    this.range = new NumericRange(
	new NumericRange(range.getStart() - extra, range.getEnd() + extra));
    this.origRange = new NumericRange(this.range);
  }

  protected abstract void layout();

  public abstract void highlight(double val);

  public void removeHighlight() {
    removeChild(highlight);
  }

  /**
   * Gets the pixel location for the data value.
   * 
   * @param val
   *          the value to get the pixel location for
   * @return the pixel location for the data value.
   */
  public abstract float getPixelLocation(double val);

  /**
   * Gets the axis value for the pixel location.
   * 
   * @param pixel
   *          the pixel to get the axis value for
   * 
   * @return the axis value for the pixel location.
   */
  public abstract double getValueForPixel(double pixel);

  /**
   * Zooms the axis such that it shows the specified range.
   * 
   * @param range
   *          the range to zoom to.
   */
  public void zoomTo(NumericRange range) {
    this.range = range;
    layout();
  }

  /**
   * Resets axis to its original range.
   */
  public void resetRange() {
    this.range = origRange;
    layout();
  }

  /**
   * Gets the pixel location for the data value in a linear scale.
   * 
   * @param val
   *          the value to get the pixel location for
   * @return the pixel location for the data value.
   */
  protected float getPixelLocationLin(double val) {
    return (float) ((val - range.getStart()) / range.getLength() * pixelExtent);
  }

  protected double getValueForPixelLin(double pixel) {
    return pixel / pixelExtent * range.getLength() + range.getStart();
  }

  /**
   * Gets the pixel location for the data value in a log scale.
   * 
   * @param val
   *          the value to get the pixel location for
   * @return the pixel location for the data value in a log scale.
   */
  protected float getPixelLocationLog(double val) {
    val = val - range.getStart();
    if (val > 0) {
      val = Math.log10(val);
      double extent = Math.log10(range.getLength());
      return (float) (val / extent * pixelExtent);
    } else {
      return 0;
    }
  }

  protected double getValueForPixelLog(double pixel) {
    double extent = Math.log10(range.getLength());
    double val = pixel / pixelExtent * extent;
    return Math.pow(10, val) + range.getStart();
  }

  public Iterable<Float> getTickLocations() {
    return tickLocations;
  }

  public float getPixelExtent() {
    return pixelExtent;
  }

  /**
   * @return the label
   */
  public String getLabel() {
    return label;
  }

  /**
   * @param label
   *          the label to set
   */
  public void setLabel(String label) {
    this.label = label;
  }
}