package anl.aida.bubble;


/**
 * Data access interface for a bubble chart. The shape of the
 * data is each category is a row and each column is some
 * numeric value of that category for that column. That shape
 * is then repeated for each timestep.
 * 
 */
public interface BubbleDataset {
  
  /**
   * Gets an iterable over the categories in this dataset.
   * 
   * @return an iterable over the categories in this dataset.
   */
  Iterable<String> categories();
  
  /**
   * Gets the number of categories in this dataset. 
   * 
   * @return the number of categories in this dataset.
   */
  int getCategoryCount();
  
  /**
   * Gets the number of timesteps.
   * 
   * @return the number of timesteps.
   */
  int getTimeStepCount();
  
  /**
   * Gets the formatted timestep (e.g. a formatted date,
   * a year, etc.)
   * 
   * @param timestep the formatted timestep
   * @return
   */
  String getFormattedTimeStep(int timestep);
  
  /**
   * Gets the index of the specified category.
   * 
   * @param category the category whose index we want
   * 
   * @return the index of the specified category.
   */
  int getCategoryIndex(String category);
  
  /**
   * Gets the x value for the particular category at the specified timestep.
   * 
   * @param categoryIndex the category index
   * @param timestep the timestep
   * 
   * @return the x value for the particular category at the specified timestep.
   */
  double getXValue(int categoryIndex, int timestep);
  
  /**
   * Gets the y value for the particular category at the specified timestep.
   * 
   * @param categoryIndex the category index
   * @param timestep the timestep
   * 
   * @return the y value for the particular category at the specified timestep.
   */
  double getYValue(int categoryIndex, int timestep);
  
  /**
   * Gets the z value for the particular category at the specified timestep.
   * 
   * @param categoryIndex the category index
   * @param timestep the timestep
   * 
   * @return the z value for the particular category at the specified timestep.
   */
  double getZValue(int categoryIndex, int timestep);
  
  /**
   * Gets the range of x values over all timesteps and categories.
   * 
   * @return the range of x values over all timesteps and categories.
   */
  NumericRange getXRange();
  
  /**
   * Gets the range of y values over all timesteps and categories.
   * 
   * @return the range of y values over all timesteps and categories.
   */
  NumericRange getYRange();
  
  /**
   * Gets the range of z values over all timesteps and categories.
   * 
   * @return the range of z values over all timesteps and categories.
   */
  NumericRange getZRange();

}
