package anl.aida.bubble;

import java.awt.Color;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.List;

import edu.umd.cs.piccolo.PCanvas;
import edu.umd.cs.piccolo.PNode;
import edu.umd.cs.piccolo.nodes.PPath;
import edu.umd.cs.piccolo.nodes.PText;
import edu.umd.cs.piccolo.util.PBounds;
import edu.umd.cs.piccolo.util.PPaintContext;

public class BubblePlot {

  private PCanvas canvas;
  private RangeAxis rAxis;
  private AbstractDomainAxis dAxis;
  private PNode plot = new PNode();
  private DefaultBubbleRenderer bubbles;
  private BubbleDataset data;
  private int timestep = 0;
  private int[] catIndices;
  private NumericRange zRange;
  private PText dateNode = new PText();
  // if true mouse over events on bubbles
  // should be ignored.
  private boolean ignoreMouseOver = false;
  private boolean initialized = false;

  private List<BubbleSelectionListener> listeners = new ArrayList<BubbleSelectionListener>();

  private interface Updater {
    PNode update(int i, float px, float py, float scale);
  }

  private Updater animateUpdate = new Updater() {
    public PNode update(int i, float px, float py, float scale) {
      return bubbles.updateAnimate(i, px, py, scale);
    }
  };

  private Updater offsetUpdate = new Updater() {
    public PNode update(int i, float px, float py, float scale) {
      return bubbles.update(i, px, py, scale);
    }
  };

  public BubblePlot(PCanvas canvas, BubbleDataset dataset) {
    this.canvas = canvas;
    data = dataset;
    zRange = dataset.getZRange();
    canvas.setPanEventHandler(null);
    canvas.setZoomEventHandler(null);
    canvas.setDefaultRenderQuality(PPaintContext.HIGH_QUALITY_RENDERING);
    canvas.setInteractingRenderQuality(PPaintContext.HIGH_QUALITY_RENDERING);
    canvas.setAnimatingRenderQuality(PPaintContext.HIGH_QUALITY_RENDERING);
    catIndices = new int[dataset.getCategoryCount()];

    int i = 0;
    for (String category : dataset.categories()) {
      catIndices[i++] = dataset.getCategoryIndex(category);
    }

    rAxis = new RangeAxis(data.getYRange(), "y axis");
    dAxis = new DomainAxis(data.getXRange(), "x axis");

    canvas.addInputEventListener(new Zoomer(this));
    canvas.addComponentListener(new ComponentAdapter() {
      @Override
      public void componentResized(ComponentEvent e) {
	if (bubbles == null)
	  init();
	layout();
	updateDateNode(data.getFormattedTimeStep(timestep));
      }
    });
  }

  private void init() {
    bubbles = new DefaultBubbleRenderer(data, new BubbleTooltip(this));
    canvas.getLayer().addChild(dAxis);
    canvas.getLayer().addChild(rAxis);
    canvas.getLayer().addChild(plot);
    bubbles.addLayers(plot);

    initialized = true;
  }

  PCanvas getCanvas() {
    return canvas;
  }

  PNode getPlot() {
    return plot;
  }

  /**
   * Gets whether or not mouse over events on bubbles should be ignored.
   * 
   * @return whether or not mouse over events on bubbles should be ignored.
   */
  public boolean isIgnoreMouseOver() {
    return ignoreMouseOver;
  }

  /**
   * Sets whether or not mouse over events on bubbles should be ignored.
   * 
   * @param ignoreMouseOver
   *          the new value
   */
  public void setIgnoreMouseOver(boolean ignoreMouseOver) {
    this.ignoreMouseOver = ignoreMouseOver;
  }

  /**
   * Adds the specified listener to listen for bubble selection events.
   * 
   * @param listener
   *          the listener to add
   */
  public void addSelectionListener(BubbleSelectionListener listener) {
    listeners.add(listener);
  }

  /**
   * Selects or unselects the bubble with the specified category index.
   * 
   * @param categoryIndex
   *          in the index of the bubble
   * @param selected
   *          whether or not the bubble is selected or unselected.
   */
  public void selectBubble(int categoryIndex, boolean selected) {
    bubbles.selectBubble(categoryIndex, selected);
    for (BubbleSelectionListener listener : listeners) {
      if (selected)
	listener.bubbleSelected(categoryIndex);
      else
	listener.bubbleUnselected(categoryIndex);
    }
  }

  public void updateTimeStep(int timestep, boolean animate) {
    if (timestep != this.timestep) {
      this.timestep = timestep;
      updateDateNode(data.getFormattedTimeStep(this.timestep));
      if (animate)
	updateBubbles(animateUpdate);
      else
	updateBubbles(offsetUpdate);
    }
  }

  private void updateDateNode(String text) {
    PBounds bounds = plot.getUnionOfChildrenBounds(null);
    dateNode.setText(text);
    PBounds textBounds = dateNode.getBoundsReference();
    AffineTransform trans = AffineTransform.getScaleInstance((bounds.width - 200)
	/ textBounds.width, (bounds.height - 200) / textBounds.height);
    dateNode.setTransform(trans);
    textBounds = dateNode.computeFullBounds(null);
    dateNode.setOffset(bounds.getCenterX() - textBounds.getCenterX(), bounds.getCenterY()
	- textBounds.getCenterY());
  }

  private void updateBubbles(Updater updater) {
    for (int i : catIndices) {
      double xVal = data.getXValue(i, timestep);
      double yVal = data.getYValue(i, timestep);
      double zVal = data.getZValue(i, timestep);

      if (!(Double.isNaN(xVal) || Double.isNaN(yVal) || Double.isNaN(zVal))) {
	float scaleVal = zRange.getLength() == 0 ? .1f
	    : (float) ((zVal - zRange.getStart()) / zRange.getLength());

	float px = dAxis.getPixelLocation(xVal);
	float py = rAxis.getPixelLocation(yVal);

	PNode node = updater.update(i, px, py, scaleVal);
	BubbleData data = (BubbleData) node.getAttribute(BubbleConstants.DATA);
	data.update(xVal, yVal, zVal);
      } else {
	bubbles.setNodeVisible(i, false);
      }
    }
  }

  /**
   * @return the rAxis
   */
  public RangeAxis getRangeAxis() {
    return rAxis;
  }

  /**
   * @param axis
   *          the rAxis to set
   */
  public void setRangeAxis(RangeAxis axis) {
    removeChild(rAxis);
    rAxis = axis;
    if (initialized) {
      canvas.getLayer().addChild(rAxis);
      layout();
    }
  }

  /**
   * Gets the domain axis.
   * 
   * @return the dAxis
   */
  public AbstractDomainAxis getDomainAxis() {
    return dAxis;
  }

  /**
   * Sets the domain axis.
   * 
   * @param axis
   *          the dAxis to set
   */
  public void setDomainAxis(AbstractDomainAxis axis) {
    removeChild(dAxis);
    dAxis = axis;
    if (initialized) {
      canvas.getLayer().addChild(dAxis);
      layout();
    }
  }

  private void removeChild(PNode node) {
    if (initialized)
      this.canvas.getLayer().removeChild(dAxis);
  }

  /**
   * Zoom to the specified bounds. The bounds are in camera coordinates.
   * 
   * @param bounds
   *          the bounds to zoom to
   */
  public void zoom(Rectangle2D bounds) {
    Rectangle2D boundsCopy = new Rectangle2D.Double();
    boundsCopy.setRect(bounds);

    dAxis.globalToLocal(bounds);
    double minX = dAxis.getValueForPixel(bounds.getMinX());
    double maxX = dAxis.getValueForPixel(bounds.getMaxX());
    dAxis.zoomTo(new NumericRange(minX, maxX));

    rAxis.globalToLocal(boundsCopy);
    // note minY is > maxY
    double maxY = rAxis.getValueForPixel(boundsCopy.getMinY());
    double minY = rAxis.getValueForPixel(boundsCopy.getMaxY());
    rAxis.zoomTo(new NumericRange(minY, maxY));

    layout();
  }

  /**
   * Reset the bounds of the plot to its original extent.
   */
  public void resetBounds() {
    dAxis.resetRange();
    rAxis.resetRange();
    layout();
  }

  private void layout() {
    PBounds rBounds = rAxis.getUnionOfChildrenBounds(null);
    PBounds dBounds = dAxis.getUnionOfChildrenBounds(null);
    float yEnd = (float) (canvas.getBounds().getHeight() - dBounds.getHeight());
    double xOffset = rBounds.width + 8;
    dAxis.setOffset(xOffset, yEnd);
    rAxis.setYEnd(yEnd);
    rAxis.setOffset(xOffset, 0);

    rAxis.layout();
    dAxis.layout();

    plot.removeAllChildren();
    plot.setOffset(xOffset, 0);

    PPath grid = new PPath();
    grid.setStrokePaint(Color.LIGHT_GRAY);
    grid.invalidatePaint();
    for (Float tick : dAxis.getTickLocations()) {
      grid.moveTo(tick, yEnd);
      grid.lineTo(tick, yEnd - rAxis.getPixelExtent());
    }

    for (Float tick : rAxis.getTickLocations()) {
      grid.moveTo(0, tick);
      grid.lineTo(dAxis.getPixelExtent(), tick);
    }
    grid.setPickable(false);
    plot.addChild(grid);

    dateNode = new PText();
    dateNode.setTextPaint(Color.LIGHT_GRAY);
    dateNode.setTransparency(.5f);
    dateNode.setPickable(false);
    updateDateNode(data.getFormattedTimeStep(timestep));
    plot.addChild(dateNode);

    updateBubbles(offsetUpdate);
    bubbles.addLayers(plot);

    /*
     * PBounds b = plot.getUnionOfChildrenBounds(null); PPath path =
     * PPath.createRectangle((float)b.x,(float) b.y, (float) b.width,
     * (float)b.height); path.setPaint(Color.RED); plot.addChild(path);
     */
  }

}
