package anl.aida.bubble;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;

import anl.aida.data.DataFrame;
import anl.aida.data.DataFrameIndex;
import anl.aida.data.DataFrameStringAxis;

/**
 * Default implementation of a BubbleDataset that adapts DataFrames to the
 * BubbleDataset interface.
 * 
 */
public class DefaultBubbleDataset implements BubbleDataset {

  private static class CategoryEntry {
    int xIndex, yIndex, zIndex;
    String category;

    CategoryEntry() {
      xIndex = yIndex = zIndex = -1;
    }
  }

  private class CategoryIter implements Iterable<String>, Iterator<String> {

    private Iterator<CategoryEntry> iter;

    CategoryIter() {
      iter = categories.iterator();
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.util.Iterator#hasNext()
     */
    @Override
    public boolean hasNext() {
      return iter.hasNext();
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.util.Iterator#next()
     */
    @Override
    public String next() {
      return iter.next().category;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.util.Iterator#remove()
     */
    @Override
    public void remove() {
      throw new UnsupportedOperationException();

    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Iterable#iterator()
     */
    @Override
    public Iterator<String> iterator() {
      return this;
    }
  }

  private DataFrame xFrame, yFrame, zFrame;
  private int[] offsets = new int[3];
  private int tsCount;
  private List<CategoryEntry> categories = new ArrayList<CategoryEntry>();
  private Calendar cal = GregorianCalendar.getInstance();
  private NumericRange xRange, yRange, zRange;

  private static int CALENDAR_FIELD = Calendar.YEAR;

  /**
   * @param xFrame
   * @param yFrame
   * @param zFrame
   */
  public DefaultBubbleDataset(DataFrame xFrame, DataFrame yFrame, DataFrame zFrame) {
    this(xFrame, yFrame, zFrame, -1);
  }

  /**
   * @param xFrame
   * @param yFrame
   * @param zFrame
   * @param categoryCount
   *          the number of categories to display. -1 means show all.
   */
  public DefaultBubbleDataset(DataFrame xFrame, DataFrame yFrame, DataFrame zFrame,
      int categoryCount) {
    this.xFrame = xFrame;
    this.yFrame = yFrame;
    this.zFrame = zFrame;

    // reconcile the timesteps

    cal.setTime(xFrame.getAxes().getStartDate());
    int maxDate = cal.get(CALENDAR_FIELD);
    int xVal = maxDate;
    cal.setTime(yFrame.getAxes().getStartDate());
    int yVal = cal.get(CALENDAR_FIELD);
    maxDate = Math.max(maxDate, yVal);

    if (zFrame != null) {
      cal.setTime(zFrame.getAxes().getStartDate());
      int zVal = cal.get(CALENDAR_FIELD);
      maxDate = Math.max(maxDate, zVal);
      offsets[2] = maxDate - zVal;
    }
    offsets[0] = maxDate - xVal;
    offsets[1] = maxDate - yVal;

    cal.setTime(xFrame.getAxes().getEndDate());
    int minDate = cal.get(CALENDAR_FIELD);
    cal.setTime(yFrame.getAxes().getEndDate());
    minDate = Math.min(minDate, cal.get(CALENDAR_FIELD));
    if (zFrame != null) {
      cal.setTime(zFrame.getAxes().getEndDate());
      minDate = Math.min(minDate, cal.get(CALENDAR_FIELD));
    }
    // this looks backwards but max is the max of the start
    // and min is min of the end so the min will be greater
    // than the max
    tsCount = minDate - maxDate + 1;

    // reconcile the categories
    Map<String, CategoryEntry> map = new HashMap<String, CategoryEntry>();
    DataFrameStringAxis axis = (DataFrameStringAxis) xFrame.getAxes().getNonTimeAxis();
    for (int i = 0; i < axis.getExtent(); i++) {
      String category = axis.getValueAsString(i);
      CategoryEntry entry = new CategoryEntry();
      entry.category = category;
      entry.xIndex = i;
      map.put(category.toLowerCase(), entry);
    }

    axis = (DataFrameStringAxis) yFrame.getAxes().getNonTimeAxis();
    for (int i = 0; i < axis.getExtent(); i++) {
      String category = axis.getValueAsString(i);
      CategoryEntry entry = map.get(category.toLowerCase());
      if (entry != null) {
	entry.yIndex = i;
      }
    }

    if (zFrame != null) {
      axis = (DataFrameStringAxis) zFrame.getAxes().getNonTimeAxis();
      for (int i = 0; i < axis.getExtent(); i++) {
	String category = axis.getValueAsString(i);
	CategoryEntry entry = map.get(category.toLowerCase());
	if (entry != null) {
	  entry.zIndex = i;
	}
      }
    } 

    for (Iterator<Map.Entry<String, CategoryEntry>> iter = map.entrySet().iterator(); iter
	.hasNext();) {
      Map.Entry<String, CategoryEntry> entry = iter.next();
      CategoryEntry cat = entry.getValue();
      if (cat.yIndex == -1 || (cat.zIndex == -1 && zFrame != null))
	iter.remove();
    }

    // map should now contain only categories common to all frames
    for (CategoryEntry entry : map.values()) {
      categories.add(entry);
    }

    if (categoryCount > 0 && categoryCount < categories.size()) {
      Random random = new Random();
      Set<CategoryEntry> tmp = new HashSet<CategoryEntry>();
      for (int i = 0; i < categoryCount; i++) {
	int index = random.nextInt(categories.size() - 1);
	tmp.add(categories.get(index));
      }
      categories.clear();
      categories.addAll(tmp);
    }

    initRanges();
  }

  private void initRanges() {
    double xMin, yMin, zMin;
    double xMax, yMax, zMax;
    xMin = yMin = zMin = Double.MAX_VALUE;
    xMax = yMax = zMax = Double.MIN_VALUE;
    for (int ts = 0, n = getTimeStepCount(); ts < n; ts++) {
      for (int index = 0; index < categories.size(); index++) {
	double xVal = getXValue(index, ts);
	if (!Double.isNaN(xVal)) {
	  xMin = Math.min(xVal, xMin);
	  xMax = Math.max(xVal, xMax);
	}

	double yVal = getYValue(index, ts);
	if (!Double.isNaN(yVal)) {
	  yMin = Math.min(yVal, yMin);
	  yMax = Math.max(yVal, yMax);
	}

	double zVal = getZValue(index, ts);
	if (!Double.isNaN(zVal)) {
	  zMin = Math.min(zVal, zMin);
	  zMax = Math.max(zVal, zMax);
	}
      }
    }
    xRange = new NumericRange(xMin, xMax);
    yRange = new NumericRange(yMin, yMax);
    zRange = new NumericRange(zMin, zMax);
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.bubble.BubbleDataset#categories()
   */
  @Override
  public Iterable<String> categories() {
    return new CategoryIter();
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.bubble.BubbleDataset#getCategoryIndex(java.lang.String)
   */
  @Override
  public int getCategoryIndex(String category) {
    for (int i = 0, n = categories.size(); i < n; i++) {
      if (categories.get(i).category.equals(category))
	return i;
    }
    return -1;
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.bubble.BubbleDataset#getFormattedTimeStep(int)
   */
  @Override
  public String getFormattedTimeStep(int timestep) {
    timestep = offsets[0] + timestep;
    cal.setTime(xFrame.getAxes().getDate(timestep));
    return String.valueOf(cal.get(CALENDAR_FIELD));
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.bubble.BubbleDataset#getCategoryCount()
   */
  @Override
  public int getCategoryCount() {
    return categories.size();
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.bubble.BubbleDataset#getTimeStepCount()
   */
  @Override
  public int getTimeStepCount() {
    return tsCount;
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.bubble.BubbleDataset#getXValue(int, int)
   */
  @Override
  public double getXValue(int categoryIndex, int timestep) {
    timestep = offsets[0] + timestep;
    int catIndex = categories.get(categoryIndex).xIndex;
    DataFrameIndex index = xFrame.getIndex();
    index.set(timestep, catIndex);
    return xFrame.getDouble(index);
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.bubble.BubbleDataset#getYValue(int, int)
   */
  @Override
  public double getYValue(int categoryIndex, int timestep) {
    timestep = offsets[1] + timestep;
    int catIndex = categories.get(categoryIndex).yIndex;
    DataFrameIndex index = yFrame.getIndex();
    index.set(timestep, catIndex);
    return yFrame.getDouble(index);
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.bubble.BubbleDataset#getZValue(int, int)
   */
  @Override
  public double getZValue(int categoryIndex, int timestep) {
    if (zFrame == null) return 1;
    
    timestep = offsets[2] + timestep;
    int catIndex = categories.get(categoryIndex).zIndex;
    DataFrameIndex index = zFrame.getIndex();
    index.set(timestep, catIndex);
    return zFrame.getDouble(index);
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.bubble.BubbleDataset#getXRange()
   */
  @Override
  public NumericRange getXRange() {
    return xRange;
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.bubble.BubbleDataset#getYRange()
   */
  @Override
  public NumericRange getYRange() {
    return yRange;
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.bubble.BubbleDataset#getZRange()
   */
  @Override
  public NumericRange getZRange() {
    return zRange;
  }

}
