package anl.aida.client;

import anl.aida.util.DateRange;
import anl.aida.util.DateRangeProducer;

import java.io.File;
import java.util.HashMap;
import java.util.Map;

/**
 * Encapsulates a Cache.
 *
 */
public class Cache {

  private File luceneLocation, cacheDirectory;
  private Map<DateRange, CacheElement> items = new HashMap<DateRange, CacheElement>();
  private DateRangeProducer ranger;
  private int totalDocs;
  private Keywords keywords = new Keywords();

  /**
   * Creates a cache.
   *
   * @param luceneLocation the location of the lucene index from which the cache was created
   * @param ranger         a DateRangeProducer that can produce the date ranges of the individual cache elements
   */
  public Cache(File luceneLocation, File cacheDirectory, DateRangeProducer ranger) {
    this.luceneLocation = luceneLocation;
    this.cacheDirectory = cacheDirectory;
    this.ranger = ranger;
  }
  
  /**
   * Gets the directory where the cache resides.
   * 
   * @return the directory where the cache resides.
   */
  public File getCacheDirectory() {
    return cacheDirectory;
  }

  /**
   * Gets the location of this cache's lucene index.
   * 
   * @return the location of this cache's lucene index.
   */
  public File getLuceneLocation() {
    return luceneLocation;
  }

  /**
   * Gets the total number of docs in this cache.
   * 
   * @return the total number of docs in this cache.
   */
  public int getTotalDocs() {
    return totalDocs;
  }

  /**
   * Adds the specified CacheElement to this cache.
   * 
   * @param item the element to add
   */
  public void addCacheElement(CacheElement item) {
    DateRange range = item.getHeader().getDateRange();
    items.put(range, item);
    if (range.equals(new DateRange(ranger.getStart(), ranger.getEnd()))) {
      totalDocs = item.getDocumentCount();
    }
  }

  /**
   * Gets the number of cache elements  in this cache.
   * 
   * @return the number of cache elements  in this cache.
   */
  public int size() {
    return items.size();
  }

  /**
   * Gets the element by its numeric id.
   * 
   * @param id the id of the element to get
   * @return the element by its numeric id.
   */
  public CacheElement getElement(int id) {
    for (CacheElement item : items.values()) {
      if (item.getID() == id) return item;
    }
    return null;
  }

  /**
   * Gets the element that covers the specified date range.
   * 
   * @param range the range of the element to get
   * @return the element that covers the specified date range.
   */
  public CacheElement getElement(DateRange range) {
    return items.get(range);
  }

  /**
   * Gets an iteratble over all the elements in this cache.
   * 
   * @return an iteratble over all the elements in this cache.
   */
  public Iterable<CacheElement> elements() {
    return items.values();
  }

  /**
   * Gets the DateRangeProducer that covers the date range interval
   * of this Cache.
   * 
   * @return the DateRangeProducer that covers the date range interval
   * of this Cache.
   */
  public DateRangeProducer getRanger() {
    return ranger;
  }
  
  /**
   * Adds the specified keyword to the specified keyword set
   * in this cache.
   * 
   * @param keywordSet the set to add the keyword to
   * @param keyword the keyword to add
   */
  public void add(String keywordSet, String keyword) {
    keywords.add(keywordSet, keyword);
  }
  
  /**
   * Gets the keywords associated with this cache.
   * 
   * @return the keywords associated with this cache.
   */
  public Keywords getKeywords() {
    return keywords;
  }
  
  /**
   * Sets the keywords for this cache.
   * 
   * @param keywords the new keywords for this cache
   */
  public void setKeywords(Keywords keywords) {
    this.keywords = keywords;
  }
}
