package anl.aida.client;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import simphony.util.messages.MessageCenter;
import anl.aida.algo.Clusterer;
import anl.aida.client.doc.DefaultDocumentsModel;
import anl.aida.client.doc.DocumentsModel;
import anl.aida.core.DocumentID;
import anl.aida.core.KeywordGenerator;
import anl.aida.core.MIFSDocument;
import anl.aida.core.STCHeader;
import anl.aida.core.STCReader;
import anl.aida.core.ScoredTermsCollection;
import anl.aida.sv.SVIndexBuilder;
import anl.aida.sv.SVKeywordGenerator;
import anl.aida.sv.SemanticVectorStore;
import anl.aida.util.AIDAException;

/**
 * Encapsulates details about a timestamped section of the cache.
 *
 */
public class CacheElement {

  private static final MessageCenter center = MessageCenter.getMessageCenter(CacheElement.class);
  public static final String SV_FILES_DIR = "vectors";

  private File stcLocation;
  private STCHeader header;

  private ScoredTermsCollection stc;
  private KeywordGenerator keyGen;
  private SemanticVectorStore svStore;
  private int id;
  private Keywords keywords = new Keywords();

  private List<CacheDocumentSet> docSets = new ArrayList<CacheDocumentSet>();

  public CacheElement(int id, File stcLocation, STCHeader header) {
    this.stcLocation = stcLocation;
    this.header = header;
    this.id = id;

    docSets.add(new AllDocs(new DocSetID(id, 0), this));
  }
  
  /**
   * Gets the keywords for this CacheElement. This may contain
   * keywords with no docs in this element.
   * 
   * @return the keywords for this CacheElement. This may contain
   * keywords with no docs in this element.
   */
  public Keywords getKeywords() {
    return keywords;
  }

  /**
   * Sets the keywords for this cache element.
   * 
   * @param keywords the new keywords
   */
  public void setKeyword(Keywords keywords) {
    this.keywords = keywords;
  }



  /**
   * Gets the document set at the specified index.
   *
   * @param index the index
   * @return the document set at the specified index.
   */
  public CacheDocumentSet getDocumentSet(int index) {
    return docSets.get(index);
  }

  /**
   * Gets the header associated with this cache element.
   *
   * @return the header associated with this cache element.
   */
  public STCHeader getHeader() {
    return header;
  }

  /**
   * Releases any resources (STC etc.) acquired by this
   * CacheElement.
   */
  public void release() {
    stc = null;
    svStore = null;
    keyGen = null;
  }
  
  private void initSTC() {
    try {
      STCReader stcReader = new STCReader();
      File location = getSTCLocation();
      String path = location.getAbsolutePath();

      stc = stcReader.readCollection(path);
    } catch (AIDAException ex) {
      center.error("Error loading cache files.", ex);
    } catch (IOException ex) {
      center.error("Error loading cache files.", ex);
    }
  }
  
  private void initSV() {
    if (stc == null) initSTC();
    try {
      File[] svFiles = getSVFiles(getSTCLocation(), stc);
      svStore = new SemanticVectorStore(svFiles[1].getAbsolutePath(), svFiles[0].getAbsolutePath());
      keyGen = new SVKeywordGenerator(svStore);
    } catch (IOException ex) {
      center.error("Error loading cache files.", ex);
    }
  }


  public int getID() {
    return id;
  }

  public ScoredTermsCollection getSTC() {
    if (stc == null) initSTC();
    return stc;
  }

  public SemanticVectorStore getSVS() {
    if (svStore == null) initSV();
    return svStore;
  }

  public KeywordGenerator getKeywordGenerator() {
    if (keyGen == null) initSV();
    return keyGen;
  }

  public Iterable<CacheDocumentSet> documentSets() {
    return docSets;
  }

  public void addDocumentSet(CacheDocumentSet set) {
    docSets.add(set);
  }

  public int getDocumentSetCount() {
    return docSets.size();
  }

  // terms is at index 0, docs at index 1
  private File[] getSVFiles(File stcLoc, ScoredTermsCollection stc) throws IOException {
    File[] files = new File[2];
    File parent = new File(stcLoc.getParentFile(), SV_FILES_DIR);
    if (!parent.exists()) {
      if (!parent.mkdirs()) throw
              new IOException("Error creating vectors directory '" + parent.getAbsoluteFile() + "'");
    }

    String stcName = stcLoc.getName().substring(0, stcLoc.getName().length() - 4);
    SVIndexBuilder builder = new SVIndexBuilder(20, new KeywordDocCountFilter(10, keywords));
    files[0] = new File(parent, stcName + "_terms.bin");
    files[1] = new File(parent, stcName + "_docs.bin");
    if (!files[0].exists())
      builder.buildIndex(files[0].getAbsolutePath(), files[1].getAbsolutePath(), stc);
    return files;
  }


  public int getDocumentCount() {
    return header.getDocCount();
  }

  public int getTermCount() {
    return header.getTermCount();
  }


  public File getSTCLocation() {
    return stcLocation;
  }

  private static class AllDocs implements CacheDocumentSet {

    private DocSetID id;
    private CacheElement element;
    private Clusterer<DocumentID> clusterer;
    private DefaultDocumentsModel model;

    AllDocs(DocSetID id, CacheElement element) {
      this.id = id;
      this.element = element;
    }

    public String getLabel() {
      return "All Documents";
    }

    public DocSetID getId() {
      return id;
    }

    /**
     * Gets the document ids of all the docs this set contains.
     *
     * @return the document ids of all the docs this set contains.
     */
    public Iterable<DocumentID> documents() {
      return new DIIter(element.getSTC().documents().iterator());
    }

    /**
     * Gets the Clusterer that should be used to Cluster this document set.
     *
     * @return the Clusterer that should be used to Cluster this document set.
     */
    public Clusterer<DocumentID> getClusterer() {
      if (clusterer == null) {
        //ClustererSV clusterer = new ClustererSV(element.getSVS(), "clusterer");
        //clusterer.addAll();
        //this.clusterer = clusterer;
      }
      return this.clusterer;
    }

    public int getDocumentCount() {
      return element.getDocumentCount();
    }

    public int getTermCount() {
      return element.getTermCount();
    }

    public String getTimeStampAsString() {
      return "REPLACE WITH DATE RANGE !!!!";
    }

    public KeywordGenerator getKeyWordGenerator() {
      return element.getKeywordGenerator();
    }

    public ScoredTermsCollection getSTC() {
      return element.getSTC();
    }

    public DocumentsModel getDocumentsModel() {
      if (model == null) {
        model = new DefaultDocumentsModel(id, element);
        model.setDescription("All Documents as of " + getTimeStampAsString());
      }

      return model;
    }

    /**
     * Gets the CacheElement that is the parent of this CacheDocumentSet.
     *
     * @return the CacheElement that is the parent of this CacheDocumentSet.
     */
    public CacheElement getParent() {
      return element;
    }
  }

  private static class DIIter implements Iterable<DocumentID>, Iterator<DocumentID> {

    private Iterator<MIFSDocument> iter;

    public DIIter(Iterator<MIFSDocument> iter) {
      this.iter = iter;
    }

    public Iterator<DocumentID> iterator() {
      return this;
    }

    public boolean hasNext() {
      return iter.hasNext();
    }

    public DocumentID next() {
      MIFSDocument doc = iter.next();
      return doc.getId();
    }

    public void remove() {
      throw new UnsupportedOperationException();
    }
  }


}
