/**
 * 
 */
package anl.aida.client;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * Stores keyword sets according to a keyword set name (e.g.
 * "BioAgent"). 
 * 
 */
public class Keywords {
  
  private Map<String, Set<Keyword>> keywords = new HashMap<String, Set<Keyword>>();
  
  /**
   * Adds the specified term to the specified keyword set.
   * 
   * @param keywordSet the set to add the keyword to
   * @param term the keyword to add
   * @param label the label of the keyword to add
   */
  public void add(String keywordSet, String term, String label) {
    Set<Keyword> set = keywords.get(keywordSet);
    if (set == null) {
      set = new HashSet<Keyword>();
      keywords.put(keywordSet, set);
    }
    set.add(new DefaultKeyword(term, label));
  }
  
  /**
   * Adds the specified keyword to the specified keyword set.
   * 
   * @param keywordSet the set to add the keyword to
   * @param keyword the keyword to add
   */
  public void add(String keywordSet, String keyword) {
    this.add(keywordSet, keyword, keyword);
  }
  
  /**
   * Gets the names of all the keyword sets
   * in this keywords collection.
   * 
   * @return the names of all the keyword sets in this keywords collection.
   */
  public Collection<String> getKeywordSetNames() {
    return Collections.unmodifiableCollection(keywords.keySet());
  }
  
  /**
   * Gets all the keywords in the specified set.
   * 
   * @param keywordSet the name of the set whose keywords we want to get
   * @return all the keywords in the specified set.
   */
  public Collection<Keyword> getKeywords(String keywordSet) {
    Set<Keyword> set = keywords.get(keywordSet);
    if (set == null) {
      return new HashSet<Keyword>();
    }
    return Collections.unmodifiableCollection(set);
  }
  
  /**
   * Gets all the keywords in every set.
   */
  public Collection<Keyword> getAllKeywords() {
    Set<Keyword> set = new HashSet<Keyword>();
    for (Set<Keyword> vals : keywords.values()) {
      set.addAll(vals);
    }
    return Collections.unmodifiableCollection(set);
  }
  
  /**
   * Gets whether or not this Keywords collection contains
   * the specified string. 
   * 
   * @param item the string to test
   * @return true if this Keyword collection contains the string
   * otherwise false.
   */
  // todo speed up this test
  public boolean contains(String item) {
    for (Set<Keyword> vals : keywords.values()) {
      for (Keyword keyword : vals) {
	if (keyword.getTerm().equals(item)) return true;
      }
    }
    return false;
  }
}
