package anl.aida.client;

import anl.aida.algo.Clusterer;
import anl.aida.client.doc.DocumentsModel;
import anl.aida.client.doc.WrappedDocumentsModel;
import anl.aida.core.DocumentID;
import anl.aida.core.KeywordGenerator;
import anl.aida.core.ScoredTermsCollection;

/**
 * CacheDocumentSet that delegates most of its
 * methods to another wrapped DocumentSet.
 *
 */
public class WrappedDocSet implements CacheDocumentSet {

  private CacheDocumentSet set;
  private DocSetID id;
  private String label;
  private DocumentsModel docsModel;

  public WrappedDocSet(DocSetID id, String label, CacheDocumentSet set) {
    this.id = id;
    this.label = label;
    this.set = set;
  }

  /**
   * Gets a descriptive label for this CacheDocumentSet.
   *
   * @return a descriptive label for this CacheDocumentSet.
   */
  public String getLabel() {
    return label;
  }

  /**
   * Gets the id of this document set.
   *
   * @return the id of this document set.
   */
  public DocSetID getId() {
    return id;
  }

  /**
   * Gets the Clusterer that should be used to Cluster this document set.
   *
   * @return the Clusterer that should be used to Cluster this document set.
   */
  public Clusterer<DocumentID> getClusterer() {
    return set.getClusterer();
  }

  /**
   * Gets the number of documents in this CacheDocumentSet.
   *
   * @return the number of documents in this CacheDocumentSet.
   */
  public int getDocumentCount() {
    return set.getDocumentCount();
  }

  /**
   * Gets the number of terms in this CacheDocumentSet.
   *
   * @return the number of terms in this CacheDocumentSet.
   */
  public int getTermCount() {
    return set.getTermCount();
  }

  /**
   * Gets the timestamp of this CacheDocumentSet.
   *
   * @return the timestamp of this CacheDocumentSet.
   */
  public String getTimeStampAsString() {
    return set.getTimeStampAsString();
  }

  /**
   * Gets a keyword generator that can be used to generate
   * keywords for this doc set.
   *
   * @return a keyword generator that can be used to generate
   *         keywords for this doc set.
   */
  public KeywordGenerator getKeyWordGenerator() {
    return set.getKeyWordGenerator();
  }

  /**
   * Gets the ScoredTermsCollection associated with this CacheDocumentSet.
   *
   * @return the ScoredTermsCollection associated with this CacheDocumentSet.
   */
  public ScoredTermsCollection getSTC() {
    return set.getSTC();
  }

  /**
   * Gets the document ids of all the docs this set contains.
   *
   * @return the document ids of all the docs this set contains.
   */
  public Iterable<DocumentID> documents() {
    return set.documents();
  }

  /**
   * Gets a DocumentsModel for the documents in this CacheDocumentSet.
   *
   * @return a DocumentsModel for the documents in this CacheDocumentSet.
   */
  public DocumentsModel getDocumentsModel() {
    if (docsModel == null) {
      docsModel = new WrappedDocumentsModel(id, set.getDocumentsModel());
    }

    return docsModel;
  }

  /**
   * Gets the CacheElement that is the parent of this CacheDocumentSet.
   *
   * @return the CacheElement that is the parent of this CacheDocumentSet.
   */
  public CacheElement getParent() {
    return set.getParent();
  }
}
