package anl.aida.client.doc;

import anl.aida.ae.lucene.LuceneConstants;
import anl.aida.client.CacheElement;
import anl.aida.client.DocSetID;
import anl.aida.client.search.PathFilter;
import anl.aida.core.DocumentID;

import org.apache.lucene.document.DateTools;
import org.apache.lucene.search.*;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * Model class for the DocumentsView that contains some subset
 * of documents.
 *
 */
public class SubsetDocumentsModel extends AbstractDocumentsModel {

  private List<DocumentID> docs = new ArrayList<DocumentID>();

  public SubsetDocumentsModel(DocSetID id, CacheElement element, List<DocumentID> docs) {
    super(id, element);
    this.docs.addAll(docs);
  }


  /**
   * Gets a lucene filter that filters out documents not contained in this
   * subset of documents and docs whose timestamp is greater than this
   * models cache element timestamp.
   *
   * @return a lucene filter that filters out documents not contained in this
   *         subset of documents and docs whose timestamp is greater than this
   *         models cache element timestamp.
   */
  public Filter getSearchFilter() {
    String end = DateTools.timeToString(range.getEnd().getTime(), DateTools.Resolution.DAY);
    String start = DateTools.timeToString(range.getStart().getTime(), DateTools.Resolution.DAY);
    RangeFilter dateFilter = new RangeFilter(LuceneConstants.TIMESTAMP_FIELD, start, end, true, true);
    BooleanFilter filter = new BooleanFilter();
    filter.add(new FilterClause(dateFilter, BooleanClause.Occur.MUST));
    filter.add(new FilterClause(new PathFilter(docs), BooleanClause.Occur.MUST));
    return filter;
  }

  /**
   * Creates a DocSets with each set of the specified size.
   *
   * @param setSize the set size.
   * @return the created DocSets.
   */
  public DocSets createDocumentSets(int setSize) {
    return new DocSetsImpl(setSize);
  }

  class DocSetsImpl implements DocSets {

    private int start, setSize, numSets;

    DocSetsImpl(int setSize) {
      this.setSize = setSize;
      if (setSize >= docs.size()) numSets = 1;
      else numSets = (int) Math.ceil(((float) docs.size()) / setSize);
    }


    public int getSetCount() {
      return numSets;
    }

    public Iterable<ClientDocument> getDocuments(int index) {
      if (index < 0 || index >= numSets) throw new ArrayIndexOutOfBoundsException();
      start = setSize * index;
      int end = start + setSize;
      if (end > docs.size()) end = docs.size();
      return new DocIterator(start, end);
    }
  }

  class DocIterator implements Iterator<ClientDocument>, Iterable<ClientDocument> {

    private int index, end;

    DocIterator(int start, int end) {
      this.index = start;
      this.end = end;
    }

    @Override
    public boolean hasNext() {
      if (index >= end) {
	element.release();
      }
      return index < end;
    }

    @Override
    public ClientDocument next() {
      DocumentID id = docs.get(index);
      String title = element.getSTC().getDocument(id.getSTCIndex()).getTitle();
      ClientDocument cd = new ClientDocument(id, title, getKeywords(id.getPath()));
      index++;
      return cd;
    }

    @Override
    public void remove() {
      throw new UnsupportedOperationException();
    }

    @Override
    public Iterator<ClientDocument> iterator() {
      return this;
    }
  }
}