package anl.aida.client.gui;

import java.awt.event.ActionEvent;
import java.util.ArrayList;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.JButton;

import anl.aida.client.doc.DocSetOperation;
import anl.aida.client.util.ClientUtils;
import anl.aida.core.DocumentID;

/**
 * A widget for graphically representing operations
 * on a selected set of MIFS documents. Each operation
 * is represented as button. This class will manage the 
 * buttons as well as tracking the set of selected documents.
 * The added DocSetOperations will operate on the set
 * of selected documents.
 * 
 */
public class DocOpsWidget {

  // inner action that trigger a doc set op on 
  // the set of documents.
  @SuppressWarnings("serial")
  private class OpAction extends AbstractAction {
    
    DocSetOperation op;

    public OpAction(DocSetOperation op) {
      this.op = op;
    }

    public void actionPerformed(ActionEvent e) {
      op.run(docs);
    }
  }
  
  private List<JButton> buttons = new ArrayList<JButton>();
  private List<DocumentID> docs = new ArrayList<DocumentID>();

  /**
   * Adds a DocSetOperation.
   * 
   * @param op the operation to add
   * @param iconFile the icon to represent the operation with
   * @return the created button that will trigger the operation
   */
  public JButton addOp(DocSetOperation op, String iconFile) {
    JButton button = new JButton(ClientUtils.loadIcon(iconFile));
    button.addActionListener(new OpAction(op));
    buttons.add(button);
    button.setEnabled(docs.size() > 0);
    button.setBorderPainted(false);
    return button;
  }

  /**
   * Clears the set of the selected documents.
   */
  public void clearDocuments() {
    docs.clear();
    buttonsEnabled(false);
  }
  
  private void buttonsEnabled(boolean enabled) {
    for (JButton button : buttons) {
      button.setEnabled(enabled);
    }
  }
  
  /**
   * Gets an iterable over the set of buttons
   * managed by this DocOpsWidget.
   * 
   * @return an iterable over the set of buttons
   * managed by this DocOpsWidget.
   */
  public Iterable<JButton> getButtons() {
    return buttons;
  }

  /**
   * Adds the specified document to the set of selected
   * documents. Any added DocSetOperations will operate on
   * this document.
   * 
   * @param doc the document to add
   */
  public void addDocument(DocumentID doc) {
    this.docs.add(doc);
    buttonsEnabled(true);
  }
}
