package anl.aida.client.io;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.Attribute;
import javax.xml.stream.events.EndElement;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

import anl.aida.client.Cache;
import anl.aida.client.CacheElement;
import anl.aida.client.Keywords;
import anl.aida.core.STCHeader;
import anl.aida.core.STCReader;
import anl.aida.util.DateRangeProducer;

/**
 * Date: Feb 5, 2009 4:04:01 PM
 */
public class CacheFileReader {

  private int cacheId = 0, rangeLength, rangeInterval;
  private long start;
  private List<CacheElement> elements = new ArrayList<CacheElement>();
  private File index, file;
  private XMLEventProcessor processor = new Switcher();
  private XMLEventProcessor defaultProc = processor;
  private Keywords keywords = new Keywords();

  public Cache read(File file) throws IOException, XMLStreamException {
    this.file = file;
    XMLInputFactory factory = XMLInputFactory.newInstance();
    XMLEventReader reader = factory.createXMLEventReader(new BufferedReader(new FileReader(file)));

    while (reader.hasNext()) {
      XMLEvent evt = reader.nextEvent();
      processor.process(reader, evt);
    }

    Cache cache = new Cache(index, file.getParentFile(), new DateRangeProducer(new Date(start), rangeLength,
	rangeInterval));
    for (CacheElement element : elements) {
      element.setKeyword(keywords);
      cache.addCacheElement(element);
    }

    cache.setKeywords(keywords);

    return cache;
  }

  private STCHeader createHeader(File file) throws IOException {
    STCReader reader = new STCReader();
    return reader.readHeader(file.getCanonicalPath());
  }

  private interface XMLEventProcessor {
    void process(XMLEventReader reader, XMLEvent evt) throws XMLStreamException, IOException;
  }

  private class Switcher implements XMLEventProcessor {

    static final String CACHE = "cache";

    @Override
    public void process(XMLEventReader reader, XMLEvent evt) throws XMLStreamException, IOException {
      if (evt.isStartElement()) {
	StartElement elmt = evt.asStartElement();
	String qName = elmt.getName().getLocalPart();
	if (qName.equals(CACHE)) {
	  Attribute attrib = elmt.getAttributeByName(new QName("lucene_index"));
	  index = new File(file.getParent(), attrib.getValue());
	} else if (qName.equals(RangeProcessor.RANGE)) {
	  processor = new RangeProcessor();
	  processor.process(reader, evt);
	} else if (qName.equals(ItemsProcessor.CACHE_ITEMS)) {
	  processor = new ItemsProcessor();
	  processor.process(reader, evt);
	} else if (qName.equals(KeywordProcessor.KEYWORDS)) {
	  processor = new KeywordProcessor();
	  processor.process(reader, evt);
	}
      }
    }
  }

  private class KeywordProcessor implements XMLEventProcessor {

    static final String KEYWORDS = "keywords";
    private static final String KEYWORD_SET = "keyword_set";
    private static final String KEYWORD = "keyword";
    private static final String TERM = "term";
    private static final String LABEL = "label";

    private String setName, term, label;

    @Override
    public void process(XMLEventReader reader, XMLEvent evt) throws XMLStreamException, IOException {
      if (evt.isStartElement()) {
	StartElement elmt = evt.asStartElement();
	String qName = elmt.getName().getLocalPart();
	if (qName.equals(KEYWORD_SET)) {
	  Attribute attrib = elmt.getAttributeByName(new QName("name"));
	  setName = attrib.getValue();
	} else if (qName.equals(KEYWORD)) {
	  //System.out.println("reader.getElementText() = " + reader.getElementText());
	  //keywords.add(setName, reader.getElementText());
	  term = label = "";
	} else if (qName.equals(TERM)) {
	  term = reader.getElementText();
	  
	} else if (qName.equals(LABEL)) {
	  label = reader.getElementText();
	}
      } else if (evt.isEndElement()) {
	EndElement elmt = evt.asEndElement();
	String qName = elmt.getName().getLocalPart();
	if (qName.equals(KEYWORDS)) {
	  processor = defaultProc;
	} else if (qName.equals(KEYWORD)) {
	  keywords.add(setName, term, label);
	}
      }
    }
  }

  private class ItemsProcessor implements XMLEventProcessor {

    static final String CACHE_ITEMS = "cache_items";
    private static final String CACHE_ITEM_NAME = "cache_item";

    @Override
    public void process(XMLEventReader reader, XMLEvent evt) throws XMLStreamException, IOException {
      if (evt.isStartElement()) {
	StartElement elmt = evt.asStartElement();
	String qName = elmt.getName().getLocalPart();
	if (qName.equals(CACHE_ITEM_NAME)) {
	  String stcName = reader.getElementText();
	  File stcFile = new File(file.getParentFile(), stcName);
	  STCHeader header = createHeader(stcFile);
	  elements.add(new CacheElement(cacheId, stcFile, header));
	  cacheId++;
	}
      } else if (evt.isEndElement()
	  && evt.asEndElement().getName().getLocalPart().equals(CACHE_ITEMS)) {
	processor = defaultProc;
      }
    }
  }

  private class RangeProcessor implements XMLEventProcessor {

    static final String RANGE = "range";
    private static final String START = "start";
    private static final String LENGTH = "length";
    private static final String INTERVAL = "interval";

    @Override
    public void process(XMLEventReader reader, XMLEvent evt) throws XMLStreamException, IOException {
      if (evt.isStartElement()) {
	StartElement elmt = evt.asStartElement();
	String qName = elmt.getName().getLocalPart();
	if (qName.equals(START)) {
	  start = Long.parseLong(reader.getElementText());
	} else if (qName.equals(LENGTH)) {
	  rangeLength = Integer.parseInt(reader.getElementText());
	} else if (qName.equals(INTERVAL)) {
	  rangeInterval = Integer.parseInt(reader.getElementText());
	}
      } else if (evt.isEndElement() && evt.asEndElement().getName().getLocalPart().equals(RANGE)) {
	processor = defaultProc;
      }
    }
  }

}
