/**
 * 
 */
package anl.aida.client.io;

import java.awt.Component;
import java.awt.image.BufferedImage;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.IOException;
import java.util.HashSet;
import java.util.Set;

import javax.imageio.ImageIO;
import javax.swing.JFileChooser;
import javax.swing.filechooser.FileFilter;

/**
 * Exports buffered images to a file.
 * 
 */
public class ImageExporter {

  public final static String JPEG = "jpeg";
  public final static String JPG = "jpg";
  public final static String TIFF = "tiff";
  public final static String TIF = "tif";
  public final static String PNG = "png";
  public final static String BMP = "bmp";

  private String currentExt = "png";

  private class ImageFileFilter extends FileFilter {

    private String description;
    private Set<String> imageExt = new HashSet<String>();
    private String ext;

    public ImageFileFilter(String description, String... exts) {
      this.description = description;
      ext = exts[0];
      for (String str : exts) {
	imageExt.add(str);
      }
    }

    public String getDescription() {
      return description;
    }

    public boolean accept(File f) {
      if (f.isDirectory())
	return true;
      String ext = findExtension(f);
      return ext != null && imageExt.contains(ext);
    }

    public String getExtension() {
      return ext;
    }
  }

  private Imageable imageable;

  public ImageExporter(Imageable imageable) {
    this.imageable = imageable;
  }

  /**
   * Run the exporter. This will show a file chooser with the current export
   * formats as a file filters and save a snapshot of the plot to those files.
   * 
   * @throws IOException
   *           if there is an error while creating the image or saving the plot.
   */
  public void run(Component parent) throws IOException {
    JFileChooser chooser = new JFileChooser();
    chooser.addChoosableFileFilter(new ImageFileFilter("BMP Image (*.bmp)", BMP));
    chooser.addChoosableFileFilter(new ImageFileFilter("JPEG Image (*.jpg, *.jpeg)", JPG, JPEG));
    chooser.addChoosableFileFilter(new ImageFileFilter("TIFF Image (*.tif, *.tiff)", TIF, TIFF));
    chooser.addChoosableFileFilter(new ImageFileFilter("PNG Image (*.png)", PNG));
    chooser.setAcceptAllFileFilterUsed(false);
    chooser.setFileFilter(chooser.getChoosableFileFilters()[3]);

    chooser.addPropertyChangeListener(new PropertyChangeListener() {
      public void propertyChange(PropertyChangeEvent evt) {
	if (evt.getPropertyName().equals(JFileChooser.FILE_FILTER_CHANGED_PROPERTY)) {
	  currentExt = ((ImageFileFilter) evt.getNewValue()).getExtension();
	}
      }
    });

    int res = chooser.showSaveDialog(parent);
    if (res == JFileChooser.APPROVE_OPTION) {
      File file = chooser.getSelectedFile();
      save(file);
    }
    currentExt = "png";
  }

  /**
   * Saves an image of the plot in the specified format to the specified file.
   * 
   * @param format
   *          the format of the image file. One of PlotExporter.JPG,
   *          PlotExporter.TIF, PlotExporter.PNG, or PlotExporter.BMP
   * 
   * @param file
   *          the file to save the image to
   * @throws IOException
   *           if there is an error while saving the image
   */
  public void save(String format, File file) throws IOException {
    BufferedImage image = imageable.createImage();
    ImageIO.write(image, format, file);
  }

  private void save(File file) throws IOException {
    String ext = findExtension(file);
    if (ext == null || !ext.equals(currentExt)) {
      file = new File(file.getAbsolutePath() + "." + currentExt);
    }

    BufferedImage image = imageable.createImage();
    ImageIO.write(image, currentExt, file);
  }

  private String findExtension(File f) {
    String ext = null;
    String s = f.getName();
    int i = s.lastIndexOf('.');

    if (i > 0 && i < s.length() - 1) {
      ext = s.substring(i + 1).toLowerCase();
    }
    return ext;
  }
}
