package anl.aida.client.net;

import java.util.ArrayList;
import java.util.List;

import anl.aida.client.Keyword;
import anl.aida.core.DocumentID;
import anl.aida.score.TermScoreFilter;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.SparseGraph;

/**
 * Builder for a concept graph.
 *
 */
public class ConceptGraphBuilder {

  private List<DocumentID> docs;
  private Graph<TermNode, CGEdge> graph = new SparseGraph<TermNode, CGEdge>();
  private List<TermNode> nodes = new ArrayList<TermNode>();

  public ConceptGraphBuilder(Iterable<DocumentID> docIter) {
    this.docs = new ArrayList<DocumentID>();
    for (DocumentID doc : docIter) {
      this.docs.add(doc);
    }
  }

  /**
   * Builds the nodes for the ConceptGraph using the specified filter.
   * Each concept becomes a node in the graph. Each document that passes the
   * filter for a concept will be added to that concept.
   *
   * @param filter   the filter to use when adding documents to nodes
   * @param keywords the list of the concepts that should be become nodes.
   * @return this ConceptGraphBuilder.
   */
  public ConceptGraphBuilder buildNodes(TermScoreFilter filter, List<? extends Keyword> keywords) {
    for (Keyword keyword : keywords) {
      TermNode node = new TermNode(keyword.getLabel());
      for (DocumentID doc : docs) {
        if (filter.evaluate(keyword.getTerm(), doc)) {
          node.addDocument(doc);
        }
      }

      graph.addVertex(node);
      nodes.add(node);
    }

    return this;
  }

  /**
   * Builds the edges using the specified edge creator. Every pair of nodes
   * will be passed into the edge creator. The edges are assumed to be
   * undirected and so the pair X,Y is identical to Y,X as far as edge
   * creation is concerned.
   *
   * @param edgeCreator the created used to create the graphs edges
   * @return this ConceptGraphBuilder
   */
  public ConceptGraphBuilder buildEdges(EdgeCreator edgeCreator) {
    int size = nodes.size();
    for (int i = 0; i < size; i++) {
      TermNode iNode = nodes.get(i);
      for (int j = i + 1; j < size; j++) {
        TermNode jNode = nodes.get(j);
        edgeCreator.createEdge(iNode, jNode, graph);
      }
    }
    return this;
  }

  /**
   * Gets the graph that is the result of the building process.
   *
   * @return the graph that is the result of the building process.
   */
  public Graph<TermNode, CGEdge> getResult() {
    return graph;
  }
}
