package anl.aida.client.net;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import edu.uci.ics.jung.graph.util.Pair;

import simphony.util.messages.MessageCenter;

import anl.aida.client.Cache;
import anl.aida.client.CacheDocumentSet;
import anl.aida.client.CacheElement;
import anl.aida.client.gui.VizPropertyBean;
import anl.aida.score.TermScoreFilter;
import anl.aida.util.DateRange;

/**
 * Class for creating text output of data contained in graphs displayed.
 */
public class NetworkOutputter extends TSConceptGraphs{
	
	/**List of DataRange objects**/
	private List<DateRange> ranges = new ArrayList<DateRange>();
	
	/**Delimiter for file outputs**/
	private static String DELIMITER= ",";
	
	/**Directory for placing file outputs**/
	private static String directory = "../Client/text_cache_output/";

	/**The document counts arranged by DateRange**/
	private Map<DateRange, Integer> documents=new HashMap<DateRange,Integer>();
	
	/**
	 * Constructor for creating network data output files.
	 * @param cache the cache of data
	 * @param bean the visualization bean
	 * @param graphs the graphs containing term nodes and links
	 */
	 public NetworkOutputter(Cache cache, VizPropertyBean bean,TSConceptGraphs graphs){
		super(cache,bean);
		 for (DateRange range : graphs.getRanger().ranges()) {
		      ranges.add(range);
		 }
	  }
	 
	 /**
	  * Method to control the graph population process as well as file outputs.
	  */
	  public void writeOutputOut(){
		  if(!new File(directory).exists())
			  new File(directory).mkdir();
		  deleteDirContent(new File(directory));
		  determineGraphRanges();
		  readGraphs();
	  }
	  
	  /**
	   * Method to place relevant graphs for the time range investigated.
	   * @param range the DateRange object.
	   */
	  private void populateGraph(DateRange range) {
		      ConceptGraph<TermNode> graph = null;
		      CacheElement element = cache.getElement(range);

		      if (element == null) {
		       
		        return;
		      }

		      CacheDocumentSet docSet = element.getDocumentSet(0);
		      int documentCounts=docSet.getSTC().getDocumentCount();
		      TermScoreFilter filter = bean.getNodeFilter().createFilter(docSet.getParent(), (float) bean.getNodeThresh());
		      ConceptGraphBuilder builder = new ConceptGraphBuilder(docSet.documents());

		      builder.buildNodes(filter, bean.getKeywords());
		      EdgeCreator creator = bean.getEdgeFilter().createEdgeCreator(docSet.getParent(), bean.getEdgeDocThresh(),
		              (float) bean.getEdgeScoreThresh());
		      builder.buildEdges(creator);
		      graph = new ConceptGraph<TermNode>(bean.getTitle(), docSet.getId(), builder.getResult());
		      // release the element to save the memory taken up by the
		      // stc etc.
		      element.release();
		      graphs.put(range, graph);
		      documents.put(range,documentCounts);
	
		    if (docOp != null) docOp.setElement(element);

		}
	  
	public static void setDirectory(String directory) {
		NetworkOutputter.directory = directory;
	}

	/**
	 * Method to determine the graph time range
	 */
	private void determineGraphRanges(){
	    for(int i=0; i < ranges.size(); i++) {
	    	DateRange range = ranges.get(i);
	    	populateGraph(range);
	    }
	  }
	  
	/**
	 * Method to read graph data for output.
	 */
	  private void readGraphs(){
		  for(int i=0; i< ranges.size(); i++){
			  DateRange range = ranges.get(i);
			  ConceptGraph<?> graph = graphs.get(range);
			  Iterator<CGEdge> ic=graph.getEdges().iterator();
			  Map<String,Integer> linkCounts = new HashMap<String,Integer>();
			  Map<String,Integer> nodeCounts = new HashMap<String,Integer>();
			  while(ic.hasNext()){
				  CGEdge e = ic.next();
				  Pair<?> p=graph.getEndpoints(e);
				  TermNode node1 = (TermNode) p.getFirst();
				  TermNode node2 = (TermNode) p.getSecond();
				  nodeCounts.put(node1.getLabel(), node1.getDocumentCount());
				  nodeCounts.put(node2.getLabel(), node2.getDocumentCount());
				  int count=e.getDocumentCount();
				  String label=node1.getLabel()+"-"+node2.getLabel();
				  linkCounts.put(label,count);
			  }
			  int docCounts=this.documents.get(range);
			  writeOutput(range.getStart(),range.getEnd(),linkCounts,"Links");
			  writeOutput(range.getStart(),range.getEnd(),nodeCounts,"Nodes");
			  writeSummary(range.getStart(),range.getEnd(),docCounts);
		  }
	  }

	  public static void setDELIMITER(String delimiter) {
		DELIMITER = delimiter;
	  }

	  /**
	   * Method deletes directory content prior to filling it with new output files.
	   * @param dir the directory
	   * @return a boolean for file deletion
	   */
	  private static boolean deleteDirContent(File dir) {
	        if (dir.isDirectory()) {
	            String[] children = dir.list();
	            for (int i=0; i<children.length; i++) {
	                boolean success = deleteDirContent(new File(dir, children[i]));
	                if (!success) {
	                    return false;
	                }
	            }
	        }
	    
	       return true;
	    } 
	  
	  /**
	   * Output provided for the summary of the total number of documents per time range
	   * @param start the start date
	   * @param end the end date
	   * @param count the total count of documents for the time range
	   */
	  private void writeSummary(Date start, Date end, int count){
		  Calendar c = new GregorianCalendar();
			c.setTime(start);
		    int yearS= c.get(Calendar.YEAR);
		    int monthS = c.get(Calendar.MONTH);
		    int dayS = c.get(Calendar.DAY_OF_MONTH);
		    
		    Calendar c2 = new GregorianCalendar();
			c2.setTime(end);
		    int yearE= c2.get(Calendar.YEAR);
		    int monthE = c2.get(Calendar.MONTH);
		    int dayE = c2.get(Calendar.DAY_OF_MONTH);
		    
		    String st=Integer.toString(monthS+1)+"-"
		    	+Integer.toString(dayS)+"-"+Integer.toString(yearS);
		    
		    String et=Integer.toString(monthE+1)+"-"
	    		+Integer.toString(dayE)+"-"+Integer.toString(yearE);
		    
		    String filename = st+"_"+et+".txt";
		    
		    FileWriter fw= null;
			PrintWriter pw= null;
			
		    File file=null;
			try
			{
				file = new File(directory+"Summary"+"_"+filename);
				fw= new FileWriter( file );
				pw=  new PrintWriter( fw );
				
				pw.print("Total Document");
				pw.println();
				pw.print(count);
				
				pw.close();
			}
			catch(IOException e){
				MessageCenter.getMessageCenter(NetworkOutputter.class).error("Problems with creating" +
						"output summary file", e);
			}	
	  }
	  
	  /**
	   * Method writes data out to a specific file.
	   * @param start the start data of a DateRange
	   * @param end the end date of a DateRange
	   * @param rangeValues values based on combined link label and their document count values
	   * @param type the type of file output (nodes or links)
	   */
	  private void writeOutput(Date start, Date end,Map<String,Integer> rangeValues,String type){
			Calendar c = new GregorianCalendar();
			c.setTime(start);
		    int yearS= c.get(Calendar.YEAR);
		    int monthS = c.get(Calendar.MONTH);
		    int dayS = c.get(Calendar.DAY_OF_MONTH);
		    
		    Calendar c2 = new GregorianCalendar();
			c2.setTime(end);
		    int yearE= c2.get(Calendar.YEAR);
		    int monthE = c2.get(Calendar.MONTH);
		    int dayE = c2.get(Calendar.DAY_OF_MONTH);
		    
		    String st=Integer.toString(monthS+1)+"-"
		    	+Integer.toString(dayS)+"-"+Integer.toString(yearS);
		    
		    String et=Integer.toString(monthE+1)+"-"
	    		+Integer.toString(dayE)+"-"+Integer.toString(yearE);
		    
		    String filename = st+"_"+et+".txt";
		    
		    FileWriter fw= null;
			PrintWriter pw= null;
			
		    File file=null;
			try
			{
				file = new File(directory+type+"_"+filename);
				fw= new FileWriter( file );
				pw=  new PrintWriter( fw );
				
				Iterator<String> ki = rangeValues.keySet().iterator();
				List<Integer>values = new ArrayList<Integer>();
				int size=rangeValues.keySet().size();
				int t=1;
				while(ki.hasNext()){
					String key = ki.next();
					pw.print(key);
					if(t<size)
						pw.print(DELIMITER);
					values.add(rangeValues.get(key));
					t++;
				}
				pw.println();
				Iterator<Integer>ic = values.iterator();
				t=1;
				while(ic.hasNext()){
					Integer i = ic.next();
					pw.print(i);
					if(t<size)
						pw.print(DELIMITER);
					t++;
				}
				pw.close();
			}
			catch(IOException e){
				MessageCenter.getMessageCenter(NetworkOutputter.class).error("Problems with creating" +
						"output file", e);
			}	
	    }
  }