package anl.aida.client.net;

import anl.aida.core.DocumentID;
import anl.aida.score.ScorerThresholdFilter;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.util.EdgeType;

import java.util.HashSet;
import java.util.Set;

/**
 * Creates an edge between two term nodes based on the
 * score of the documents in the two nodes. Specifically,
 * <pre>
 * 1. For each document, if the document passes the filter for both terms, then
 * increase the document count.
 * 2. If after processing all the documents, the doc count
 * is greater than some threshold then create the edge.
 * </pre>
 *
 */
public class ScorerThresholdEdgeCreator implements EdgeCreator {

  private int dcThreshold;
  private ScorerThresholdFilter filter;


  /**
   * Creates an OccurenceEdgeCreator with the specified thresholds.
   *
   * @param docThreshold the document count threshold
   * @param filter       the filter that determines whether the doc count should be
   *                     increased
   */
  public ScorerThresholdEdgeCreator(int docThreshold, ScorerThresholdFilter filter) {
    this.dcThreshold = docThreshold;
    this.filter = filter;

  }

  /**
   * Creates an edge between the two nodes if the
   * number of documents that occur in both nodes is
   * greater than or equal to the threshold.
   *
   * @param node1 the first node
   * @param node2 the second node
   * @param graph the graph where the edge will be created
   * @return boolean if the edge created, otherwise false.
   */
  public boolean createEdge(TermNode node1, TermNode node2, Graph<TermNode, CGEdge> graph) {

    Set<DocumentID> set1 = node1.documents();
    if (set1.size() < 1) return false;
    Set<DocumentID> set2 = node2.documents();
    if (set2.size() < 1) return false;
    
    String term1 = node1.getLabel();
    String term2 = node2.getLabel();
    
    

    Set<DocumentID> docs = new HashSet<DocumentID>();

    //System.out.println("... Evaluating ...");
    for (DocumentID id : set1) {
      if (filter.evaluate(term1, id) && filter.evaluate(term2, id)) docs.add(id);
    }

    for (DocumentID id : set2) {
      if (!docs.contains(id) && filter.evaluate(term1, id) && filter.evaluate(term2, id)) docs.add(id);
    }
    //System.out.println("... Done Evaluating ...");

    if (docs.size() >= dcThreshold) {
      graph.addEdge(new DefaultCGEdge(node1.getLabel() + " <->" + node2.getLabel(),
              docs), node1, node2, EdgeType.UNDIRECTED);
      return true;
    }

    return false;
  }
}