package anl.aida.client.net;

import anl.aida.client.Cache;
import anl.aida.client.CacheDocumentSet;
import anl.aida.client.CacheElement;
import anl.aida.client.ShowDocViewOp;
import anl.aida.client.doc.DocSetOperation;
import anl.aida.client.gui.AIDAGui;
import anl.aida.client.gui.VizPropertyBean;
import anl.aida.score.TermScoreFilter;
import anl.aida.util.DateRange;
import anl.aida.util.DateRangeProducer;
import simphony.util.messages.MessageCenter;

import java.util.HashMap;
import java.util.Map;

/**
 * Produces ConceptGraphs representing different timeslices. The
 * graph's term node concepts remain the same, but the  doc's
 * associated and edges created will change from graph to graph.
 *
 */
public class TSConceptGraphs {

	 protected static final MessageCenter msg = MessageCenter.getMessageCenter(TSConceptGraphs.class);

  protected Map<DateRange, ConceptGraph<TermNode>> graphs = new HashMap<DateRange, ConceptGraph<TermNode>>();
  protected VizPropertyBean bean;
  protected Cache cache;
  protected ShowDocViewOp docOp;
  // the current graph.
  private ConceptGraph<TermNode> graph;

  public TSConceptGraphs(Cache cache, VizPropertyBean bean) {
    this.bean = bean;
    this.cache = cache;
  }
  
  /**
   * Gets a DocSet operation for a Canvas that displays
   * docs produced by this TSConceptGraphs. 
   * 
   * @param gui the gui mediator
   * @return a DocSet operation for a Canvas that displays
   * graphs produced by this TSConceptGraphs. 
   */
  public DocSetOperation getDocSetOp(AIDAGui gui) {
    docOp = new ShowDocViewOp(gui, null);
    if (graph != null) {
      CacheElement element = cache.getElement(graph.getDocSetID().getCacheElementId());
      docOp.setElement(element);
    }
    
    return docOp;
  }

  /**
   * Gets a DateRangeProducer for the entire set of ranges covered
   * by these graphs.
   *
   * @return a DateRangeProducer for the entire set of ranges covered
   *         by these graphs.
   */
  public DateRangeProducer getRanger() {
    return cache.getRanger();
  }

  /**
   * Gets the graph for the specified DateRange.
   *
   * @param range the date range
   * @return the graph for the specified DateRange.
   */
  public ConceptGraph<TermNode> getGraph(DateRange range) {
    graph = graphs.get(range);
    if (graph == null) {
      CacheElement element = cache.getElement(range);

      if (element == null) {
        msg.info("TO DO: fulfill request for synthetic range");
        return null;
      }

      CacheDocumentSet docSet = element.getDocumentSet(0);
      TermScoreFilter filter = bean.getNodeFilter().createFilter(docSet.getParent(), (float) bean.getNodeThresh());
      ConceptGraphBuilder builder = new ConceptGraphBuilder(docSet.documents());

      builder.buildNodes(filter, bean.getKeywords());
      EdgeCreator creator = bean.getEdgeFilter().createEdgeCreator(docSet.getParent(), bean.getEdgeDocThresh(),
              (float) bean.getEdgeScoreThresh());
      builder.buildEdges(creator);
      graph = new ConceptGraph<TermNode>(bean.getTitle(), docSet.getId(), builder.getResult());
      // release the element to save the memory taken up by the
      // stc etc.
      element.release();
      graphs.put(range, graph);
    } 
    
    CacheElement element = cache.getElement(range);
    if (docOp != null) docOp.setElement(element);

    return graph;
  }
}
