package anl.aida.client.search;

import anl.aida.ae.lucene.LuceneConstants;
import anl.aida.core.DocumentID;
import anl.aida.core.MIFSDocument;
import anl.aida.core.ScoredTermsCollection;

import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.queryParser.MultiFieldQueryParser;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.queryParser.QueryParser;
import org.apache.lucene.search.*;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

/**
 * Searches a Lucene index and returning results as DocumentIDs.
 *
 */
public class LuceneSearcher {

  private ScoredTermsCollection stc;
  private IndexSearcher searcher;

  public LuceneSearcher(String index, ScoredTermsCollection stc) throws IOException {
    this.stc = stc;
    this.searcher = new IndexSearcher(index);
  }

  /**
   * Searches the index using the specified query, returning no more than
   * numDocs results. The results are limited to those docs specified in ids.
   *
   * @param ids         any result must be part of this list of document ids
   * @param queryString the query
   * @param numDocs     the maximum number of docs to return
   * @return an index of document ids for the docs returned by the search
   * @throws ParseException if there is an error in the query
   * @throws IOException    if there is an error searching the index
   */
  public List<DocumentID> search(List<DocumentID> ids, String queryString, int numDocs) throws ParseException,
          IOException {
    return search(queryString, numDocs, new PathFilter(ids));
  }


  /**
   * Searches the index using the specified query, returning no more than
   * numDocs results.
   *
   * @param queryString the query
   * @param numDocs     the maximum number of docs to return
   * @return an index of document ids for the docs returned by the search
   * @throws ParseException if there is an error in the query
   * @throws IOException    if there is an error searching the index
   */
  public List<DocumentID> search(String queryString, int numDocs) throws ParseException, IOException {
    return search(queryString, numDocs, null);

  }

  /**
   * Searches the index using the specified query, returning no more than
   * numDocs results.
   *
   * @param queryString the query
   * @param numDocs     the maximum number of docs to return
   * @param filter      the filter to filter query results. This can be null.
   * @return an index of document ids for the docs returned by the search
   * @throws ParseException if there is an error in the query
   * @throws IOException    if there is an error searching the index
   */
  public List<DocumentID> search(String queryString, int numDocs, Filter filter) throws ParseException, IOException {
    QueryParser parser = new MultiFieldQueryParser(new String[]{LuceneConstants.CONTENTS_FIELD,
            LuceneConstants.TITLE_FIELD}, new StandardAnalyzer());
    Query query = parser.parse(queryString);

    TopDocCollector collector = new TopDocCollector(numDocs);
    if (filter == null) searcher.search(query, collector);
    else searcher.search(query, filter, collector);

    ScoreDoc[] hits = collector.topDocs().scoreDocs;
    List<DocumentID> ids = new ArrayList<DocumentID>();
    for (ScoreDoc hit : hits) {
      int id = hit.doc;
      Document doc = searcher.doc(id);
      String path = doc.get(LuceneConstants.PATH_FIELD);
      ids.add(findDocument(path));
    }

    return ids;
  }

  /**
   * Closes the searcher. It cannot be used after this has been called
   *
   * @throws IOException if there is an error closing the searcher.
   */
  public void close() throws IOException {
    searcher.close();
  }

  // this searches the stc item by item
  // this may be too slow when we are dealing with
  // large doc collections.
  private DocumentID findDocument(String path) {
    for (int i = 0, n = stc.getDocumentCount(); i < n; i++) {
      MIFSDocument document = stc.getDocument(i);
      if (path.equals(document.getPath())) {
        return new DocumentID(i, path);
      }
    }
    return null;
  }
}
