package anl.aida.gis;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.geotools.data.DataStore;
import org.geotools.data.DataStoreFinder;
import org.geotools.data.FeatureSource;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureIterator;
import org.geotools.geometry.jts.JTS;
import org.geotools.referencing.CRS;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.geometry.MismatchedDimensionException;
import org.opengis.referencing.FactoryException;
import org.opengis.referencing.NoSuchAuthorityCodeException;
import org.opengis.referencing.crs.CoordinateReferenceSystem;
import org.opengis.referencing.operation.MathTransform;
import org.opengis.referencing.operation.TransformException;

import simphony.util.messages.MessageCenter;

import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.index.ItemVisitor;
import com.vividsolutions.jts.index.SpatialIndex;
import com.vividsolutions.jts.index.quadtree.Quadtree;

/**
 * Stores a feature in a spatial index. The feature geometry is converted to
 * WGS84.
 * 
 */
public class FeatureSpatialMap {

  private static MessageCenter msg = MessageCenter.getMessageCenter(FeatureSpatialMap.class);

  private SpatialIndex index = new Quadtree();
  private CoordinateReferenceSystem crs;
  private CoordinateReferenceSystem wgs84;
  private MathTransform wgsToLocal;

  private void initCRS() {
    try {
      wgs84 = CRS.decode("EPSG:4326");
    } catch (NoSuchAuthorityCodeException e) {
      wgs84 = DefaultGeographicCRS.WGS84;
    } catch (FactoryException e) {
      msg.error("Unable to create WGS84 coordinate reference system", e);
      return;
    }
  }

  public FeatureSpatialMap(File shapeFile) {
    initCRS();
    try {
      Map<String, Serializable> params = new HashMap<String, Serializable>();
      params.put("url", shapeFile.toURI().toURL());
      params.put("create spatial index", true);
      DataStore dataStore = DataStoreFinder.getDataStore(params);

      String typeName = dataStore.getTypeNames()[0];
      FeatureSource<SimpleFeatureType, SimpleFeature> featureSource = dataStore
	  .getFeatureSource(typeName);
      addFeatures(featureSource.getFeatures());
    } catch (IOException ex) {
      msg.error("Error reading shapefile", ex);
    }
  }
  
  public FeatureSpatialMap(FeatureCollection<SimpleFeatureType, SimpleFeature> features) {
    initCRS();
    addFeatures(features);
  }

  private void addFeatures(FeatureCollection<SimpleFeatureType, SimpleFeature> features) {
    FeatureIterator<SimpleFeature> iter = null;
    try {
      crs = features.getSchema().getCoordinateReferenceSystem();
      MathTransform localToWGS84 = CRS.findMathTransform(crs, wgs84, true);
      wgsToLocal = CRS.findMathTransform(wgs84, crs, true);

      for (iter = features.features(); iter.hasNext();) {
	SimpleFeature feature = iter.next();
	Geometry geom = (Geometry) feature.getDefaultGeometry();

	/*
	 * System.out.printf("%s: Untransformed - %s%n",
	 * feature.getAttribute("NAME"), geom.getEnvelope());
	 * System.out.printf("%s: Trans geom - %s%n",
	 * feature.getAttribute("NAME"), JTS.transform(geom.getEnvelope(),
	 * localToWGS84)); Coordinate coord =
	 * geom.getEnvelope().getCoordinates()[0];
	 * System.out.println(JTS.transform(coord, null, localToWGS84));
	 * System.out.println();
	 */

	Envelope env = JTS.toEnvelope(JTS.transform(geom.getEnvelope(), localToWGS84));
	// System.out.printf("%s: %s%n", feature.getAttribute("NAME"), env);
	index.insert(env, feature);
      }
    } catch (FactoryException e) {
      msg.error("Error creating transform", e);
    } catch (MismatchedDimensionException e) {
      msg.error("Error transforming geometry", e);
    } catch (TransformException e) {
      msg.error("Error transforming geometry", e);
    } finally {
      if (iter != null)
	iter.close();
    }
  }

  /**
   * Gets the list of features that contain the specified point. Assumes the
   * point is in WGS84.
   * 
   * @param pt
   *          the point
   * 
   * @return
   */
  public List<SimpleFeature> findFeature(Point pt) {
    ContainsPtVisitor visitor = null;
    try {
      Geometry localPt = JTS.transform(pt, wgsToLocal);
      visitor = new ContainsPtVisitor(localPt);
      Envelope searchEnv = new Envelope(pt.getCoordinate());
      // System.out.printf("Env: %s%n", searchEnv);
      index.query(searchEnv, visitor);
    } catch (MismatchedDimensionException e) {
      msg.error("Error finding feature", e);
    } catch (TransformException e) {
      msg.error("Error finding feature", e);
    }

    if (visitor != null)
      return visitor.items;
    return new ArrayList<SimpleFeature>();
  }

  // visitor that does a within check
  private class ContainsPtVisitor implements ItemVisitor {

    List<SimpleFeature> items = new ArrayList<SimpleFeature>();
    Geometry pt;

    public ContainsPtVisitor(Geometry pt) {
      this.pt = pt;
    }

    public void visitItem(Object o) {
      SimpleFeature feature = (SimpleFeature) o;
      if (pt.intersects((Geometry) feature.getDefaultGeometry())) {
	items.add(feature);
      }
    }
  }
}
