package anl.aida.gis;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionAdapter;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.geom.Point2D;
import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;

import javax.swing.Action;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JToolBar;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.geotools.data.DataStore;
import org.geotools.data.DataStoreFinder;
import org.geotools.data.FeatureSource;
import org.geotools.geometry.jts.ReferencedEnvelope;
import org.geotools.map.DefaultMapContext;
import org.geotools.map.DefaultMapLayer;
import org.geotools.map.MapContext;
import org.geotools.map.MapLayer;
import org.geotools.styling.Style;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import simphony.util.messages.MessageCenter;
import anl.aida.client.util.ClientUtils;
import anl.aida.data.DataFrame;
import anl.aida.plot.LegendPanel;
import anl.aida.plot.PlotCreationException;
import anl.aida.plot.TimeStepWidget;
import edu.umd.cs.piccolo.PCamera;
import edu.umd.cs.piccolo.PCanvas;
import edu.umd.cs.piccolo.event.PInputEventListener;
import edu.umd.cs.piccolo.util.PBounds;

public class GISDisplay {

  private static MessageCenter msg = MessageCenter.getMessageCenter(GISDisplay.class);

  class TimeStepListener implements ChangeListener {

    private TimeStepWidget ts;
    private Styler styler;

    public TimeStepListener(TimeStepWidget ts, Styler styler) {
      this.ts = ts;
      this.styler = styler;
    }

    @Override
    public void stateChanged(ChangeEvent e) {
      int timestep = ts.getTime();
      try {
	Style style = styler.getStyle(timestep);
	mapContext.getLayer(0).setStyle(style);
      } catch (PlotCreationException ex) {
	msg.error("Unable to update map style for new timestep", ex);
      }
    }
  }

  private PiccoloMapPanel mapPanel;
  private JPanel panel;
  private MapContext mapContext;
  private PositionTool positionTool;
  private LegendPanel legendPanel;

  public GISDisplay(CoordinateReferenceSystem crs, LegendPanel legendPanel) {
    this(crs);
    this.legendPanel = legendPanel;
  }

  public GISDisplay(CoordinateReferenceSystem crs) {
   this(new DefaultMapContext(crs));
    
  }
  
  public GISDisplay(MapContext context) {
    mapContext = context;
    mapPanel = new PiccoloMapPanel(mapContext);
    initTools();
  }

  public TimeStepWidget initTimeStepWidget(DataFrame frame, Styler styler) {
    TimeStepWidget ts = new TimeStepWidget();
    ts.init(frame.getAxes(), 0);
    mapPanel.getToolBar().addSeparator();
    mapPanel.getToolBar().add(ts);
    ts.addSpinnerListener(new TimeStepListener(ts, styler));
    return ts;
  }

  public JToolBar getToolbar() {
    return mapPanel.getToolBar();
  }

  public void addLayer(File file, Style style) {
    try {
      Map<String, Serializable> params = new HashMap<String, Serializable>();
      params.put("url", file.toURI().toURL());
      params.put("create spatial index", true);
      DataStore dataStore = DataStoreFinder.getDataStore(params);

      String typeName = dataStore.getTypeNames()[0];
      msg.info("Reading content: " + typeName);
      FeatureSource<SimpleFeatureType, SimpleFeature> featureSource = dataStore
	  .getFeatureSource(typeName);
      MapLayer layer = new DefaultMapLayer(featureSource, style);
      mapContext.addLayer(layer);
      mapContext.setAreaOfInterest(layer.getBounds());
      mapPanel.getCanvas().zoomToAreaOfInterest();
    } catch (IOException ex) {
      msg.error("Error adding gis layer", ex);
    }
  }

  public JPanel getPanel() {
    if (panel == null) {
      panel = new JPanel(new BorderLayout());
      panel.add(mapPanel, BorderLayout.CENTER);
      if (legendPanel != null) {
	panel.add(legendPanel, BorderLayout.EAST);
      }
      panel.add(mapPanel.getToolBar(), BorderLayout.NORTH);
      mapPanel.getCanvas().zoomToAreaOfInterest();
      panel.addComponentListener(new ComponentAdapter() {
	@Override
	public void componentResized(ComponentEvent e) {
	  mapPanel.getCanvas().zoomToAreaOfInterest();
	}
      });
    }

    return panel;
  }

  public void destroy() {
    PGISCanvas canvas = mapPanel.getCanvas();
    for (MouseWheelListener listener : canvas.getMouseWheelListeners()) {
      canvas.removeMouseWheelListener(listener);
    }

    for (MouseMotionListener listener : canvas.getMouseMotionListeners()) {
      canvas.removeMouseMotionListener(listener);
    }
    positionTool.cleanUp();
    mapPanel.removeAll();
    mapContext = null;
  }

  public void addLocationListener(LocationChangedListener listener) {
    positionTool.addLocationChangedListener(listener);
  }

  public void addInputEventListener(PInputEventListener listener) {
    mapPanel.getCanvas().addInputEventListener(listener);
  }

  public void addMouseMotionListener(MouseMotionListener listener) {
    mapPanel.getCanvas().addMouseMotionListener(listener);
  }

  public void addMouseListener(MouseListener listener) {
    mapPanel.getCanvas().addMouseListener(listener);
  }

  public PCamera getCamera() {
    return mapPanel.getCanvas().getCamera();
  }

  public PCanvas getCanvas() {
    return mapPanel.getCanvas();
  }

  private void initTools() {
    Map<String, Object> toolParams = new HashMap<String, Object>();
    toolParams.put(ToolManager.TOGGLE, true);
    toolParams.put(Action.SMALL_ICON, new ImageIcon(PMarqueeZoomIn.class
	.getResource("viewmag+.png")));
    toolParams.put(Action.SHORT_DESCRIPTION, "Zoom In");
    mapPanel.addTool(new PMarqueeZoomIn(mapContext), toolParams);

    // zoomOut
    toolParams = new HashMap<String, Object>();
    toolParams.put(ToolManager.TOGGLE, true);
    toolParams.put(Action.SMALL_ICON, new ImageIcon(PMarqueeZoomOut.class
	.getResource("viewmag-.png")));
    toolParams.put(Action.SHORT_DESCRIPTION, "Zoom Out");
    mapPanel.addTool(new PMarqueeZoomOut(mapContext), toolParams);

    // pan
    PGISCanvas canvas = mapPanel.getCanvas();
    toolParams = new HashMap<String, Object>();
    toolParams.put(Action.SHORT_DESCRIPTION, "Pan the map");
    toolParams.put(ToolManager.TOGGLE, true);
    toolParams.put(Action.SMALL_ICON, ClientUtils.loadIcon("/move.png"));
    toolParams.put("DEFAULT", Boolean.TRUE);
    toolParams.put(ToolManager.SELECTED, Boolean.TRUE);
    mapPanel.addTool(new PGISPanTool(mapContext, canvas), toolParams);

    JButton button = new JButton(ClientUtils.loadIcon("InitialExtent.png"));
    button.setBorderPainted(false);
    button.setToolTipText("Zoom to Full Map");
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
	mapContext.setAreaOfInterest(mapContext.getLayer(0).getBounds());
	mapPanel.getCanvas().zoomToAreaOfInterest();
      }
    });
    mapPanel.getToolBar().add(button);

    positionTool = new PositionTool(mapContext.getCoordinateReferenceSystem());
    mapPanel.getCanvas().addInputEventListener(positionTool);
    mapPanel.getCanvas().addMouseWheelListener(new GISMouseWheelListener());
    mapPanel.getCanvas().addMouseMotionListener(new GISMouseMotionListener());

  }

  private double mousePositionX;
  private double mousePositionY;

  public class GISMouseWheelListener implements MouseWheelListener {
    private double minScale = 0.00000001;
    private double maxScale = Double.MAX_VALUE;

    public void mouseWheelMoved(MouseWheelEvent e) {
      PGISCanvas canvas = mapPanel.getCanvas();
      PCamera camera = canvas.getCamera();

      double dx = -e.getWheelRotation();

      double scaleDelta = (1.0 + (0.1 * dx));

      double currentScale = camera.getViewScale();
      double newScale = currentScale * scaleDelta;

      if (newScale < minScale)
	scaleDelta = minScale / currentScale;

      if ((maxScale > 0) && (newScale > maxScale))
	scaleDelta = maxScale / currentScale;

      Point2D point = camera.localToView(new Point2D.Double(mousePositionX, mousePositionY));

      camera.scaleViewAboutPoint(scaleDelta, point.getX(), point.getY());
      PBounds viewBounds = camera.getViewBounds();
      canvas.setAreaOfInterest(new ReferencedEnvelope(viewBounds.getMinX(), viewBounds.getMaxX(),
	  viewBounds.getMinY(), viewBounds.getMaxY(), canvas.getCRS()));
    }
  }

  /**
   * Assists mouse wheel zoom control by updating the center coord
   * 
   * @author tatara
   */
  public class GISMouseMotionListener extends MouseMotionAdapter {

    public void mouseMoved(MouseEvent e) {
      mousePositionX = e.getX();
      mousePositionY = e.getY();
    }
  }

}
