package anl.aida.gis;

import java.awt.Cursor;
import java.util.ArrayList;
import java.util.List;

import org.geotools.geometry.jts.JTS;
import org.geotools.referencing.CRS;
import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.opengis.referencing.FactoryException;
import org.opengis.referencing.NoSuchAuthorityCodeException;
import org.opengis.referencing.crs.CoordinateReferenceSystem;
import org.opengis.referencing.operation.MathTransform;
import org.opengis.referencing.operation.TransformException;

import simphony.util.messages.MessageCenter;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.Point;

import edu.umd.cs.piccolo.event.PBasicInputEventHandler;
import edu.umd.cs.piccolo.event.PInputEvent;

/**
 * GIS Tool for firing position changed events to a listener. This translates a
 * mouse coordinate into a WGS 84 lon and lat location.
 * 
 */
public class PositionTool extends PBasicInputEventHandler {

  MessageCenter center = MessageCenter.getMessageCenter(getClass());
  protected boolean active = false;

  protected List<LocationChangedListener> listeners = new ArrayList<LocationChangedListener>();
  protected CoordinateReferenceSystem crs;
  protected CoordinateReferenceSystem wgs84;
  protected MathTransform transform;
  protected Point pt;

  public PositionTool(CoordinateReferenceSystem crs, LocationChangedListener listener) {
    this(crs);
    listeners.add(listener);
  }

  public PositionTool(CoordinateReferenceSystem crs) {
    this.crs = crs;
    pt = new GeometryFactory().createPoint(new Coordinate(0, 0));
    try {
      wgs84 = CRS.decode("EPSG:4326");
      transform = CRS.findMathTransform(crs, wgs84, true);
    } catch (NoSuchAuthorityCodeException e) {
      wgs84 = DefaultGeographicCRS.WGS84;
    } catch (FactoryException e) {
      center.warn("Unable to create transform", e);
    }
  }

  public void cleanUp() {
    this.listeners.clear();
    this.crs = null;
  }

  @Override
  public void mouseEntered(PInputEvent event) {
    active = true;
  }

  @Override
  public void mouseExited(PInputEvent event) {
    active = false;
    fireListeners(null);
  }

  public void fireListeners(Point pt) {
    //System.out.printf("Point: %s%n", pt);
    for (LocationChangedListener listener : listeners) {
      listener.locationChanged(pt);
    }
  }

  @Override
  public void mouseMoved(PInputEvent event) {
    if (active) {
      Coordinate coord = pt.getCoordinate();
      coord.x = event.getPosition().getX();
      coord.y = event.getPosition().getY();
      //System.out.printf("Untransformed Point: %s%n", pt);
      if (transform == null || crs.equals(wgs84)) {
	fireListeners(pt);
      } else {
	try {
	  JTS.transform(coord, coord, transform);
	  fireListeners(pt);
	} catch (TransformException e) {
	  center.warn("Error transforming location", e);
	  fireListeners(null);
	}
      }
    }
  }

  public void activate(PiccoloMapPanel panel) {
  }

  public void deactivate() {
  }

  public Cursor getCursor() {
    return Cursor.getDefaultCursor();
  }

  public void addLocationChangedListener(LocationChangedListener listener) {
    listeners.add(listener);
  }

}
