package anl.aida.gis3D;

import gov.nasa.worldwind.AnaglyphSceneController;
import gov.nasa.worldwind.BasicModel;
import gov.nasa.worldwind.Configuration;
import gov.nasa.worldwind.Model;
import gov.nasa.worldwind.View;
import gov.nasa.worldwind.avlist.AVKey;
import gov.nasa.worldwind.awt.WorldWindowGLCanvas;
import gov.nasa.worldwind.geom.Angle;
import gov.nasa.worldwind.geom.LatLon;
import gov.nasa.worldwind.geom.Position;
import gov.nasa.worldwind.geom.Sector;
import gov.nasa.worldwind.layers.AirspaceLayer;
import gov.nasa.worldwind.layers.Layer;
import gov.nasa.worldwind.layers.SurfaceShapeLayer;
import gov.nasa.worldwind.render.BasicShapeAttributes;
import gov.nasa.worldwind.render.Material;
import gov.nasa.worldwind.render.SurfacePolygon;
import gov.nasa.worldwind.render.UserFacingText;
import gov.nasa.worldwind.render.airspaces.Airspace;
import gov.nasa.worldwind.render.airspaces.Polygon;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JToolBar;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureIterator;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;

import anl.aida.client.util.ClientUtils;
import anl.aida.data.Axes;
import anl.aida.data.AxisAttributeConstants;
import anl.aida.data.AxisType;
import anl.aida.data.DataFrame;
import anl.aida.data.DataFrameAxis;
import anl.aida.data.DataFrameIndex;
import anl.aida.data.DataFrameStringAxis;
import anl.aida.data.DataUtilities;
import anl.aida.data.StringCoordAxis;
import anl.aida.data.DataUtilities.MinMax;
import anl.aida.gis.FeaturePropertyFinder;
import anl.aida.gis.ShapefileSimplifier;
import anl.aida.palette.Palette;
import anl.aida.palette.PaletteCreator;
import anl.aida.plot.PlotCreationException;
import anl.aida.plot.TimeStepWidget;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.Point;

/**
 * Display producer for 3D gis worldwind based displays.
 * 
 */
public class GIS3DDisplay {

  private static final double MIN_ALT = 5000;
  private static final double MAX_ALT = 100000;
  private static final String NAME = "name";

  static {
    // prevents the incompatible xerces lib from being loaded from
    // somewhere ...
    System.setProperty("javax.xml.parsers.DocumentBuilderFactory",
	"com.sun.org.apache.xerces.internal.jaxp.DocumentBuilderFactoryImpl");
  }

  private Model model;
  private WorldWindowGLCanvas wwglCanvas;
  private String displayMode = AnaglyphSceneController.DISPLAY_MODE_MONO;
  private JPanel panel;
  private DataFrame frame;
  private int tsIndex = 0;
  private MinMax minMax;
  private Map<String, Polygon> polys = new HashMap<String, Polygon>();
  private FeatureOverlay overlay;
  private Sector bounds;

  public GIS3DDisplay() {
    Configuration.setValue(AVKey.SCENE_CONTROLLER_CLASS_NAME, AnaglyphSceneController.class
	.getName());
    // Adjust configuration values before instanciation
    Configuration.setValue(AVKey.INITIAL_LATITUDE, 0);
    Configuration.setValue(AVKey.INITIAL_LONGITUDE, 0);
    Configuration.setValue(AVKey.INITIAL_ALTITUDE, 5000e3);
    Configuration.setValue(AVKey.INITIAL_PITCH, 40);
    // Configuration.setValue(AVKey.GLOBE_CLASS_NAME,
    // EarthFlat.class.getName());
    // Configuration.setValue(AVKey.VIEW_CLASS_NAME,
    // FlatOrbitView.class.getName());

    // Configuration.setValue(AVKey.GLOBE_CLASS_NAME,
    // EarthFlat.class.getName());
    // Configuration.setValue(AVKey.VIEW_CLASS_NAME,
    // FlatOrbitView.class.getName());

    model = new BasicModel();

    String[] offLayers = { "Stars", "Atmosphere", "Compass", "World Map" };
    for (Layer layer : model.getLayers()) {
      layer.setEnabled(true);
      // System.out.println(layer.getName());
    }
    for (String name : offLayers) {
      model.getLayers().getLayerByName(name).setEnabled(false);
    }
  }

  private void calcPolyAltitude(Polygon poly) {
    DataFrameStringAxis axis = (DataFrameStringAxis) frame.getAxes().getAxis(
	AxisType.NAMED_LOCATION);
    double interval = minMax.getMax() - minMax.getMin();
    DataFrameIndex index = frame.getIndex();
    String name = poly.getValue(NAME).toString();
    int catIndex = axis.getIndex(name.toLowerCase());
    double val = minMax.getMin();
    if (catIndex != -1) {
      index.set(tsIndex, catIndex);
      val = frame.getDouble(index);
    }
    double alt = ((val - minMax.getMin()) / interval) * MAX_ALT + MIN_ALT;
    // System.out.println(name + ": alt = " + alt);
    poly.setAltitudes(0, alt);
  }

  private void updatePolygons() {
    for (Polygon poly : polys.values()) {
      calcPolyAltitude(poly);
    }
    wwglCanvas.redraw();
  }

  public void init(DataFrame dataFrame) throws PlotCreationException {
    this.frame = dataFrame;
    Axes<DataFrameAxis> axes = dataFrame.getAxes();
    DataFrameStringAxis locationAxis = (DataFrameStringAxis) axes.getAxis(AxisType.NAMED_LOCATION);
    if (locationAxis == null)
      throw new PlotCreationException(new IllegalArgumentException(
	  "DataFrame does not contain a named location axis."));

    String mapFile = locationAxis.getAttribute(AxisAttributeConstants.MAP_FILE);
    if (mapFile == null)
      throw new PlotCreationException(new IllegalArgumentException(
	  "Named location axis does not contain a map file attribute."));

    String nameAttribute = locationAxis.getAttribute(AxisAttributeConstants.NAME_ATTRIBUTE);
    if (nameAttribute == null)
      throw new PlotCreationException(new IllegalArgumentException(
	  "Named location axis does not contain a attribute name attribute."));

    minMax = DataUtilities.minMax(dataFrame);

    FeatureIterator<SimpleFeature> fIter = null;
    FeatureCollection<SimpleFeatureType, SimpleFeature> features = null;

    try {
      File file = new File(mapFile);
      ShapefileSimplifier simplifier = new ShapefileSimplifier();
      features = simplifier.simplify(file, .005f);
      overlay = new FeatureOverlay(new FeaturePropertyFinder(nameAttribute, features), dataFrame, model);

      SurfaceShapeLayer noDataLayer = new SurfaceShapeLayer();
      noDataLayer.setName("Entities (No Data)");
      BasicShapeAttributes noDataAttribs = new BasicShapeAttributes();
      noDataAttribs.setDrawInterior(false);
      noDataAttribs.setDrawOutline(true);
      noDataAttribs.setOutlineMaterial(new Material(Color.GRAY));
      noDataAttribs.setOutlineOpacity(1);
      noDataAttribs.setOutlineWidth(3);

      AirspaceLayer layer = new AirspaceLayer();
      layer.setName("Entities");
      ElevatedTextLayer textLayer = new ElevatedTextLayer();
      textLayer.setName("Entity Labels");

      fIter = features.features();
      PaletteCreator creator = new PaletteCreator();
      Palette palette = creator.createPalettes(features.size()).get(0);
      int colorIndex = 0;

      SectorCreator sectorCreator = new SectorCreator();
      while (fIter.hasNext()) {
	SimpleFeature feature = fIter.next();
	Geometry geom = (Geometry) feature.getDefaultGeometry();
	Color color = palette.getColor(colorIndex++);
	for (int i = 0, n = geom.getNumGeometries(); i < n; i++) {
	  Geometry polygon = geom.getGeometryN(i);
	  polygon = polygon.buffer(-.05);
	  List<LatLon> coords = new ArrayList<LatLon>();
	  for (Coordinate coord : polygon.getCoordinates()) {
	    LatLon latLon = new LatLon(Angle.fromDegrees(coord.y), Angle.fromDegrees(coord.x));
	    coords.add(latLon);
	    sectorCreator.addLatLon(latLon);
	  }
	  

	  Point pt = polygon.getCentroid();
	  String name = feature.getAttribute(nameAttribute).toString();
	  if (locationAxis.getIndex(name.toLowerCase()) == StringCoordAxis.INDEX_NOT_FOUND) {
	    SurfacePolygon shape = new SurfacePolygon(coords);
	    shape.setAttributes(noDataAttribs);
	    noDataLayer.addRenderable(shape);
	    
	    Position pos = new Position(Angle.fromDegrees(pt.getY()), Angle.fromDegrees(pt.getX()), MIN_ALT);
	    textLayer.addText(new UserFacingText(name, pos));
	  } else {
	    Polygon poly = new Polygon();
	    poly.setLocations(coords);
	    poly.setTerrainConforming(true, true);
	    poly.setValue(NAME, name);
	    setupDefaultMaterial(poly, color);
	    calcPolyAltitude(poly);
	    polys.put(name, poly);
	    layer.addAirspace(poly);
	    Position pos = new Position(Angle.fromDegrees(pt.getY()), Angle.fromDegrees(pt.getX()), poly.getAltitudes()[1] + 10000);
	    UserFacingText text = new UserFacingText(name, pos);
	    text.setColor(new Color(1f, 1f, 1f, 1f));
	    textLayer.addText(text);
	  }
	}
      }

      bounds = sectorCreator.createSector();
      model.getLayers().add(layer);
      model.getLayers().add(noDataLayer);
      model.getLayers().add(textLayer);
      

    } catch (Exception ex) {
      throw new PlotCreationException(ex);
    } finally {
      if (fIter != null)
	features.close(fIter);
    }
  }

  private void setupDefaultMaterial(Airspace a, Color color) {
    a.getAttributes().setDrawOutline(false);
    a.getAttributes().setMaterial(new Material(color));
    a.getAttributes().setOpacity(0.8);
    a.getAttributes().setOutlineOpacity(0.9);
    a.getAttributes().setOutlineWidth(3.0);
  }
  
  private JToolBar createToolBar() {
    JToolBar toolbar = new JToolBar();
    toolbar.setFloatable(false);
    
    JButton button = new JButton(ClientUtils.loadIcon("InitialExtent.png"));
    button.setBorderPainted(false);
    button.setToolTipText("Zoom to Initial Extent");
    button.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent evt) {
	flyToBounds();
      }
    });
    toolbar.add(button);
    
    OptionsWidget widget = new OptionsWidget(model);
    toolbar.add(widget.getButton());
    
    toolbar.addSeparator();
    
    TimeStepWidget ts = new TimeStepWidget();
    ts.init(frame.getAxes(), 0);
    ts.setTimerDelay(100);
    toolbar.add(ts);
    ts.addSpinnerListener(new TimeStepListener(ts));
    
    return toolbar;
  }

  private void createPanel() {
    wwglCanvas = new WorldWindowGLCanvas();
    wwglCanvas.setModel(model);
    // InfoOverlay overlay = new InfoOverlay(model);
    // wwglCanvas.addSelectListener(overlay);
    wwglCanvas.addPositionListener(overlay);
    AnaglyphSceneController asc = (AnaglyphSceneController) wwglCanvas.getSceneController();
    asc.setDisplayMode(this.displayMode);

    panel = new JPanel();
    panel.setLayout(new BorderLayout());
    wwglCanvas.setPreferredSize(new Dimension(800, 600));
    panel.add(wwglCanvas, BorderLayout.CENTER);
    panel.add(createToolBar(), BorderLayout.NORTH);
    
    zoomToBounds();
    wwglCanvas.getView().setPitch(Angle.fromDegrees(40));
    
    
    // StatusBar statusBar = new StatusBar();
    // statusBar.setEventSource(wwglCanvas);
    // panel.add(statusBar, BorderLayout.PAGE_END);

    // JPanel leftPanel = new JPanel(new BorderLayout());

    // layerPanel = new LayerPanel(wwglCanvas);
    // leftPanel.add(layerPanel, BorderLayout.WEST);

    // panel.add(leftPanel, BorderLayout.WEST);

  }

  public void zoomToBounds() {
    Position pos = GIS3DUtils.getEyePositionForSector(wwglCanvas, bounds);
    View view = wwglCanvas.getView();
    view.setEyePosition(pos);
    view.firePropertyChange(AVKey.VIEW,  null, view);
  }
  
  public void flyToBounds() {
    Position pos = GIS3DUtils.getEyePositionForSector(wwglCanvas, bounds);
    View view = wwglCanvas.getView();
    view.goTo(pos, pos.getElevation());
  }

  public JPanel getPanel() {
    if (panel == null) {
      createPanel();
    }
    return panel;
  }

  class TimeStepListener implements ChangeListener {

    private TimeStepWidget ts;

    public TimeStepListener(TimeStepWidget ts) {
      this.ts = ts;
    }

    @Override
    public void stateChanged(ChangeEvent e) {
      if (tsIndex != ts.getTime()) {
	tsIndex = ts.getTime();
	overlay.timestepChanged(tsIndex);
	updatePolygons();
      }
    }
  }

}
