package anl.aida.palette;

import java.awt.Color;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

/**
 * Maps colors to a range of values.
 * 
 * @version $Revision$ $Date$
 */
public class ColorMap {

  public enum IntervalType {
    CUSTOM, AUTOMATIC
  }

  public enum PaletteType {
    SEQUENTIAL, QUALITATIVE, DIVERGING
  }

  public static IntervalType getIntervalType(String type) {
    for (IntervalType iType : IntervalType.values()) {
      if (iType.toString().equalsIgnoreCase(type))
	return iType;
    }

    return null;
  }

  public static PaletteType getPaletteType(String type) {
    for (PaletteType pType : PaletteType.values()) {
      if (pType.toString().equalsIgnoreCase(type))
	return pType;
    }

    return null;
  }

  private Palette palette;
  private double[] intervals;
  private IntervalType intervalType = IntervalType.AUTOMATIC;
  private PaletteType paletteType = PaletteType.QUALITATIVE;
  private double min, max;
  private DecimalFormat format;

  public ColorMap() {
    palette = new Palette(new Color[0], "");
  }

  public ColorMap(Palette palette, double min, double max) {
    this.min = min;
    this.max = max;
    this.palette = new Palette(palette);
    calcIntervals(palette, min, max);
  }

  private void calcIntervals(Palette palette, double min, double max) {
    int colorCount = palette.getColorCount();
    this.intervals = new double[colorCount + 1];
    double interval = (max - min) / colorCount;
    for (int i = 0; i < colorCount; i++) {
      intervals[i] = min + (i * interval);
    }
  }

  public ColorMap(Palette palette, List<Double> steps) {
    Collections.sort(steps);
    this.min = steps.get(0);
    this.max = steps.get(steps.size() - 1);
    this.palette = new Palette(palette);
    int colorCount = palette.getColorCount();
    this.intervals = new double[colorCount + 1];

    for (int i = 0; i < steps.size(); i++) {
      intervals[i] = steps.get(i);
    }
  }
  
  public ColorMap(Palette palette, double... steps) {
    Arrays.sort(steps);
    this.min = steps[0];
    this.max = steps[steps.length - 1];
    this.palette = new Palette(palette);
    int colorCount = palette.getColorCount();
    this.intervals = new double[colorCount + 1];

    for (int i = 0; i < steps.length; i++) {
      intervals[i] = steps[i];
    }
  }

  public int getMaxIndex() {
    return intervals.length - 2;
  }
  
  public Color getColorForValue(double val) {
    if (val <= min) return palette.getColor(0);
    if (val >= max) return palette.getColor(getMaxIndex());
    
    for (int index = 0; index < intervals.length - 1; index++) {
      double lowerBound = intervals[index];
      double upperBound = intervals[index + 1];
      if (val >= lowerBound && val < upperBound) return getColor(index);
    }
    return null;
  }


  public double getStep(int index) {
    if (index > intervals.length - 1)
      return intervals[intervals.length - 1];

    if (index < 0)
      return intervals[0];

    return intervals[index];
  }

  public double getMax() {
    return max;
  }

  public double getMin() {
    return min;
  }

  public double[] getIntervals() {
    return intervals;
  }

  public PaletteType getPaletteType() {
    return paletteType;
  }

  public void setPaletteType(PaletteType paletteType) {
    this.paletteType = paletteType;
  }

  public Palette getPalette() {
    return palette;
  }

  public void setPalette(Palette palette) {
    int prevIntervals = this.palette.getColorCount();
    int curIntervals = palette.getColorCount();
    this.palette = palette;

    if (prevIntervals != curIntervals)
      calcIntervals(palette, min, max);
  }

  public Color getColor(int index) {
    return palette.getColor(index);
  }
  
  public void setColor(int index, Color color) {
    palette.setColor(index, color);
  }

  public int getColorCount() {
    return palette.getColorCount();
  }

  public double getIntervalStart(int index) {
    return intervals[index];
  }

  public void setIntervalStart(int index, double start) throws NumberFormatException {
    int last = intervals.length - 1;

    if (paletteType == PaletteType.SEQUENTIAL && index > 0) {
      if (index < last && start <= intervals[index - 1])
	return;

      if (index < last && start >= intervals[index + 1])
	return;

      if (index == 0)
	min = start;

      if (index == last)
	max = start + (start - intervals[index - 1]);
    }

    intervals[index] = start;
  }

  public IntervalType getIntervalType() {
    return intervalType;
  }

  public void setIntervalType(IntervalType intervalType) {
    this.intervalType = intervalType;
  }

  public NumberFormat getNumberFormat() {
    if (format != null)
      return format;

    String strMin = String.valueOf(min);
    String strMax = String.valueOf(max);

    if (strMin.contains("E-") || strMax.contains("E-"))
      return new DecimalFormat("0.000E0");

    return new DecimalFormat("0.000");
  }

  public void setNumberFormat(NumberFormat format) throws Exception {
    if (!(format instanceof DecimalFormat))
      throw new Exception("Number format: " + format.toString() + " is not supported.");

    this.format = (DecimalFormat) format;
  }
}
