/**
 * 
 */
package anl.aida.plot;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.jfree.data.time.FixedMillisecond;
import org.jfree.data.time.TimeSeries;
import org.jfree.data.time.TimeSeriesCollection;

import simphony.util.messages.MessageCenter;
import anl.aida.client.Cache;
import anl.aida.client.CacheElement;
import anl.aida.client.Keyword;
import anl.aida.core.ScoredTermsCollection;
import anl.aida.util.DateRange;

/**
 * 
 */
public class CacheTimeSeriesParams implements TimeSeriesParameters {

  private static MessageCenter msg = MessageCenter.getMessageCenter(CacheTimeSeriesParams.class);

  private TimeSeriesCollection dataset = new TimeSeriesCollection();
  private double min, max;

  private List<ProgressListener<DateRange>> listeners = new ArrayList<ProgressListener<DateRange>>();
  
  public static CacheTimeSeriesParams createParams(Cache cache, Collection<? extends Keyword> keywords, 
      Iterable<DateRange> ranges, ProgressListener<DateRange>... listeners) {
    
    CacheTimeSeriesParams params = new CacheTimeSeriesParams();
    for (ProgressListener<DateRange> listener : listeners) {
      params.addProgressListener(listener);
    }
    params.init(cache, keywords, ranges);
    return params;
  }

  private CacheTimeSeriesParams() {
    min = Double.POSITIVE_INFINITY;
    max = Double.NEGATIVE_INFINITY;
  }
  
  private void init(Cache cache, Collection<? extends Keyword> keywords, Iterable<DateRange> ranges) {
    for (DateRange range : ranges) {
      fireUpdate(range); 
      CacheElement element = cache.getElement(range);
      if (element == null) {
	msg.warn("Synthetic range required; not yet implemented");
      } else {
	ScoredTermsCollection stc = element.getSTC();
	for (Keyword keyword : keywords) {
	  int docCount = 0;
	  int index = stc.getTermIndex(keyword.getTerm());
	  if (index != -1) {
	    docCount = stc.getTerm(index).getDocumentCount();
	  }
	  updateSeries(keyword, range, docCount);
	}
	element.release();
      }
    }
  }

  private void updateSeries(Keyword keyword, DateRange range, int docCount) {
    TimeSeries series = dataset.getSeries(keyword.getLabel());
    if (series == null) {
      series = new TimeSeriesMinMax(keyword.getLabel());
      dataset.addSeries(series);
    }
    series.add(new FixedMillisecond(range.getEnd()), docCount);
    min = Math.min(min, docCount);
    max = Math.max(max, docCount);
  }

  /**
   * Adds listener that will list for progress in the dataset creation.
   * 
   * @param listener
   *          the listener to add.
   */
  public void addProgressListener(ProgressListener<DateRange> listener) {
    listeners.add(listener);
  }

  private void fireUpdate(DateRange range) {
    for (ProgressListener<DateRange> listener : listeners) {
      listener.progressUpdated(range);
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.plot.TimeSeriesParameters#getDataset()
   */
  @Override
  public TimeSeriesCollection getDataset() {
    return dataset;
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.plot.TimeSeriesParameters#getMax()
   */
  @Override
  public double getMax() {
    return max;
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.plot.TimeSeriesParameters#getMin()
   */
  @Override
  public double getMin() {
    return min;
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.plot.TimeSeriesParameters#getSeriesHeading()
   */
  @Override
  public String getSeriesHeading() {
    return "Keyword";
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.plot.TimeSeriesParameters#getTitle()
   */
  @Override
  public String getTitle() {
    return "Keyword Document Counts";
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.plot.TimeSeriesParameters#getUnitsLabel()
   */
  @Override
  public String getUnitsLabel() {
    return "";
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.plot.TimeSeriesParameters#getXAxisLabel()
   */
  @Override
  public String getXAxisLabel() {
    return "Time Slices (Ending Date)";
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.aida.plot.TimeSeriesParameters#getYAxisLabel()
   */
  @Override
  public String getYAxisLabel() {
    return "Document Count";
  }

}
