package anl.aida.plot;

import java.awt.Dimension;
import java.awt.geom.Point2D;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import anl.aida.client.net.ValueNode;
import edu.uci.ics.jung.graph.Graph;

/**
 *Feb 11, 2009 11:25:50 AM
 */
public class CircleLayout<T extends ValueNode> {

  private double radius;
  protected Map<T, Point2D> locations = new HashMap<T, Point2D>();
  protected Dimension size;

  public CircleLayout(Collection<T> objs, Dimension size) {
    this.size = size;
    initNodes(objs);
  }

  protected void initNodes(Collection<T> objs) {
    Point2D.Double tmp = new Point2D.Double(0, 0);
    locations.clear();
    for (T obj : objs) {
      locations.put(obj, tmp);
    }
  }

  public Point2D getLocation(T obj) {
    return locations.get(obj);
  }

  /**
   * Called when the display area size has changed.
   * 
   * @param size
   *          the new size
   */
  public void sizeChanged(Dimension size) {
    this.size = size;
    this.radius = 0;

    initialize();
  }

  /**
   * Swaps the layouts current graph with the specified one. This assumes the
   * graphs have equivalent term nodes and it preserves the locations.
   * 
   * @param graph
   *          the graph to swap in.
   */
  public void swapGraph(Graph<ValueNode, Object> graph) {
    Map<String, Point2D> locs = new HashMap<String, Point2D>();
    for (Map.Entry<T, Point2D> entry : locations.entrySet()) {
      locs.put(entry.getKey().getLabel(), entry.getValue());
    }
    this.locations.clear();

    for (ValueNode node : graph.getVertices()) {
      Point2D pt = locs.get(node.getLabel());
      locations.put((T) node, pt);
    }
  }

  /**
   * Initializes the layout with the specified size.
   * 
   * @param size
   *          the size
   */
  public void initialize(Dimension size) {
    this.size = size;
    radius = 0;
    initialize();
  }

  /**
   * Sets the location of the specified node.
   * 
   * @param node
   *          the node whose location we want to set
   * @param x
   *          the x coordinate
   * @param y
   *          the y coordinate
   */
  public void setLocation(T node, double x, double y) {
    Point2D location = locations.get(node);
    if (location == null) {
      location = new Point2D.Double();
      locations.put(node, location);
    }
    location.setLocation(x, y);
  }

  public void initialize() {
    double width = size.getWidth();
    double height = size.getHeight();

    if (radius <= 0) {
      radius = 0.45 * (height < width ? height : width);
    }

    int i = 0;
    int nodeCount = locations.size();
    for (Map.Entry<T, Point2D> entry : locations.entrySet()) {
      Point2D point = new Point2D.Double();
      float angle = (float) (2 * Math.PI * i) / nodeCount;
      point
	  .setLocation(Math.cos(angle) * radius + width / 2, Math.sin(angle) * radius + height / 2);
      entry.setValue(point);
      i++;
    }
  }

  public Point2D getCentroid() {
    double x = 0, y = 0;
    for (Point2D point : locations.values()) {
      x += point.getX();
      y += point.getY();
    }

    return new Point2D.Double(x / locations.size(), y / locations.size());
  }
}
