/**
 * 
 */
package anl.aida.plot;

import java.util.List;

import org.jfree.data.time.FixedMillisecond;
import org.jfree.data.time.TimeSeries;
import org.jfree.data.time.TimeSeriesCollection;

import anl.aida.data.Axes;
import anl.aida.data.DataFrame;
import anl.aida.data.DataFrameAxis;
import anl.aida.data.DataFrameIndex;
import anl.aida.data.Units;

/**
 * Adapts one or more DataFrames to the TimeSeriesParameters 
 * interface.
 * 
 */
public class DFTimeSeriesParams implements TimeSeriesParameters {
  
  private double min, max;
  private TimeSeriesCollection dataset;
  private String seriesHeading, unitsLabel, xAxisLabel, yAxisLabel,
  		title;
  
  public DFTimeSeriesParams(List<DataFrame> frames) {
    this(frames, new int[0]);
  }
  
  public DFTimeSeriesParams(List<DataFrame> frames, int[] otherAxisIndices) {
    createDataset(frames, otherAxisIndices);
    seriesHeading = "Series Name";
    if (frames.size() == 1) {
      seriesHeading = frames.get(0).getVariable().getName();

      if (frames.get(0).getAxes().size() > 1) {
	seriesHeading = frames.get(0).getAxes().getNonTimeAxis().getName();
      }
    }
    unitsLabel = Units.getFormattedName(frames.get(0).getVariable().getUnit());
    yAxisLabel = seriesHeading;
    if (unitsLabel.length() > 0) yAxisLabel += " (" + unitsLabel + ")";
    
    xAxisLabel = frames.get(0).getAxes().getTimeAxis().getName();
    title = frames.get(0).getVariable().getName();
  }
  
  private void createDataset(List<DataFrame> frames, int[] otherAxisIndices) {
    min = Double.POSITIVE_INFINITY;
    max = Double.NEGATIVE_INFINITY;
    dataset = new TimeSeriesCollection();
    for (DataFrame frame : frames) {
      if (frame.getAxes().size() == 1) {
	addFrame1Axis(frame);
      } else {
	if (otherAxisIndices.length == 0) {
	  addFrame2Axis(frame);
	} else {
	  addFrame2Axis(frame, otherAxisIndices);
	}
      }
    }
  }

  private void addFrame1Axis(DataFrame frame) {
    DataFrameIndex index = frame.getIndex();
    Axes<DataFrameAxis> axes = frame.getAxes();
    // only a time axis, so only a single series, use the variables as the
    // series name
    TimeSeries series = new TimeSeriesMinMax(frame.getVariable().getName());
    int extent = axes.getTimeAxis().getExtent();
    for (int t = 0; t < extent; t++) {
      index.setTime(t);
      double val = frame.getDouble(index);
      if (val < min)
	min = val;
      if (val > max)
	max = val;
      series.add(new FixedMillisecond(axes.getDate(t)), val);
    }
    dataset.addSeries(series);
  }

  private void addFrame2Axis(DataFrame frame) {
    DataFrameIndex index = frame.getIndex();
    Axes<DataFrameAxis> axes = frame.getAxes();
    DataFrameAxis axis = axes.getNonTimeAxis();

    for (int i = 0; i < axis.getExtent(); i++) {
      String axisVal = axis.getValueAsString(i);
      TimeSeries series = new TimeSeriesMinMax(axisVal);
      for (int t = 0; t < axes.getTimeAxis().getExtent(); t++) {
	index.set(t, i);
	double val = frame.getDouble(index);
	if (val < min)
	  min = val;
	if (val > max)
	  max = val;
	series.add(new FixedMillisecond(axes.getDate(t)), val);
      }
      dataset.addSeries(series);

    }
  }

  private void addFrame2Axis(DataFrame frame, int[] otherAxisIndices) {
    DataFrameIndex index = frame.getIndex();
    Axes<DataFrameAxis> axes = frame.getAxes();
    DataFrameAxis axis = axes.getNonTimeAxis();
    for (int axisIndex : otherAxisIndices) {
      String axisVal = axis.getValueAsString(axisIndex);
      TimeSeries series = new TimeSeriesMinMax(axisVal);
      for (int t = 0; t < axes.getTimeAxis().getExtent(); t++) {
	index.set(t, axisIndex);
	double val = frame.getDouble(index);
	if (val < min)
	  min = val;
	if (val > max)
	  max = val;
	series.add(new FixedMillisecond(axes.getDate(t)), val);
      }
      dataset.addSeries(series);
    }
  }

  /* (non-Javadoc)
   * @see anl.aida.plot.TimeSeriesParameters#getDataset()
   */
  @Override
  public TimeSeriesCollection getDataset() {
    return dataset;
  }

  /* (non-Javadoc)
   * @see anl.aida.plot.TimeSeriesParameters#getMax()
   */
  @Override
  public double getMax() {
    return max;
  }

  /* (non-Javadoc)
   * @see anl.aida.plot.TimeSeriesParameters#getMin()
   */
  @Override
  public double getMin() {
    return min;
  }

  /* (non-Javadoc)
   * @see anl.aida.plot.TimeSeriesParameters#getSeriesHeading()
   */
  @Override
  public String getSeriesHeading() {
    return seriesHeading;
  }

  /* (non-Javadoc)
   * @see anl.aida.plot.TimeSeriesParameters#getTitle()
   */
  @Override
  public String getTitle() {
    return title;
  }

  /* (non-Javadoc)
   * @see anl.aida.plot.TimeSeriesParameters#getUnitsLabel()
   */
  @Override
  public String getUnitsLabel() {
    return unitsLabel;
  }

  /* (non-Javadoc)
   * @see anl.aida.plot.TimeSeriesParameters#getXAxisLabel()
   */
  @Override
  public String getXAxisLabel() {
    return xAxisLabel;
  }

  /* (non-Javadoc)
   * @see anl.aida.plot.TimeSeriesParameters#getYAxisLabel()
   */
  @Override
  public String getYAxisLabel() {
    return yAxisLabel;
  }

}
