package anl.aida.plot;

import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.geom.Ellipse2D;

import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JToolBar;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.labels.XYItemLabelGenerator;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.Range;
import org.jfree.data.xy.XYDataset;
import org.jfree.ui.RectangleInsets;

import anl.aida.data.DataFrame;
import anl.aida.data.DataFrameAxis;
import anl.aida.data.DataFrameIndex;
import anl.aida.data.DataUtilities;
import anl.aida.data.Units;
import anl.aida.palette.Palette;
import anl.aida.palette.PaletteCreator;

/**
 * Scatter plot that plots a single point for each group / category in two data
 * frames. This assumes that the category / group and time axes of the two
 * DataFrames are compatible.
 * 
 */
public class GroupScatterPlot {
  
  private static class LabelGenerator implements XYItemLabelGenerator {

    @Override
    public String generateLabel(XYDataset dataset, int series, int item) {
      return dataset.getSeriesKey(series).toString();
    }
  }

  private ScatterGroupDataset dataset;
  private JFreeChart chart;
  private DataFrame xFrame, yFrame;
  private ChartPanel chartPanel;
  private JPanel panel;
  private Range xRange, yRange;
  

  public GroupScatterPlot(DataFrame xFrame, DataFrame yFrame) {
    this.xFrame = xFrame;
    this.yFrame = yFrame;
    createDataset(xFrame, yFrame);
    createChart();
    chartPanel = new ChartPanel(chart);
    chartPanel.setMouseWheelEnabled(true);
    chartPanel.setFillZoomRectangle(true);
  }

  public JPanel getPanel() {
    if (panel == null) {
      panel = new JPanel(new BorderLayout());
      JToolBar bar = new JToolBar();
      JButton button = new JButton("Reset Range");
      bar.add(button);
      button.addActionListener(new ActionListener() {
	public void actionPerformed(ActionEvent evt) {
	  XYPlot plot = (XYPlot) chart.getPlot();
	  plot.getDomainAxis().setRange(xRange, true, true);
	  plot.getRangeAxis().setRange(yRange, true, true);
	}
      });
      bar.addSeparator();
      final TimeStepWidget widget = new TimeStepWidget();
      widget.init(xFrame.getAxes(), 0);
      widget.setTimerDelay(250);
      widget.addSpinnerListener(new ChangeListener() {
	@Override
	public void stateChanged(ChangeEvent e) {
	  int timestep = widget.getTime();
	  dataset.setTimeStep(timestep);
	}
      });
      bar.add(widget);
      bar.setFloatable(false);
      panel.add(bar, BorderLayout.NORTH);
      panel.add(chartPanel, BorderLayout.CENTER);
    }

    return panel;
  }

  private void createDataset(DataFrame xFrame, DataFrame yFrame) {
    dataset = new ScatterGroupDataset(xFrame, yFrame);
    dataset.setTimeStep(0);
    DataFrameAxis groupAxis = xFrame.getAxes().getNonTimeAxis();
    for (int group = 0; group < groupAxis.getExtent(); group++) {
      dataset.addSeries(group);
    }
  }

  private void createChart() {
    String xAxisLabel = xFrame.getVariable().getName() + " "
	+ Units.getFormattedName(xFrame.getVariable().getUnit());
    String yAxisLabel = yFrame.getVariable().getName() + " "
	+ Units.getFormattedName(yFrame.getVariable().getUnit());
    String title = xFrame.getVariable().getName() + " x " + yFrame.getVariable().getName();
    chart = ChartFactory.createScatterPlot(title, xAxisLabel, yAxisLabel, dataset,
	PlotOrientation.VERTICAL, true, true, false);

    chart.setBackgroundPaint(Color.white);
    // chart.addProgressListener(this);

    XYPlot plot = (XYPlot) chart.getPlot();
    plot.setBackgroundPaint(Color.WHITE);
    plot.setDomainGridlinePaint(Color.GRAY);
    plot.setRangeGridlinePaint(Color.GRAY);
    plot.setAxisOffset(new RectangleInsets(5.0, 5.0, 5.0, 5.0));
    plot.setDomainCrosshairVisible(true);
    plot.setRangeCrosshairVisible(true);

    PaletteCreator creator = new PaletteCreator();
    Palette palette = creator.createPalettes(dataset.getSeriesCount()).get(0);
    XYLineAndShapeRenderer renderer = (XYLineAndShapeRenderer) plot.getRenderer();
    renderer.setBaseItemLabelGenerator(new LabelGenerator());
    renderer.setItemLabelAnchorOffset(4);
    renderer.setUseFillPaint(true);
   
    for (int i = 0; i < dataset.getSeriesCount(); i++) {
      renderer.setSeriesShape(i, new Ellipse2D.Double(-6.0, -6.0, 12.0, 12.0));
      Color pColor = palette.getColor(i);
      Color color = new Color(pColor.getRed() / 255f, pColor.getGreen() / 255f, pColor.getBlue() / 255f, .65f);
      renderer.setSeriesFillPaint(i, color);
      renderer.setSeriesOutlineStroke(i, new BasicStroke(1.0f));
      renderer.setUseOutlinePaint(true);
      renderer.setSeriesItemLabelsVisible(i, true);
    }

    DataUtilities.MinMax minMax = getMinMax(xFrame);
    double min = minMax.getMin();
    double max = minMax.getMax();
    NumberAxis axis = (NumberAxis) plot.getDomainAxis();
    double tenth = (max - min) / 10;
    if (tenth == 0) {
      axis.setRangeAboutValue(min, .01);
    } else {
      axis.setRange(min - tenth, max + tenth);
    }
    xRange = new Range(axis.getRange().getLowerBound(), axis.getRange().getUpperBound());
    axis.setAutoRange(false);

    minMax = getMinMax(yFrame);
    min = minMax.getMin();
    max = minMax.getMax();
    axis = (NumberAxis) plot.getRangeAxis();
    tenth = (max - min) / 10;
    if (tenth == 0) {
      axis.setRangeAboutValue(min, .01);
    } else {
      axis.setRange(min - tenth, max + tenth);
    }
    yRange = new Range(axis.getRange().getLowerBound(), axis.getRange().getUpperBound());
    axis.setAutoRange(false);
  }

  private DataUtilities.MinMax getMinMax(DataFrame frame) {
    double min = Double.POSITIVE_INFINITY;
    double max = Double.NEGATIVE_INFINITY;

    DataFrameIndex index = frame.getIndex();
    DataFrameAxis axis = frame.getAxes().getNonTimeAxis();
    DataFrameAxis timeAxis = frame.getAxes().getTimeAxis();
    for (int t = 0; t < timeAxis.getExtent(); t++) {
      for (int i = 0; i < axis.getExtent(); i++) {
	index.set(t, i);
	double val = frame.getDouble(index);
	if (val < min)
	  min = val;
	if (val > max)
	  max = val;
      }
    }

    return new DataUtilities.MinMax(min, max);
  }

}
