package anl.aida.plot;

import java.awt.BorderLayout;
import java.awt.Color;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.ExecutionException;

import javax.swing.JPanel;
import javax.swing.SwingWorker;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.geotools.filter.text.cql2.CQL;
import org.geotools.filter.text.cql2.CQLException;
import org.geotools.referencing.CRS;
import org.geotools.styling.FeatureTypeStyle;
import org.geotools.styling.Rule;
import org.geotools.styling.Style;
import org.geotools.styling.StyleBuilder;
import org.geotools.styling.Symbolizer;
import org.geotools.styling.TextSymbolizer;
import org.opengis.filter.Filter;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import simphony.util.messages.MessageCenter;

import anl.aida.client.gui.CreateTSWorker;
import anl.aida.client.gui.AIDAGui;
import anl.aida.client.gui.ProgressLayer;
import anl.aida.data.Axes;
import anl.aida.data.AxisAttributeConstants;
import anl.aida.data.AxisType;
import anl.aida.data.DataFrame;
import anl.aida.data.DataFrameAxis;
import anl.aida.data.DataFrameIndex;
import anl.aida.data.DataUtilities;
import anl.aida.data.Units;
import anl.aida.data.DataUtilities.MinMax;
import anl.aida.gis.FeaturePropertyFinder;
import anl.aida.gis.GISDisplay;
import anl.aida.palette.ColorMap;
import anl.aida.palette.Palette;

public class MapPlotCreator extends SwingWorker<Object, Object> {
  
  private static MessageCenter center = MessageCenter.getMessageCenter(MapPlotCreator.class);
  
  private JPanel panel;
  private DataFrame dataFrame;
  private ColorMap map;
  private CoordinateReferenceSystem crs;
  private double min, max;
  private String mapFile;
  private GISTimeStepStyler styler;
  private AIDAGui gui;
  private String dockableId;
  
  public MapPlotCreator(String dockableId, DataFrame frame, JPanel panel, AIDAGui gui) {
    this.dataFrame = frame;
    this.panel = panel;
    this.gui = gui;
    this.dockableId = dockableId;
  }
  
  static {
    System.setProperty("org.geotools.referencing.forceXY", "true");
  }
  
  @Override
  protected Object doInBackground() throws Exception {
    Axes<DataFrameAxis> axes = dataFrame.getAxes();
    DataFrameAxis axis = axes.getAxis(AxisType.NAMED_LOCATION);
    if (axis == null)
	throw new PlotCreationException(new IllegalArgumentException("DataFrame does not contain a named location axis."));
    String crsCode = axis.getAttribute(AxisAttributeConstants.CRS_CODE);
    mapFile = axis.getAttribute(AxisAttributeConstants.MAP_FILE);
    if (crsCode == null)
	throw new PlotCreationException(new IllegalArgumentException(
	    "Named location axis does not contain an EPSG code attribute."));
    if (mapFile == null)
	throw new PlotCreationException(new IllegalArgumentException(
	    "Named location axis does not contain a map file attribute."));

    MinMax minMax = DataUtilities.minMax(dataFrame);
    min = minMax.getMin();
    max = minMax.getMax();
    //ColorMap map = new ColorMap(Palette.getPalette("OrRd", 8), minMax.getMin(), minMax.getMax());
    map = new ColorMap(Palette.getPalette("OrRd", 8), minMax.getMin(), 2.5, 5, 10, 15, 20, 25, 50, minMax.getMax());
    crs = CRS.decode(crsCode);
    styler = new GISTimeStepStyler(dataFrame, map);
    return null;
  }

  @Override
  protected void done() {
    try {
      get();
      LegendPanel legend = new LegendPanel(map, min, max,  Units.getFormattedName(dataFrame.getVariable().getUnit()));
      GISDisplay display = new GISDisplay(crs, legend);
      File file = new File(mapFile);
      
      display.addLayer(file, styler.getStyle(0));
      TimeStepWidget widget = display.initTimeStepWidget(dataFrame, styler);
      
      FeaturePropertyFinder finder = new FeaturePropertyFinder("NAME", new File(mapFile));
      display.addLocationListener(finder);
      MapToolTipper tipper = new MapToolTipper(dataFrame, finder, display.getCanvas());
      display.addInputEventListener(tipper);
      display.addMouseMotionListener(tipper);
      widget.addSpinnerListener(new TimeStepListener(widget, tipper));
      
      MapMenuMaker menuMaker = new MapMenuMaker(display.getCanvas(), dataFrame, finder, gui);
      display.addMouseListener(menuMaker);
      
      gui.visualizationAdded(dockableId, display);
      
      panel.add(display.getPanel(), BorderLayout.CENTER);
      panel.validate();
      
    } catch (InterruptedException e) {
      center.error("Error while creating map plot", e);
    } catch (ExecutionException e) {
      center.error("Error while creating map plot", e);
    } catch (PlotCreationException e) {
      center.error("Error while creating map plot", e);
    } finally {
     ((ProgressLayer)panel.getClientProperty(ProgressLayer.CLIENT_KEY)).setEnabled(true);
    }
  }

  private Style createPolygonStyle(ColorMap map, DataFrame frame) throws CQLException {
    StyleBuilder builder = new StyleBuilder();
    TextSymbolizer label = builder.createTextSymbolizer(Color.BLACK, builder
	.createFont("Arial", 10), "NAME");

    DataFrameIndex index = frame.getIndex();
    DataFrameAxis axis = frame.getAxes().getAxis("County");
    List<Rule> rules = new ArrayList<Rule>();
    for (int i = 0; i < axis.getExtent(); i++) {
      String county = axis.getValueAsString(i);
      index.set(0, i);
      int val = frame.getInt(index);

      Rule rule = builder.getStyleFactory().createRule();
      rule.setSymbolizers(new Symbolizer[] {
	  builder.createPolygonSymbolizer(map.getColorForValue(val), Color.BLACK, 1.0), label });
      Filter filter = CQL.toFilter(String.format("NAME = '%s'", county));
      rule.setFilter(filter);
      rules.add(rule);
    }

    Rule rule2 = builder.getStyleFactory().createRule();
    Symbolizer poly = builder.createPolygonSymbolizer(Color.GRAY, Color.BLACK, 1.0);
    rule2.setSymbolizers(new Symbolizer[] { poly, label });
    rule2.setIsElseFilter(true);
    rules.add(rule2);

    FeatureTypeStyle fts = builder.createFeatureTypeStyle("tl_2008_55_county", rules
	.toArray(new Rule[rules.size()]));
    Style style = builder.createStyle();
    style.addFeatureTypeStyle(fts);
    return style;
  }
  
  private static class TimeStepListener implements ChangeListener {
    
    private TimeStepWidget widget;
    private MapToolTipper tipper;
    /**
     * @param widget
     * @param tipper
     */
    public TimeStepListener(TimeStepWidget widget, MapToolTipper tipper) {
      this.widget = widget;
      this.tipper = tipper;
    }
    
    /* (non-Javadoc)
     * @see javax.swing.event.ChangeListener#stateChanged(javax.swing.event.ChangeEvent)
     */
    @Override
    public void stateChanged(ChangeEvent e) {
      int tsIndex = widget.getTime();
      tipper.setTimeStepIndex(tsIndex);
    }
  }
}
