package anl.aida.plot;

import java.util.ArrayList;
import java.util.List;

import org.jfree.data.DomainOrder;
import org.jfree.data.general.AbstractDataset;
import org.jfree.data.general.DatasetChangeEvent;
import org.jfree.data.xy.XYDataset;

import anl.aida.data.DataFrame;
import anl.aida.data.DataFrameIndex;

/**
 * Dataset appropriate for a scatter plot. Each series has two frames. The
 * xValue is the value from one frame and the yValue from another. Each frame
 * provides a single point per group / category axis value in that frame.
 * 
 * @version $Revision$ $Date$
 */
@SuppressWarnings("serial")
public class ScatterGroupDataset extends AbstractDataset implements XYDataset {

  private static class SeriesData {
    String name;
    int group;

    public SeriesData(String name, int group) {
      this.name = name;
      this.group = group;
    }
  }

  private DataFrame xFrame, yFrame;
  private List<SeriesData> series = new ArrayList<SeriesData>();
  private int timestep = 0;
  
  /**
   * Creates a ScatterGroupDataset that will plot the equivalent groups in each frame against 
   * each other.
   * 
   * @param xFrame
   * @param yFrame
   */
  public ScatterGroupDataset(DataFrame xFrame, DataFrame yFrame) {
    this.xFrame = xFrame;
    this.yFrame = yFrame;
  }

  public void addSeries(int group) {
    this.xFrame = xFrame;
    this.yFrame = yFrame;
    String name = xFrame.getAxes().getNonTimeAxis().getValueAsString(group);
    SeriesData seriesData = new SeriesData(name, group);
    series.add(seriesData);
    notifyListeners(new DatasetChangeEvent(this, this));
  }

  /**
   * Returns the order of the domain (or X) values returned by the dataset.
   * 
   * @return The order (never <code>null</code>).
   */
  public DomainOrder getDomainOrder() {
    return DomainOrder.NONE;
  }

  /**
   * Returns the number of items in a series. <br>
   * <br>
   * It is recommended that classes that implement this method should throw an
   * <code>IllegalArgumentException</code> if the <code>series</code> argument
   * is outside the specified range.
   * 
   * @param series
   *          the series index (in the range <code>0</code> to
   *          <code>getSeriesCount() - 1</code>).
   * @return The item count.
   */
  public int getItemCount(int series) {
    // always 1 for now
    return 1;
  }
  
  private double getValue(DataFrame frame, int group) {
    DataFrameIndex index = frame.getIndex();
    index.set(timestep, group);
    return frame.getDouble(index);
  }

  /**
   * Returns the x-value for an item within a series. The x-values may or may
   * not be returned in ascending order, that is up to the class implementing
   * the interface.
   * 
   * @param series
   *          the series index (in the range <code>0</code> to
   *          <code>getSeriesCount() - 1</code>).
   * @param item
   *          the item index (in the range <code>0</code> to
   *          <code>getItemCount(series)</code>).
   * @return The x-value (never <code>null</code>).
   */
  public Number getX(int series, int item) {
    return new Double(getXValue(series, item));
  }

  /**
   * Returns the x-value for an item within a series.
   * 
   * @param series
   *          the series index (in the range <code>0</code> to
   *          <code>getSeriesCount() - 1</code>).
   * @param item
   *          the item index (in the range <code>0</code> to
   *          <code>getItemCount(series)</code>).
   * @return The x-value.
   */
  public double getXValue(int series, int item) {
    SeriesData data = this.series.get(series);
    // only a single item so we can ignore that.
    return getValue(xFrame, data.group);
  }

  /**
   * Returns the y-value for an item within a series.
   * 
   * @param series
   *          the series index (in the range <code>0</code> to
   *          <code>getSeriesCount() - 1</code>).
   * @param item
   *          the item index (in the range <code>0</code> to
   *          <code>getItemCount(series)</code>).
   * @return The y-value (possibly <code>null</code>).
   */
  public Number getY(int series, int item) {
    return new Double(getYValue(series, item));
  }

  /**
   * Returns the y-value (as a double primitive) for an item within a series.
   * 
   * @param series
   *          the series index (in the range <code>0</code> to
   *          <code>getSeriesCount() - 1</code>).
   * @param item
   *          the item index (in the range <code>0</code> to
   *          <code>getItemCount(series)</code>).
   * @return The y-value.
   */
  public double getYValue(int series, int item) {
    SeriesData data = this.series.get(series);
    // only a single item so we can ignore that.
    return getValue(yFrame, data.group);
  }

  /**
   * Returns the number of series in the dataset.
   * 
   * @return The series count.
   */
  public int getSeriesCount() {
    return series.size();
  }
  
  /**
   * Gets the current timestep of this dataset.
   * 
   * @return the current timestep of this dataset.
   */
  public int getTimeStep() {
    return timestep;
  }
  
  /**
   * Sets the current timestep of this dataset.
   * 
   * @param timestep the new timestep
   */
  public void setTimeStep(int timestep) {
    this.timestep = timestep;
    notifyListeners(new DatasetChangeEvent(this, this));
  }

  /**
   * Returns the key for a series.
   * 
   * @param series
   *          the series index (in the range <code>0</code> to
   *          <code>getSeriesCount() - 1</code>).
   * @return The key for the series.
   */
  public Comparable getSeriesKey(int series) {
    if ((series < 0) || (series >= getSeriesCount())) {
      throw new IllegalArgumentException("Series index out of bounds");
    }
    SeriesData data = this.series.get(series);
    return data.name;
  }

  public void setSeriesKey(int series, String key) {
    if ((series < 0) || (series >= getSeriesCount())) {
      throw new IllegalArgumentException("Series index out of bounds");
    }
    SeriesData data = this.series.get(series);
    data.name = key;
    fireDatasetChanged();
  }

  /**
   * Returns the index of the series with the specified key, or -1 if there is
   * no such series in the dataset.
   * 
   * @param seriesKey
   *          the series key (<code>null</code> permitted).
   * @return The index, or -1.
   */
  public int indexOf(Comparable seriesKey) {
    for (int i = 0; i < series.size(); i++) {
      SeriesData data = series.get(i);
      if (data.name.equals(seriesKey))
	return i;
    }
    return -1;
  }
}
