package anl.aida.viz;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Paint;
import java.awt.event.ActionEvent;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.swing.AbstractAction;
import javax.swing.ButtonGroup;
import javax.swing.JCheckBox;
import javax.swing.JToolBar;

import anl.aida.client.doc.DocumentContainer;
import anl.aida.client.net.CGEdge;
import anl.aida.client.net.ConceptGraph;
import anl.aida.client.net.TermNode;
import anl.aida.client.widget.GraphSwapable;
import anl.aida.core.DocumentID;
import anl.aida.util.DateRange;
import anl.aida.viz.layout.Layout2D;
import edu.uci.ics.jung.graph.util.Pair;
import edu.umd.cs.piccolo.PLayer;
import edu.umd.cs.piccolo.PNode;
import edu.umd.cs.piccolo.nodes.PPath;
import edu.umd.cs.piccolo.util.PBounds;
import edu.umd.cs.piccolo.util.PNodeFilter;
import edu.umd.cs.piccolox.util.PBoundsLocator;

/**
 * 
 *         Date: Feb 11, 2009 11:43:03 AM
 */
public class NetworkDisplay implements AIDA2DLayer, GraphSwapable {

  private static final String OBJ = "OBJ";
  public static final String EDGES = "EDGES";
  public static final String NODES = "NODES";

  private Layout2D<TermNode> layout;
  private ConceptGraph<TermNode> graph;

  private PLayer nodeLayer, edgeLayer;
  private Paint nodePaint; 
  private boolean hideEmptyNodes = true;

  protected SelectionManager<DocumentContainer> selection;
  protected ItemSelectionHandler nodeSelectionHandler, edgeSelectionHandler;
  protected Canvas2D canvas;

  private NodeStyler styler;
  private EdgeStyler edgeStyler;
  private List<PPath> hiddenNodes = new ArrayList<PPath>();
  private boolean labelsVisible = true;


  public NetworkDisplay(ConceptGraph<TermNode> graph, Layout2D<TermNode> layout) {
    this.graph = graph;
    this.layout = layout;
    
    nodePaint = new Color(Colors.NODE_COLOR.getRed(), Colors.NODE_COLOR.getGreen(), Colors.NODE_COLOR.getBlue(),
	Colors.NODE_ALPHA);

    nodeLayer = new PLayer();
    edgeLayer = new PLayer();
    edgeLayer.setChildrenPickable(true);

    NodeDragSelectionHandler nodeSel = new NodeDragSelectionHandler(nodePaint, nodeLayer, edgeLayer, this.layout);
    nodeLayer.addInputEventListener(nodeSel);
    EdgeSelectionHandler edgeSel = new EdgeSelectionHandler(nodePaint, nodeLayer, edgeLayer);
    edgeLayer.addInputEventListener(edgeSel);

    selection = new SelectionManager<DocumentContainer>(nodePaint, Colors.SELECTED_COLOR);
    nodeSelectionHandler = new ItemSelectionHandler(this);
    nodeSel.addSelectionManager(selection);
    edgeSelectionHandler = new ItemSelectionHandler(this);
    edgeSel.addSelectionManager(selection);

    styler = new DefaultNodeStyler(nodePaint, Colors.TEXT_COLOR, graph);
    edgeStyler = new GrayScaleEdgeStyler(graph);
  }

  public static Object getObject(PNode node) {
    return node.getAttribute(OBJ);
  }

  public void initialize() {
    layout.initialize();
    reset();
  }

  
  private void create() {
    Map<TermNode, PPath> nodeMap = new HashMap<TermNode, PPath>();
    
    List<PPath> pNodes = new ArrayList<PPath>();

    for (TermNode tNode : graph.getVertices()) {
      Point2D coord = layout.getLocation(tNode);
      PPath node = styler.createNode(tNode, coord);
      if(tNode.getDocumentCount() == 0 && hideEmptyNodes) {
	node.setVisible(false);
	hiddenNodes.add(node);
      }
      NodeUtils.setObject(node, tNode);
      NodeUtils.setAsNode(node);
      node.addAttribute(EDGES, new ArrayList<PPath>());
      pNodes.add(node);
      //nodeLayer.addChild(node);
      nodeMap.put(tNode, node);
    }
    
    // sort so that larger nodes are drawn first
    Collections.sort(pNodes, new Comparator<PPath>() {
      @Override
      public int compare(PPath o1, PPath o2) {
	double area1 = o1.getBoundsReference().width * o1.getBoundsReference().height;
	double area2 = o2.getBoundsReference().width * o2.getBoundsReference().height;
	
	return area1 < area2 ? 1 : area1 == area2 ? 0 : -1;
      }
    });
    for (PPath path : pNodes) {
      nodeLayer.addChild(path);
    }

    for (CGEdge tEdge : graph.getEdges()) {
      Pair<TermNode> nodes = graph.getEndpoints(tEdge);
      PPath node1 = nodeMap.get(nodes.getFirst());
      PPath node2 = nodeMap.get(nodes.getSecond());

      Point2D.Double bound1 = (Point2D.Double) node1.getBounds().getCenter2D();
      Point2D.Double bound2 = (Point2D.Double) node2.getBounds().getCenter2D();
      
      PPath edge = edgeStyler.createEdge(tEdge, node1.localToGlobal(bound1), node2.localToGlobal(bound2));
      NodeUtils.setObject(edge, tEdge);
      NodeUtils.setAsEdge(edge);

      java.util.List<PPath> list = new ArrayList<PPath>();
      list.add(node1);
      list.add(node2);
      edge.addAttribute(NODES, list);

      //edge.setStroke(new BasicStroke(2 * tEdge.getDocumentCount() / avg));

      list = (java.util.List<PPath>) node1.getAttribute(EDGES);
      list.add(edge);
      list = (java.util.List<PPath>) node2.getAttribute(EDGES);
      list.add(edge);

      edgeLayer.addChild(edge);
    }
  }

  public void reset() {
    if (nodeLayer.getChildrenCount() == 0) {
      create();
    } else {
      for (int i = 0, n = nodeLayer.getChildrenCount(); i < n; i++) {
        PNode node = nodeLayer.getChild(i);
        TermNode tNode = (TermNode) NodeUtils.getObject(node);
        Point2D coord = layout.getLocation(tNode);
        node.setX(coord.getX());
        node.setY(coord.getY());

        PNode child = node.getChild(0);
        PBoundsLocator loc = PBoundsLocator.createNorthEastLocator(node);
        child.setX(loc.locateX());
        child.setY(loc.locateY());
      }

      for (int i = 0, n = edgeLayer.getChildrenCount(); i < n; i++) {
        PPath edge = (PPath) edgeLayer.getChild(i);
        java.util.List<PPath> nodes = (java.util.List<PPath>) edge.getAttribute(NetworkDisplay.NODES);
        PNode node1 = nodes.get(0);
        PNode node2 = nodes.get(1);

        edge.reset();

        // Note that the node's "FullBounds" must be used (instead of just the "Bound")
        // because the nodes have non-identity transforms which must be included when
        // determining their position.
        PNode c1 = node1.removeChild(0);
        PNode c2 = node2.removeChild(0);

        Point2D.Double bound1 = (Point2D.Double) node1.getFullBounds().getCenter2D();
        Point2D.Double bound2 = (Point2D.Double) node2.getFullBounds().getCenter2D();

        node1.addChild(c1);
        node2.addChild(c2);


        edge.moveTo((float) bound1.getX(), (float) bound1.getY());
        edge.lineTo((float) bound2.getX(), (float) bound2.getY());
      }
    }

    nodeLayer.repaint();
    edgeLayer.repaint();
  }

  public void swapGraph(ConceptGraph<TermNode> graph, DateRange range) {
    // swap the graph in the layout
    selection.unselectAll();
    canvas.clearSelection();
    layout.swapGraph(graph);
    this.graph = graph;
    edgeStyler.update(graph.getEdges());
   
    // on the first swap we want to change the styler
    // to use the original sizes as the base against which
    // to scale.
    //if (!(styler instanceof ScaledNodeStyler)) styler = new ScaledNodeStyler(nodePaint, TEXT_COLOR, nodeLayer);

    hiddenNodes.clear();
    edgeLayer.removeAllChildren();
    nodeLayer.removeAllChildren();
    create();
    if (!labelsVisible) doLabelsVisible(labelsVisible);
    nodeLayer.repaint();
    edgeLayer.repaint();
  }
  
  private void doLabelsVisible(boolean visible) {
    for (int i = 0; i < nodeLayer.getChildrenCount(); i++) {
	nodeLayer.getChild(i).getChild(0).setVisible(visible);
    }
    nodeLayer.repaint();
  }
  
  public void setLabelsVisible(boolean visible) {
    if (labelsVisible != visible) {
      doLabelsVisible(visible);
      labelsVisible = visible;
    }
  
  }
  
  public boolean isLabelsVisible() {
    return labelsVisible;
  }

  @SuppressWarnings("unchecked")
  public void lensOver(PBounds bounds, double radius, Canvas2D canvas) {
    /*
    if (lensListener != null) {
      List nodes = new ArrayList();
      PNodeFilter filter = new BoundsNodeEdgeFilter(bounds, radius);
      this.getAllNodes(filter, nodes);

      List<T> items = new ArrayList<T>();
      //StringBuilder buf = new StringBuilder();
      for (Object node : nodes) {
        PNode pnode = (PNode) node;
        Object type = pnode.getAttribute(TYPE_LABEL);
        if (type != null && type.equals("NODE")) {
          T obj = (T) pnode.getAttribute(OBJ);
          items.add(obj);
        }
      }
      if (items.size() > 0) lensListener.selectionOccurred(items, canvas);
    }
    */
  }

  /**
   * Called when selection mode is enabled or disabled.
   *
   * @param enabled state of selection mode.
   */
  public void selectionEnabled(boolean enabled) {
    if (enabled) {
      nodeLayer.addInputEventListener(nodeSelectionHandler);
      edgeLayer.addInputEventListener(edgeSelectionHandler);
    } else {
      selection.unselectAll();
      nodeLayer.removeInputEventListener(nodeSelectionHandler);
      edgeLayer.removeInputEventListener(edgeSelectionHandler);
    }
  }

  /**
   * Called when an individual node has been selected.
   *
   * @param node     the selected node
   * @param ctrlDown whether or not the the ctrl key was down
   *                 when the node was selected
   */
  public void nodeSelected(PNode node, boolean ctrlDown) {
    if (NodeUtils.isNode(node) || NodeUtils.isEdge(node)) {
      if (!ctrlDown) selection.unselectAll();
      selection.addSelectedItem((PPath) node);
    }

    addSelectionsToCanvas();
  }

  public void marqueeOver(PBounds bounds, boolean ctrlDown) {
    if (bounds == null && !ctrlDown) {
      selection.unselectAll();
      canvas.clearSelection();
    } else {
      if (!ctrlDown)
	selection.unselectAll();
      java.util.List nodes = new ArrayList();
      PNodeFilter filter = new BoundsNodeEdgeFilter(bounds, 0);
      edgeLayer.getAllNodes(filter, nodes);
      nodeLayer.getAllNodes(filter, nodes);
      for (Object node : nodes) {
	selection.addSelectedItem((PPath) node);
      }
      addSelectionsToCanvas();
    }
  }

  private void addSelectionsToCanvas() {
    canvas.clearSelection();
    Set<DocumentID> ids = new HashSet<DocumentID>();
    for (DocumentContainer docs : selection.getSelectedObjects()) {
      for (DocumentID id : docs.documents()) {
        ids.add(id);
      }
    }

    for (DocumentID id : ids) {
      canvas.addSelection(id);
    }

  }

  public void addToCanvas(Canvas2D canvas) {
    canvas.getRoot().addChild(nodeLayer);
    canvas.getRoot().addChild(0, edgeLayer);

    canvas.getCamera().addLayer(nodeLayer);
    canvas.getCamera().addLayer(0, edgeLayer);
    this.canvas = canvas;
  }

  public void initToolBar(JToolBar bar, ButtonGroup grp) {
    JCheckBox checkBox = new JCheckBox(new AbstractAction("Hide Empty Nodes") {
      @Override
      public void actionPerformed(ActionEvent e) {
	hideEmptyNodes(((JCheckBox)e.getSource()).isSelected());
      }
    });
    checkBox.setSelected(hideEmptyNodes);
    bar.addSeparator();
    bar.add(checkBox);
  }
  
  private void hideEmptyNodes(boolean hide) {
    if (hide != hideEmptyNodes) {
      hideEmptyNodes = hide;
      for (PPath node : hiddenNodes) {
	node.setVisible(!hideEmptyNodes);
      }
    }
  }

  public void resized(Dimension size) {
    if (nodeLayer.getChildrenCount() == 0) {
      layout.initialize(size);
      reset();
    } else {
      layout.sizeChanged(size);
      reset();
    }
  }

  private class BoundsNodeEdgeFilter implements PNodeFilter {

    PBounds bounds;
    PBounds localBounds = new PBounds();
    double radius2 = 0;

    private BoundsNodeEdgeFilter(PBounds bounds, double radius) {
      this.bounds = bounds;
      this.radius2 = radius * radius;
    }

    public boolean accept(PNode pNode) {
      //localBounds.setRect(bounds);
      //pNode.globalToLocal(localBounds);
      //System.out.println("pNode.getBounds() = " + pNode.getBounds());
      //System.out.println("bounds = " + localBounds);
      boolean intersect = pNode.intersects(bounds) && (NodeUtils.isEdge(pNode) ||
              NodeUtils.isNode(pNode));
      if (radius2 == 0) return intersect;
      return radius2 >= bounds.getCenter2D().distanceSq(pNode.getBounds().getCenter2D());


    }

    public boolean acceptChildrenOf(PNode pNode) {
      return true;
    }
  }
}