package anl.aida.viz;

import anl.aida.algo.Clusters;
import anl.aida.viz.layout.CircleLayout;

import java.awt.*;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Lays out items depending on their similarity to N number
 * of centroids.
 *
 */
public class SimilarityClusterLayout<T> {

  private Point2D[] clusterPts;

  private static class PointScore {
    Point2D pt;
    float score;

    private PointScore(Point2D pt, float score) {
      this.pt = pt;
      this.score = score;
    }
  }

  private static class Points {

    private float min = Float.POSITIVE_INFINITY;
    private List<PointScore> pts = new ArrayList<PointScore>();

    public void addPoint(Point2D pt, float score) {
      pts.add(new PointScore(pt, score));
      if (score < min) min = score;
    }

    private Point2D avg() {
      double x = 0, y = 0;
      int total = pts.size();

      for (PointScore ps : pts) {
        double ptx = ps.pt.getX();
        double pty = ps.pt.getY();
        x += ptx;
        y += pty;

        // add the more times to exacerabate
        // difference with lowest
        int weight = (int) Math.ceil(ps.score / min);
        if (weight > 0) {
          x += ptx * weight;
          y += pty * weight;
          total += weight;
        }
      }

      return new Point2D.Double(x / total, y / total);
    }
  }

  private Clusters<T> clusters;


  private Map<T, Point2D> locations = new HashMap<T, Point2D>();
  private CircleLayout<Integer> centroidLayout;


  public SimilarityClusterLayout(Clusters<T> clusters, Dimension size) {
    this.clusters = clusters;
    java.util.List<Integer> centroids = new ArrayList<Integer>();
    for (int i = 0; i < clusters.size(); i++) {
      centroids.add(i);
    }
    centroidLayout = new CircleLayout<Integer>(centroids, size);
  }

  public void resetSize(Dimension size) {
    centroidLayout.sizeChanged(size);
  }


  public void initialize() {
    centroidLayout.initialize();
    clusterPts = new Point2D[clusters.size()];

    for (int i = 0, n = clusters.size(); i < n; i++) {
      clusterPts[i] = centroidLayout.getLocation(i);
    }
    reset();
  }

  public void reset() {
    Point2D center = getClusterCenter();

    Map<T, Points> pointMap = new HashMap<T, Points>();
    for (int i = 0; i < clusterPts.length; i++) {
      Point2D clusterPt = clusterPts[i];
      Point2D diff = new Point2D.Double();
      diff(clusterPt, center, diff);
      for (T item : clusters.items()) {
        Points pts = getPointsFor(item, pointMap);
        float score = clusters.getScore(item, i);
        //if (score > 0) {
        // todo some scores are < 0, not sure why this should ever be the case ...
        double x = diff.getX() * score;
        double y = diff.getY() * score;
        Point2D pt = new Point2D.Double(x + center.getX(), y + center.getY());
        pts.addPoint(pt, score);
        //}
      }
    }


    locations.clear();
    for (Map.Entry<T, Points> entries : pointMap.entrySet()) {
      Point2D avg = entries.getValue().avg();
      locations.put(entries.getKey(), avg);
    }
  }

  private Points getPointsFor(T obj, Map<T, Points> map) {
    Points pts = map.get(obj);
    if (pts == null) {
      pts = new Points();
      map.put(obj, pts);
    }

    return pts;
  }

  public Point2D getLocation(T obj) {
    return locations.get(obj);
  }

  public Point2D getCentroidLocation(int index) {
    return clusterPts[index];
  }

  public void setCentroidLocation(int index, Point2D loc) {
    clusterPts[index] = loc;
  }

  public Point2D getClusterCenter() {
    double x = 0, y = 0;
    for (Point2D point : clusterPts) {
      x += point.getX();
      y += point.getY();
    }

    return new Point2D.Double(x / clusterPts.length, y / clusterPts.length);
  }


  private void diff(Point2D pt1, Point2D pt2, Point2D result) {
    result.setLocation(pt1.getX() - pt2.getX(), pt1.getY() - pt2.getY());
  }
}
