package anl.aida.viz3d;

import javax.media.opengl.GL;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

/**
 * An item in a StackDisplay.
 * 
 */
public class StackItem2 extends VSpatial implements Comparable<StackItem2> {
  
  private static float MAX_X = .9f;
  private static float MAX_Z = .7f;
  private static float MAX_ROT = 90f;
  
  private float rotation, x, z;
  private float xOff, zOff;
  private StackDrawable2 drawable;
  private long index;
  
  /**
   * @param item
   */
  public StackItem2(StackDrawable2 drawable, long index) {
    this.drawable = drawable;
    this.index = index;
    this.drawable.setParent(this);
    
    id = "Stack Item";
  }
  
  /**
   * Notifies this StackItem of its position in the stack prior to 
   * drawing. If the index is -1, then this StackItem is off the stack.
   * Index 0 is the top of the stack.
   * 
   * @param index the index of this StackItem in the stack 
   * @param stackSize the current size of the stack
   */
  public void stackPosition(int index, int stackSize) {
    drawable.stackPosition(index, stackSize);
    
  }

  /**
   * Gets an index for this StackItem that can be used
   * to sort this item with respect to other items.
   * 
   * @return an sorting index for this StackItem
   */
  
  public long getSortingIndex() {
    return index;
  }

  /**
   * Puts this StackItem completely off the stack.
   */
  public void putOffStack() {
    this.x = MAX_X;
    this.rotation = MAX_ROT;
    this.z = MAX_Z;
    markAsDirty();
  }
  
  /**
   * Puts this StackItem completely on the stack.
   */
  public void putOnStack() {
    rotation = x = z = 0;
    markAsDirty();
  }

  /**
   * Gets whether or not this slice is currently moving. A moving
   * StackItem is neither fully on or off the stack.
   * 
   * @return whether or not this slice is currently moving.
   */
  public boolean isMoving() {
    return rotation > .001f && rotation < 90;
  }

  /**
   * Moves this StackItem off the stack.
   * 
   * @return true if this item is completely off the stack
   * and will move no further, otherwise false.
   */
  public synchronized boolean moveOff() {
    if (rotation < 90) {
      rotation += 5f;
      if (rotation > 20f)
	z += .05f;
      x += .05f;
    }
    markAsDirty();

    return rotation >= 90;
  }
  
  /**
   * Sets the offsets for this StackItem. 
   * 
   * @param xOff the x offset
   * @param zOff the y offset
   */
  public void setOffsets(float xOff, float zOff) {
    this.xOff = xOff;
    this.zOff = zOff;
    markAsDirty();
  }

  /**
   * Moves this StackItem onto the stack.
   * 
   * @return true if this item is completely on the stack
   * and will move no further, otherwise false.
   */
  public synchronized boolean moveOn() {
    if (rotation > 0) {
      if (rotation > 20f)
	z -= .05f;
      rotation -= 5f;
      x -= 0.05f;
    }
    markAsDirty();

    return rotation <= 0.001;
  }
  
  public String toString() {
    return String.format("StackItem: x: %f, z: %f, rotation: %f", x, z, rotation);
  }
  
  /**
   * Draws this StackItem.
   * 
   * @param gl
   * @param zOff the z offset for this StackItem
   * @param xOff the x offset for this StackItem
   */
  public void draw(GL gl) {
    drawable.translate(-x + xOff, 0f, z + zOff);
    drawable.rotate(rotation, 0f, .5f, 0f);
    super.draw(gl);
  }
  
  public void updateWorldTransformation() {
    super.updateWorldTransformation();
    drawable.updateWorldTransformation();
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.mifs.viz3d.AbstractVNode#doDraw(javax.media.opengl.GL)
   */
  @Override
  protected void doDraw(GL gl) {
    if (visible) {
      drawable.draw(gl);
    }
  }
  
  /* (non-Javadoc)
   * @see anl.mifs.viz3d.VSpatial#markAsDirty()
   */
  @Override
  protected void markAsDirty() {
    super.markAsDirty();
    drawable.markAsDirty();
  }

  /* (non-Javadoc)
   * @see anl.mifs.viz3d.VSpatial#getBoundingSphere()
   */
  @Override
  public BoundingSphere getBoundingSphere() {
    return drawable.getBoundingSphere();
  }
  
  protected BoundingSphere doGetBoundingSphere() {
    return drawable.doGetBoundingSphere();
  }

  /* (non-Javadoc)
   * @see anl.mifs.viz3d.VSpatial#getLocalBoundingSphere()
   */
  @Override
  public BoundingSphere getLocalBoundingSphere() {
    return drawable.getBoundingSphere();
  }

  /* (non-Javadoc)
   * @see anl.mifs.viz3d.VSpatial#intersects(javax.vecmath.Point3f, javax.vecmath.Vector3f, anl.mifs.viz3d.Accumulator)
   */
  @Override
  public void intersects(Point3f rayOrigin, Vector3f rayDirection, Accumulator accumulator) {
    drawable.intersects(rayOrigin, rayDirection, accumulator);
  }

  public int compareTo(StackItem2 o) {
    long i1 = getSortingIndex();
    long i2 = o.getSortingIndex();
    
    return i1 < i2 ? -1 : i1 == i2 ? 0 : 1;
  }
}
