/**
 * 
 */
package anl.aida.viz3d;

import java.util.ArrayList;
import java.util.List;

import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

import anl.aida.viz3d.StackSorter.SortOrder;

/**
 * Renders items in a stack.
 * 
 */
public class StackRenderer2 extends VSpatial {
  
  private static float Z_INTERVAL = .15f;
  private static float X_INTERVAL = .1f;
  
  private List<StackItem2> items = new ArrayList<StackItem2>();
  private int stackIndex = 0;
  private boolean atTop, atBottom;
  private GLAutoDrawable drawable;
  private boolean initialized = false;
  private StackItemTextFormatter itemFormatter;
  private StackSorter sorter = new StackSorter();
  
  private Axis axis = new Axis(Z_INTERVAL);
  
  public StackRenderer2(GLAutoDrawable drawable) {
    this(drawable, new StackItemDateFormatter());
    
  }
  
  public StackRenderer2(GLAutoDrawable drawable, StackItemTextFormatter itemFormatter) {
    this.drawable = drawable;
    this.itemFormatter = itemFormatter;
    sorter.setSortOrder(SortOrder.DESCENDING);
  }
  
  /**
   * Gets the drawable used by this StackRenderer.
   * 
   * @return the drawable used by this StackRenderer.
   */
  public GLAutoDrawable getDrawable() {
    return drawable;
  }

  /**
   * Adds a StackItem to this StackRenderer.
   * 
   * @param item
   *          the item to add
   */
  public void addStackItem(StackItem2 item) {
    items.add(item);
    item.setParent(this);
    sorter.sort(items);
    if (initialized) {
      int index = items.indexOf(item);
      resetItems();
      jumpTo(index, false);
    }
  }
  
  private void resetItems() {
    for (StackItem2 item : items) {
      item.putOnStack();
    }
    stackIndex = 0;
  }
  
  /* (non-Javadoc)
   * @see anl.mifs.viz3d.VSpatial#markAsDirty()
   */
  @Override
  protected void markAsDirty() {
    super.markAsDirty();
    for (StackItem2 item : items) {
      item.markAsDirty();
    }
  }
  
  /*
   * (non-Javadoc)
   * 
   * @see anl.mifs.viz3d.VSpatial#updateWorldTransformation()
   */
  @Override
  public void updateWorldTransformation() {
    super.updateWorldTransformation();
    for (StackItem2 item : items) {
      item.updateWorldTransformation();
    }
  }
  
  /**
   * Pops the current stack item completely off the stack and make the current
   * stack item the next one.
   */
  public void next() {
    if (stackIndex < items.size() - 1) {
      items.get(stackIndex).putOffStack();
      stackIndex++;
    }
  }

  /**
   * If the current stack item is current moving onto / off the stack, then this
   * pushes it completely onto the stack. If it is not moving, then the previous
   * item is pushed onto the stack.
   */
  public void prev() {
    if (stackIndex > 0) {
      boolean isMoving = items.get(stackIndex).isMoving();
      items.get(stackIndex).putOnStack();
      if (!isMoving) {
	stackIndex--;
	items.get(stackIndex).putOnStack();
      }
    } else if (stackIndex == 0) {
      items.get(0).putOnStack();
    }
  }

  /**
   * Moves the current stack item incrementaly off the stack. If the movement
   * takes the item completely off the stack then this increments the stack
   * index.
   */
  public void moveOff() {
    if (!atBottom) {
      atTop = false;
      boolean doneMoving = items.get(stackIndex).moveOff();
      if (doneMoving) {
	atBottom = stackIndex == items.size() - 1;
	if (!atBottom)
	  stackIndex++;
      }
    }
  }

  /**
   * Moves the current stack item incrementaly on to the stack. If the movement
   * takes the item completely on to the stack then this decrements the stack
   * index.
   */
  public void moveOn() {
    if (!atTop) {
      atBottom = false;
      boolean doneMoving = items.get(stackIndex).moveOn();
      if (doneMoving) {
	atTop = stackIndex == 0;
	if (!atTop)
	  --stackIndex;
      }
    }
  }
  
  /**
   * Jumps the top of the stack to the StackItem
   * 
   * @param item the item to jump to
   * @param updateDisplay whether or not to update the display as the stack 
   * pops / pushes to the specified index
   */
  public void jumpTo(StackItem2 item, boolean updateDisplay) {
    int index = items.indexOf(item);
    if (index != -1) {
      jumpTo(index, updateDisplay);
    }
  }
  
  
  /**
   * Jumps the top of the stack to the specified index.
   * 
   * @param index the index to jump to.
   * @param updateDisplay whether or not to update the display as the stack 
   * pops / pushes to the specified index
   */
  public void jumpTo(int index, boolean updateDisplay) {
    if (index == stackIndex) {
      items.get(stackIndex).putOnStack();
      drawable.display();
    } else if (index >= 0 && index < items.size()) {

      if (index < stackIndex) {
	while (stackIndex != index) {
	  prev();
	  if (updateDisplay) drawable.display();
	}
      } else {
	while (stackIndex != index) {
	  next();
	  if (updateDisplay) drawable.display();
	}
      }
      drawable.display();
    }
  }

  
  /* (non-Javadoc)
   * @see anl.mifs.viz3d.VSpatial#doDraw(javax.media.opengl.GL)
   */
  @Override
  protected void doDraw(GL gl) {
    initialized = true;
    gl.glEnable(GL.GL_BLEND);
    gl.glBlendFunc(GL.GL_ONE, GL.GL_ONE_MINUS_SRC_ALPHA);
    
    // draws the off stack panels; draw in
    // z buffer order for proper transparency
    float xOff = -X_INTERVAL * (stackIndex - 1) - .5f;
    for (int i = 0; i < stackIndex; i++) {
      //System.out.printf("item %d at xoff %f%n", i, xOff);
      StackItem2 item = items.get(i);
      item.stackPosition(-1, -1);
      item.setOffsets(xOff, 0);
      item.draw(gl);
      xOff += X_INTERVAL;
    }

    // draws on stack panels; draw in
    // z buffer order for proper transparency
    float zOff = -(Z_INTERVAL * (items.size() - stackIndex));
    float zStart = zOff;
    axis.clearLabels();
    int stackSize = items.size() - stackIndex;
    for (int i = items.size() - 1; i >= stackIndex; i--) {
      StackItem2 item = items.get(i);
      item.stackPosition(i - stackIndex, stackSize);
      item.setOffsets(0, zOff);
      item.draw(gl);
      axis.addTickLabel(itemFormatter.format(item));
      zOff += Z_INTERVAL;
    }
    
    axis.draw(gl, new Point3f(-.51f, -.405f, 0), zStart, zOff - Z_INTERVAL);
    gl.glDisable(GL.GL_BLEND);
  }

  /* (non-Javadoc)
   * @see anl.mifs.viz3d.VNode#getBoundingBox()
   */
  @Override
  protected BoundingSphere doGetBoundingSphere() {
    BoundingSphere sphere = new BoundingSphere(items.get(0).doGetBoundingSphere());
    for (int i = 1, n = items.size(); i < n; i++) {
      sphere.merge(items.get(i).doGetBoundingSphere());
    }
    return sphere;
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.mifs.viz3d.VSpatial#intersects(javax.vecmath.Point3f,
   * javax.vecmath.Vector3f, anl.mifs.viz3d.Accumulator)
   */
  @Override
  public void intersects(Point3f rayOrigin, Vector3f rayDirection, Accumulator accumulator) {
    // ray is in world coordinates, so first check if the ray intersects with
    // the BoundingSphere if not, then do nothing, if it does then pass to children.
    BoundingSphere sphere = getBoundingSphere();
    if (sphere.intersects(rayOrigin, rayDirection)) {
      for (VSpatial child : items) {
	child.intersects(rayOrigin, rayDirection, accumulator);
      }
    }
  }
}
