/**
 * 
 */
package anl.aida.viz3d;

import java.awt.Color;
import java.awt.geom.Rectangle2D;
import java.nio.DoubleBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;

import javax.media.opengl.GL;
import javax.media.opengl.glu.GLU;
import javax.vecmath.Point3f;
import javax.vecmath.Quat4f;
import javax.vecmath.Tuple3f;
import javax.vecmath.Vector3f;

import com.sun.opengl.util.BufferUtil;

/**
 * Static utililty methods for 3D.
 * 
 *
 */
public class Utils3D {
  
  public enum Location {NORTH_EAST};
  
  private static final float K_EPSILON = 1.0e-6f;
  private static final float NE_SIN = (float)Math.sin(Math.toRadians(45));
  private static final float NE_COS = (float)Math.cos(Math.toRadians(45));
  
  public static void colorToFloats(float[] array, Color color) {
    array[0] = color.getRed() / 255f;
    array[1] = color.getGreen() / 255f;
    array[2] = color.getBlue() / 255f;
    array[3] = color.getAlpha() / 255f;
  }
  
  /**
   * Checks if equals zero within some epsilon.
   *
   * @param val
   * @return true if the val is close enough to zero otherwise false
   */
  public static boolean isZero(float val) {
    return Math.abs(val) < K_EPSILON;
  }
  
  public static Color floatToColor(float[] color) {
    return new Color(color[0], color[1], color[2], color[3]);
  }
  /**
   * Assumes quat is normalized. Rotates in tuple by the
   * rotation defined by the quaterion.
   * 
   * @param rotation
   * @param in
   * @param out
   * @return
   */
  public static Tuple3f mult(Quat4f rotation, Tuple3f in, Tuple3f out) {
    if (out == null) out = new Vector3f();
    float w = rotation.w;
    float x = rotation.x;
    float y = rotation.y;
    float z = rotation.z;
    
    float pMult = w*w - x*x - y*y - z*z;
    float vMult = 2.0f*(x*in.x + y*in.y + z*in.z);
    float crossMult = 2.0f*w;

    out.set( pMult*in.x + vMult*x + crossMult*(y*in.z - z*in.y),
                      pMult*in.y + vMult*y + crossMult*(z*in.x - x*in.z),
                      pMult*in.z + vMult*z + crossMult*(x*in.y - y*in.x) );
    return out;
  }
  
  /**
   * Gets the location relative to the specified VSpatial
   * in terms of the VSpatial's local frame.
   * 
   * @param item the VSpatial to get location relative to
   * @param location the type of location we want
   * @return the location relative to the specified VSpatial.
   */
  public static Point3f getRelativeLocation(VSpatial item, Location location) {
    
    BoundingSphere sphere = item.getLocalBoundingSphere();
    Point3f pt = new Point3f(sphere.getCenterRef());
    float radius = sphere.getRadius();
    if (location == Location.NORTH_EAST) {
      float x = radius * NE_SIN;
      float y = radius * NE_COS;
      pt.x += x;
      pt.y += y;
      
      return pt;
    }
   

    return null;
  }
  
  /**
   * Creates a bounding box that will encompass the vertices in the
   * specified buffer. 
   * 
   * @param buf the Buffer containing the vertices' coordinates
   * @param size the number of coordinates per vertex
   * @return
   */
  public static BoundingSphere createBoundingSphere(FloatBuffer buf) {
    // compute minimal and maximal bounds 
    float[] vertex = new float[3];
    buf.get(vertex);
    Point3f min = new Point3f(vertex);
    Point3f max = new Point3f(vertex);
    
    while (buf.hasRemaining()) {
      buf.get(vertex);
      float x = vertex[0];
      float y = vertex[1];
      float z = vertex[2];
      
      if (min.x > x) min.x = x;
      else if (max.x < x) max.x = x;
      
      if (min.y > y) min.y = y;
      else if (max.y < y) max.y = y;
      
      if (min.z > z) min.z = z;
      else if (max.z < z) max.z = z;
    }
    
    // center half-way between min and max
    Point3f center = new Point3f(min);
    center.add(max);
    center.scale(.5f);
    
    // compute radius
    buf.rewind();
    buf.get(vertex);
    Point3f pt = new Point3f(vertex);
    float maxDist = center.distanceSquared(pt);
    while (buf.hasRemaining()) {
      buf.get(vertex);
      pt.set(vertex);
      float dist = center.distanceSquared(pt);
      if (dist > maxDist) {
	maxDist = dist;
      }
    }
    
    buf.rewind();
    return new BoundingSphere(center, (float)Math.sqrt(maxDist));
  }
  
  /**
   * Computes whether or not the triangle defined by P0, P1, and P2
   * is intersected by the the ray.
   * 
   * @param p0
   * @param p1
   * @param p2
   * @param rayOrigin
   * @param rayDirection
   * 
   * @return true if the triangle is intersected by the ray, otherwise false.
   */
  public static boolean triangleIntersect(Point3f p0, Point3f p1, Point3f p2,
      Point3f rayOrigin, Vector3f rayDirection) {
    Vector3f e1 = new Vector3f();
    e1.sub(p1, p0);
    Vector3f e2 = new Vector3f();
    e2.sub(p2, p0);
    
    Vector3f p = new Vector3f();
    p.cross(rayDirection, e2);
    float a = e1.dot(p);
    
    // if result zero, no intersection or infinite intersections
    // (ray parallel to triangle plane)
    if (isZero(a)) return false;
    
    // compute denominator
    float f = 1.0f/a;
    
    // compute barycentric coordinates
    Vector3f s = new Vector3f();
    s.sub(rayOrigin, p0); 
    float u = f*s.dot(p);
    
    // ray falls outside triangle
    if (u < 0.0f || u > 1.0f) return false;

    
    Vector3f q = new Vector3f();
    q.cross(s, e1); 
    float v = f* rayDirection.dot(q);
    
    // ray falls outside triangle
    if (v < 0.0f || u+v > 1.0f) return false;

    // compute line parameter
    float t = f*e2.dot(q);
    return (t >= 0.0f);
  }
  
  /**
   * Computes the normal of the specified triangle.
   * 
   * @param pt1
   * @param pt2
   * @param pt3
   */
  public static void triangleNormal(Point3f pt1, Point3f pt2, Point3f pt3, Vector3f result) {
    result.set(pt2);
    result.sub(pt1);
    
    Vector3f v2 = new Vector3f(pt3);
    v2.sub(pt2);
    
    result.cross(result, v2);
    result.normalize();
  }
  
  /**
   * Creates a BoundingSphere that will enclose the rectangle.
   * The z coordinate is assumed to be 0. Center is the rect's
   * center and radius is the distance from center to rect corner.
   * 
   * @param rect the rectangle to bound
   * @return the created sphere
   */
  public static BoundingSphere createBoundingSphere(Rectangle2D rect) {
    Point3f center = new Point3f((float)rect.getCenterX(), (float)rect.getCenterY(), 0f);
    double xExtent = rect.getWidth();
    double yExtent = rect.getHeight();
    // distance from one corner to opposite divided by 2
    float radius = (float)(Math.sqrt(xExtent * xExtent + yExtent * yExtent) / 2f);
    return new BoundingSphere(center, radius);
  }
  
  public static Box getWorldSizeAtZ(GL gl, float nearClipZ, float farClipZ, float z) {
    
    IntBuffer viewport = BufferUtil.newIntBuffer(4);
    DoubleBuffer mvMatrix = BufferUtil.newDoubleBuffer(16);
    DoubleBuffer projMatrix = BufferUtil.newDoubleBuffer(16);
    DoubleBuffer output = BufferUtil.newDoubleBuffer(3);
    GLU glu = new GLU();
    
    gl.glGetIntegerv(GL.GL_VIEWPORT, viewport);
    gl.glGetDoublev(GL.GL_MODELVIEW_MATRIX, mvMatrix);
    gl.glGetDoublev(GL.GL_PROJECTION_MATRIX, projMatrix);
    
    //System.out.printf("cursor coords: %d, %d%n",x, ogY);
    glu.gluUnProject(0, 0, 0, mvMatrix, projMatrix, viewport, output);
    //System.out.printf("world coords at z = 1.0 are (%f, %f, %f)%n", output.get(0), output.get(1), output.get(2));
   
    Point3f nearLower = new Point3f();
    nearLower.x = (float)output.get(0);
    nearLower.y = (float)output.get(1);
    nearLower.z = (float)output.get(2);
    //System.out.println("nearLower = " + nearLower);
    
    glu.gluUnProject(viewport.get(2), viewport.get(3), 0, mvMatrix, projMatrix, viewport, output);
    
    
    Point3f nearUpper = new Point3f();
    nearUpper.x = (float)output.get(0);
    nearUpper.y = (float)output.get(1);
    nearUpper.z = (float)output.get(2);
    
    //System.out.println("nearUpper = " + nearUpper);
    
    glu.gluUnProject(0, 0, 1f, mvMatrix, projMatrix, viewport, output);
    //System.out.printf("world coords at z = 1.0 are (%f, %f, %f)%n", output.get(0), output.get(1), output.get(2));
   
    Point3f farLower = new Point3f();
    farLower.x = (float)output.get(0);
    farLower.y = (float)output.get(1);
    farLower.z = (float)output.get(2);
    //System.out.println("farLower = " + farLower);
    
    glu.gluUnProject(viewport.get(2), viewport.get(3), 1f, mvMatrix, projMatrix, viewport, output);
    
    Point3f farUpper = new Point3f();
    farUpper.x = (float)output.get(0);
    farUpper.y = (float)output.get(1);
    farUpper.z = (float)output.get(2);
    
    float interval = (nearClipZ - z) / Math.abs((farClipZ - nearClipZ));
    
    Vector3f dir = new Vector3f(farLower);
    dir.sub(nearLower);
    
    Point3f lower = new Point3f();
    lower.x = nearLower.x + (dir.x * interval);
    lower.y = nearLower.y + (dir.y * interval);
    lower.z = z;
    
    dir.set(farUpper);
    dir.sub(nearUpper);
    
    Point3f upper = new Point3f();
    upper.x = nearUpper.x + (dir.x * interval);
    upper.y = nearUpper.y + (dir.y * interval);
    upper.z = z;
    
    return new Box(lower, upper);
  }
}
