package anl.aida.viz3d;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.locks.ReentrantLock;

import javax.media.opengl.GL;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

/**
 * VSpatial with children.
 * 
 * 
 */
public class VNode extends VSpatial {

  protected List<VSpatial> children = new ArrayList<VSpatial>();
  private ReentrantLock lock = new ReentrantLock();

  public void addChild(VSpatial child) {
    lock.lock();
    try {
      if (child.getParent() != null) {
	((VNode)child.getParent()).removeChild(child);
      }
      child.setParent(this);
      children.add(child);
    } finally {
      lock.unlock();
    }
  }
  
  public Iterable<VSpatial> children() {
    return children;
  }
  
  public void addChildren(VSpatial... items) {
    lock.lock();
    try {
      for (VSpatial child : items) {
	if (child.getParent() != null) {
	  ((VNode) child.getParent()).removeChild(child);
	}
	child.setParent(this);
	this.children.add(child);
      }
    } finally {
      lock.unlock();
    }
  }
  
  public void removeChild(VSpatial child) {
    lock.lock();
    try {
      children.remove(child);
    } finally {
      lock.unlock();
    }
  }

  public int getChildCount() {
    return children.size();
  }

  public VSpatial get(int childIndex) {
    VSpatial child;
    lock.lock();
    try {
      child = children.get(childIndex);
    } finally {
      lock.unlock();
    }
    return child;
  }

  public void removeAllChildren() {
    lock.lock();
    try {
      for (VSpatial child : children) {
	child.setParent(null);
      }
      children.clear();
    } finally {
      lock.unlock();
    }
  }
  
  /* (non-Javadoc)
   * @see anl.mifs.viz3d.VSpatial#markAsDirty()
   */
  @Override
  protected void markAsDirty() {
    super.markAsDirty();
    lock.lock();
    try {
      for (VSpatial child : children) {
	child.markAsDirty();
      }
    } finally {
      lock.unlock();
    }
  }
  
  /*
   * (non-Javadoc)
   * 
   * @see anl.mifs.viz3d.VSpatial#updateWorldTransformation()
   */
  @Override
  public void updateWorldTransformation() {
    super.updateWorldTransformation();
    lock.lock();
    try {
      for (VSpatial child : children) {
	child.updateWorldTransformation();
      }
    } finally {
      lock.unlock();
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.mifs.viz3d.AbstractVNode#doDraw(javax.media.opengl.GL)
   */
  @Override
  protected void doDraw(GL gl) {
    if (visible) {
      lock.lock();
      try {
	for (VSpatial child : children) {
	  child.draw(gl);
	}
      } finally {
	lock.unlock();
      }
    }
  }

  /* (non-Javadoc)
   * @see anl.mifs.viz3d.VNode#getBoundingBox()
   */
  @Override
  protected BoundingSphere doGetBoundingSphere() {
    if (children.isEmpty()) return new BoundingSphere(new Point3f(), 0);
    BoundingSphere sphere = new BoundingSphere(children.get(0).doGetBoundingSphere());
    for (int i = 1, n = children.size(); i < n; i++) {
      sphere.merge(children.get(i).doGetBoundingSphere());
    }
    return sphere;
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.mifs.viz3d.VSpatial#intersects(javax.vecmath.Point3f,
   * javax.vecmath.Vector3f, anl.mifs.viz3d.Accumulator)
   */
  @Override
  public void intersects(Point3f rayOrigin, Vector3f rayDirection, Accumulator accumulator) {
    // ray is in world coordinates, so first check if the ray intersects with
    // the BoundingSphere if not, then do nothing, if it does then pass to children.
    BoundingSphere sphere = getBoundingSphere();
    if (sphere.intersects(rayOrigin, rayDirection)) {
      lock.lock();
      try {
	for (VSpatial child : children) {
	  child.intersects(rayOrigin, rayDirection, accumulator);
	}
      } finally {
	lock.unlock();
      }
    }
  }
}
