/**
 * 
 */
package anl.aida.viz3d;

import java.awt.Font;
import java.nio.FloatBuffer;
import java.util.HashMap;
import java.util.Map;

import javax.media.opengl.GLContext;
import javax.vecmath.Vector3f;

import com.sun.opengl.util.BufferUtil;
import com.sun.opengl.util.j2d.TextRenderer;

/**
 * Factory for generation VNodes for particular shapes.
 * 
 * 
 */
public class VShapeFactory {
  
  private static VShapeFactory instance;
  
  private Map<GLContext, TextRenderer> trMap = new HashMap<GLContext, TextRenderer>();
  // to force text renderer to use gl immediate mode
  // may help on some video cards
  private boolean forceImmMode = false;
  
  private VShapeFactory() {
    forceImmMode = Boolean.getBoolean("anl.aida.forceTRIM");
  }
  
  public void glInit() {
    for (TextRenderer renderer : trMap.values()) {
      renderer.dispose();
    }
    trMap.clear();
 
  }
  
  public TextRenderer getTextRenderer() {
    GLContext context = GLContext.getCurrent();
    TextRenderer renderer = trMap.get(context);
    if (renderer == null) {
      renderer = new TextRenderer(new Font("Helvetica", Font.BOLD, 14), true); 
      renderer.setUseVertexArrays(!forceImmMode);
      trMap.put(context, renderer);
    }
    return renderer;
  }
  
  /**
   * Initializes the VNodeFactory. This should be called in the
   * GLEventListener.init method.
   * 
   */
  public static void init() {
    instance = new VShapeFactory();
  }
  
  public static VShapeFactory getInstance() {
    return instance;
  }
  
  public VLabel createLabel(String text) {
    VLabel label = new VLabel(text);
    return label;
  }
  
  private GeometryData createCircleGeom(float radius, int slices, float normal) {
    final int count = (slices + 2) * 3;
    FloatBuffer buf = BufferUtil.newFloatBuffer(count);
    // first vertex at 0
    buf.put(0);
    buf.put(0);
    buf.put(0);
    float interval = (float)(Math.PI * 2.0 / slices);
    // counter clockwise
    for (int s = slices; s >= 0; s--) {
      float angle = 0f;
      if (s != slices) angle = s * interval;
      float sa = (float)Math.sin(angle);
      float ca = (float)Math.cos(angle);
      buf.put(sa * radius); 
      buf.put(ca * radius);
      buf.put(0f);
    }
    buf.rewind();
    
    FloatBuffer normals = BufferUtil.newFloatBuffer(count);
    for (int i = 2; i < count; i +=3) {
      normals.put(i, normal);
    }
    
    return new GeometryData(buf, normals);
  }
  
  private GeometryData createCylinderGeom(float radius, float height, int slices) {
    // for each slice: two triangles + the overlapping two at the end.
    FloatBuffer buf = BufferUtil.newFloatBuffer((slices * 2 + 2) * 3);
    float interval = (float)(Math.PI * 2.0 / slices);
    float halfHeight = height / 2;
    
    for (int s = 0; s <= slices; s++) {
      float angle = 0f;
      if (s != slices) angle = s * interval;
      float x = (float)Math.sin(angle) * radius;
      float y = (float)Math.cos(angle) * radius;
      
      buf.put(x); 
      buf.put(y);
      buf.put(halfHeight);
      
      buf.put(x);
      buf.put(y);
      buf.put(-halfHeight);
    }
    buf.rewind();
    
    FloatBuffer normals = BufferUtil.newFloatBuffer(buf.limit());
    float[] pt = new float[3];
    Vector3f normal = new Vector3f();
    while (buf.hasRemaining()) {
      buf.get(pt);
      normal.set(pt);
      normal.z = 0;
      normal.normalize();
      normals.put(normal.x);
      normals.put(normal.y);
      normals.put(normal.z);
    }
    
    return new GeometryData(buf, normals);
  }
  
  /*
  private FloatBuffer createNormalsTS(FloatBuffer vertices) {
    float[] vectors = new float[vertices.limit()];
    vertices.rewind();

    // get the first point
    float[] pt = new float[3];
    Point3f p0 = new Point3f();
    vertices.get(pt);
    p0.set(pt);

    // get the second point
    Point3f p1 = new Point3f();
    vertices.get(pt);
    p1.set(pt);

    Point3f p2 = new Point3f();
    // iterate through the remaining points
    // with triangle of p0, p1 and p2
    // and after check p1 is set to p2.
    // and p2 is then the next point
    Vector3f normal = new Vector3f();
    int start = 0;
    while (vertices.hasRemaining()) {
      vertices.get(pt);
      p2.set(pt);
      Utils3D.triangleNormal(p0, p1, p2, normal);
      
      for (int i = start; i < start + 9; i += 3) {
	vectors[i] += normal.x;
	vectors[i + 1] += normal.y;
	vectors[i + 2] += normal.z;
      }
      
      start += 3;

      p0.set(p1);
      p1.set(p2);
    }
    
    //System.out.println(Arrays.toString(vectors));
    
   
    // first two and last two vertices are the same so we need to add their values
    int end = vectors.length - 1;
    for (int i = 0; i < 6; i++) {
      float val = vectors[i];
      vectors[i] += vectors[end - i];
      vectors[end - i] += val;
    } 
  
    
    FloatBuffer normals = BufferUtil.newFloatBuffer(vertices.limit());
    for (int i = 0; i < vectors.length; i += 3) {
      normal.set(vectors[i], vectors[i + 1], vectors[i + 2]);
      normal.normalize();
      normals.put(normal.x);
      normals.put(normal.y);
      normals.put(normal.z);
      System.out.println("normal = " + normal);
    }
    System.out.println();
    normals.rewind();
   
    return normals;
  }
	*/
  
  public VSpatial createCircle(float radius, int slices) {
    GeometryData geom = createCircleGeom(radius, slices, 1f);
    geom.rewind();
    return new VShapeTF(geom);
  }
 
  public VSpatial createCylinder(float radius, float height, int slices) {
    GeometryData geom = createCylinderGeom(radius, height, slices);
    geom.rewind();
    return new VShapeTS(geom);
  }
  
  public VSpatial createDisk(float radius, float height, int slices) {
    // share the circle geometry
    GeometryData geom = createCircleGeom(radius, slices, 1);
    geom.rewind();
    VGeometry top = new VShapeTF(geom);
    // bottom needs normal of -1 so its back face
    // is the outside
    int count = geom.getNormals().limit();
    FloatBuffer normals = BufferUtil.newFloatBuffer(count);
    for (int i = 2; i < count; i +=3) {
      normals.put(i, -1);
    }
   
    geom = new GeometryData(geom.getVertices(), normals);
    geom.rewind();
    VGeometry bottom = new VShapeTF(geom);
    
    geom = createCylinderGeom(radius, height, slices);
    geom.rewind();
    VGeometry cylinder = new VShapeTS(geom);
    bottom.translate(0, 0, -height / 2);
    top.translate(0, 0, height / 2);
    CompositeGeometry disk = new CompositeGeometry(cylinder.getLocalBoundingSphere(), bottom, cylinder, top);
    return disk;
  }
}
