package anl.aida.data;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import java.io.IOException;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import javax.measure.unit.NonSI;

import org.junit.Test;

import anl.aida.AIDADatasetProducer;
import anl.aida.data.Axes;
import anl.aida.data.AxisType;
import anl.aida.data.DataFrame;
import anl.aida.data.DataFrameAxis;
import anl.aida.data.DataFrameIndex;
import anl.aida.data.DataFrameStringAxis;
import anl.aida.data.DataFrameTimeAxis;
import anl.aida.data.Dataset;
import anl.aida.data.Range;
import anl.aida.data.Variable;
import anl.aida.data.mock.ERDataFeed;
import anl.aida.data.mock.WIDataFeed;
import anl.aida.data.mock.WIHealthEmployeeFeed;
import anl.aida.data.mock.WIPCRFeed;

public class DataTest {
  
  @Test
  public void testH2CAS() throws IOException {
    AIDADatasetProducer producer = new AIDADatasetProducer(AxisType.CATEGORY, "Driver", NonSI.DAY);
    Dataset dataset = producer.createDataset("./data/H2CAS.bin", "H2CAS Model Data");
    assertEquals(26, dataset.getVariables().size());
    
    Variable var = dataset.getVariable("dTotDist");
    assertNotNull(var);
    DataFrame frame = dataset.getDataFrame(var);
    assertNotNull(frame);
    
    Axes<DataFrameAxis> axes = frame.getAxes();
    DataFrameAxis axis = axes.getTimeAxis();
    Range range = axis.getRange();
    assertEquals(0, range.getOrigin());
    assertEquals(140, range.getExtent());
    
    DataFrameStringAxis catAxis = (DataFrameStringAxis) axes.getAxis(AxisType.CATEGORY);
    assertNotNull(catAxis);
    int catIndex = catAxis.getIndex("206");
    
    DataFrameIndex index = frame.getIndex();
    assertNotNull(index);
    index.set(0, catIndex);
    assertEquals(2066.9f, frame.getFloat(index), 0f);
    
    index.set(98, catIndex);
    assertEquals(2543.3f, frame.getFloat(index), 0f);
  }
  
  @Test
  public void testGMAIDAFormat() throws IOException {
    AIDADatasetProducer producer = new AIDADatasetProducer(AxisType.NAMED_LOCATION, "Country", NonSI.YEAR);
    Dataset dataset = producer.createDataset("./data/gap_minder_data.bin", "GAP Minder Data");
    assertEquals(197, dataset.getVariables().size());
    
    Variable var = dataset.getVariable("Life_expectancy_at_birth");
    assertNotNull(var);
    DataFrame frame = dataset.getDataFrame(var);
    assertNotNull(frame);
    
    Axes<DataFrameAxis> axes = frame.getAxes();
    DataFrameAxis axis = axes.getTimeAxis();
    Range range = axis.getRange();
    assertEquals(0, range.getOrigin());
    assertEquals(2007 - 1800 + 1, range.getExtent());
    Calendar cal = GregorianCalendar.getInstance();
    Date date = axes.getDate(0);
    cal.setTime(date);
    assertEquals(cal.get(Calendar.YEAR), 1800);
    date = axes.getDate((int)range.getUpperBound());
    cal.setTime(date);
    assertEquals(cal.get(Calendar.YEAR), 2007);
    
    DataFrameIndex index = frame.getIndex();
    assertNotNull(index);
    
    DataFrameStringAxis catAxis = (DataFrameStringAxis) axes.getAxis(AxisType.NAMED_LOCATION);
    int catIndex = catAxis.getIndex("Sweden");
    assertEquals(165, catIndex);
    
    index.set(0, catIndex);
    assertEquals(32.16f, frame.getFloat(index), 0f);
    index.set(103, catIndex);
    assertEquals(55.09f, frame.getFloat(index), 0f);
  }
  
  @Test
  public void testFluCountsData() throws IOException {
    WIDataFeed feed = new WIDataFeed();
    DataFrame frame = feed.getDataFrame(feed.getVariables().get(0));
    assertNotNull(frame);
    
    DataFrameTimeAxis axis = (DataFrameTimeAxis) frame.getAxes().getTimeAxis();
    assertEquals(0, axis.getOrigin());
    assertEquals(108, axis.getExtent());
  }
  
  @Test
  public void testERData() throws IOException {
    ERDataFeed feed = new ERDataFeed();
    assertEquals(2, feed.getVariables().size());
    Variable var = feed.getVariable("Fever_Flu");
    DataFrame dataFrame = feed.getDataFrame(var);
    assertEquals(304, dataFrame.getAxes().getTimeAxis().getExtent());
    assertEquals(8, dataFrame.getAxes().getAxis("Age Range").getExtent());
    
    DataFrameIndex index = dataFrame.getIndex();
    index.set(1, 1);
    assertEquals(2, dataFrame.getInt(index));
  }
  
  @Test
  public void testPCRData() throws IOException {
    WIPCRFeed feed = new WIPCRFeed();
    assertEquals(6, feed.getVariables().size());
    Variable var = feed.getVariable("A_H1");
    DataFrame frame = feed.getDataFrame(var);
    assertEquals(44, frame.getAxes().getTimeAxis().getExtent());
    DataFrameIndex index = frame.getIndex();
    index.setTime(14);
    assertEquals(12, frame.getInt(index));
  }
  
  @Test
  public void testNAICSData() throws IOException {
    WIHealthEmployeeFeed feed = new WIHealthEmployeeFeed();
    assertEquals(5, feed.getVariables().size());
    Variable var = feed.getVariable("Population");
    DataFrame dataFrame = feed.getDataFrame(var);
    DataFrameIndex index = dataFrame.getIndex();
    index.set(0, 4);
    assertEquals(245018, dataFrame.getInt(index));
    
    var = feed.getVariable("AHCS_Employees");
    dataFrame = feed.getDataFrame(var);
    index = dataFrame.getIndex();
    index.set(0, 4);
    assertEquals(6650, dataFrame.getInt(index));
    
    
    
  }
}
