/**
 * 
 */
package anl.aida.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Reads the gap minder csv data and writes it to a binary file that we can load
 * as a data frame.
 * 
 */
public class GapMinderCSVReader {

  private static class Row {

    String category, indicator;
    int year;
    float value;

    public Row(String row) {
      row = row.trim();
      String delimiter = "\",\"";
      int start = 0;
      int end = row.indexOf(delimiter);
      category = row.substring(start + 1, end).trim();

      start = end + delimiter.length();
      end = row.indexOf(delimiter, start);
      year = Integer.valueOf(row.substring(start, end).trim());

      delimiter = "\",";
      start = end + delimiter.length();
      end = row.indexOf(delimiter, start);
      indicator = row.substring(start + 1, end).trim();

      start = end + delimiter.length();
      value = Float.valueOf(row.substring(start, row.length()).trim());
    }

    public String toString() {
      return category + " " + indicator + " at " + year + ": " + value;
    }
  }

  public static class Value {

    int tIndex, cIndex;
    float value;

    public Value(int timestep, int category, float value) {
      tIndex = timestep;
      cIndex = category;
      this.value = value;
    }
  }

  private static class Indicator {

    String indicator;
    Map<String, Integer> catMap = new HashMap<String, Integer>();
    List<String> catList = new ArrayList<String>();
    List<Value> values = new ArrayList<Value>();
    long filePos;

    int max = Integer.MIN_VALUE;
    int min = Integer.MAX_VALUE;

    public Indicator(String name) {
      this.indicator = name;
    }

    public void add(Row row) {
      catMap.put(row.category, 0);
      if (row.year < min)
	min = row.year;
      if (row.year > max)
	max = row.year;
    }

    public String toString() {
      return indicator + " " + min + ", " + max + " " + catMap.size() + " count";
    }

    public void initForValues() {
      catList.addAll(catMap.keySet());
      Collections.sort(catList);
      int i = 0;
      for (String category : catList) {
	catMap.put(category, i++);
      }
    }

    public void addValue(Row row) {
      int ts = row.year - min;
      int cIndex = catMap.get(row.category);
      values.add(new Value(ts, cIndex, row.value));
    }

    public void writeHeader(RandomAccessFile out) throws IOException {
      out.writeUTF(indicator);
      out.writeInt(catList.size());
      for (String cat : catList) {
	out.writeUTF(cat);
      }
      out.writeInt(min);
      out.writeInt(max);
      out.writeInt(values.size());
      filePos = out.getFilePointer();
      out.writeLong(0);
    }

    public void writeData(RandomAccessFile out) throws IOException {
      long pointer = out.getFilePointer();
      out.seek(filePos);
      out.writeLong(pointer);
      out.seek(pointer);
      for (Value value : values) {
	out.writeInt(value.tIndex);
	out.writeInt(value.cIndex);
	out.writeFloat(value.value);
      }
    }
  }

  public void run(String source, String outFile) throws IOException {
    BufferedReader reader = new BufferedReader(new FileReader(source));
    // skip the first line
    String line = reader.readLine();
    Map<String, Indicator> indicators = new HashMap<String, Indicator>();
    // StringBuffer buf = new StringBuffer();
    while ((line = reader.readLine()) != null) {
      Row row = new Row(line);
      Indicator indicator = indicators.get(row.indicator);
      if (indicator == null) {
	indicator = new Indicator(row.indicator);
	indicators.put(row.indicator, indicator);
      }
      indicator.add(row);
      // buf.append("\"");
      // buf.append(indicator.indicator);
      // buf.append("\",");
    }
    reader.close();

    for (Indicator indicator : indicators.values()) {
      indicator.initForValues();
    }

    reader = new BufferedReader(new FileReader(source));
    // skip the first line
    line = reader.readLine();
    while ((line = reader.readLine()) != null) {
      Row row = new Row(line);
      Indicator indicator = indicators.get(row.indicator);
      indicator.addValue(row);
    }
    reader.close();

    RandomAccessFile out = new RandomAccessFile(new File(outFile), "rw");

    out.writeInt(indicators.size());
    for (Indicator indicator : indicators.values()) {
      indicator.writeHeader(out);
    }

    for (Indicator indicator : indicators.values()) {
      indicator.writeData(out);
    }
    out.close();

  }

  public static void main(String[] args) throws IOException {
    GapMinderCSVReader reader = new GapMinderCSVReader();
    reader.run("/Users/nick/Desktop/data_set_1018194.csv", "./data/gap_minder_data.bin");
  }

}
