package anl.aida.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

public class H2CasCSVReader {

  private class Header {
    String name;
    int dataCount, index;
    // dcOffset so we can return to write the data count
    // once we know how much data
    long dcOffset, dataOffset;

    public Header(String name, int index) {
      this.name = name;
      this.index = index;
    }

    public void write(RandomAccessFile out, Set<Integer> ids, int minTS, int maxTS)
	throws IOException {
      out.writeUTF(name);
      out.writeInt(ids.size());
      for (Integer id : ids) {
	out.writeUTF(String.valueOf(id));
      }
      out.writeInt(minTS);
      out.writeInt(maxTS);
      dcOffset = out.getFilePointer();
      out.writeInt(0);
      out.writeLong(0);
    }

    public void writeData(RandomAccessFile out, String[] items, Map<Integer, Integer> idMap)
	throws IOException {
      dataCount++;
      int id = Integer.parseInt(items[1]);
      int idIndex = idMap.get(id);
      // first timestep is 1, so the index of that
      // timestep in the ts dimension is 0
      int tIndex = Integer.parseInt(items[0]) - 1;
      out.writeInt(tIndex);
      out.writeInt(idIndex);
      out.writeFloat(Float.valueOf(items[index]));
    }

    public String toString() {
      return name;
    }

    public void writeOffsets(RandomAccessFile out) throws IOException {
      out.seek(dcOffset);
      out.writeInt(dataCount);
      out.writeLong(dataOffset);
    }

  }

  public void run(String source, String outFile) throws IOException {
    BufferedReader reader = new BufferedReader(new FileReader(source));
    // read the headers
    String line = reader.readLine().trim();
    String[] names = line.split(",");
    List<Header> headers = new ArrayList<Header>();
    for (int i = 2; i < names.length; i++) {
      String name = names[i].trim();
      int index = name.indexOf("[INPUT]");
      if (index != -1)
	name = name.substring(0, index);
      name = name.trim();
      name = name.replace(" - ", "_");
      name = name.replace(" ", "_");
      
      headers.add(new Header(name, i));
    }
    
    
    Set<Integer> ids = new HashSet<Integer>();
    int minTS = Integer.MAX_VALUE;
    int maxTS = Integer.MIN_VALUE;

    int i = 0;
    while ((line = reader.readLine()) != null) {
      String[] items = line.split(",");
      int timestep = Integer.parseInt(items[0]);
      int id = Integer.parseInt(items[1]);
      ids.add(id);
      minTS = Math.min(minTS, timestep);
      maxTS = Math.max(maxTS, timestep);
      if (i % 10000 == 0)
	System.out.println(i);
      if (i == 0) {
	// test for non-numeric data.
	for (int j = 2; j < items.length; j++) {
	  try {
	    Double.parseDouble(items[j]);
	    //
	  } catch (NumberFormatException ex) {
	    // if here then non-numeric so remove that header.
	    for (Iterator<Header> iter = headers.iterator(); iter.hasNext();) {
	      Header header = iter.next();
	      if (header.index == j) {
		iter.remove();
		break;
	      }
	    }
	  }
	}
      }
      i++;
    }
    reader.close();

    RandomAccessFile out = new RandomAccessFile(new File(outFile), "rw");

    out.writeInt(headers.size());
    for (Header header : headers) {
      header.write(out, ids, minTS, maxTS);
    }

    // make map of ids for each lookup
    // key is the id and value is the index of that id
    Map<Integer, Integer> idMap = new HashMap<Integer, Integer>();
    int index = 0;
    for (Integer id : ids) {
      idMap.put(id, index);
      index++;
    }

    for (Header header : headers) {
      System.out.println("Processing '" + header.name + "'");
      reader = new BufferedReader(new FileReader(source));
      // skip the first line
      reader.readLine();
      header.dataOffset = out.getFilePointer();
      while ((line = reader.readLine()) != null) {
	String[] items = line.split(",");
	header.writeData(out, items, idMap);
      }
      reader.close();
    }
    
    for (Header header : headers) {
      header.writeOffsets(out);
    }
    
    out.close();
  }

  public static void main(String[] args) {
    try {
      new H2CasCSVReader().run("./data/H2CAS_GapMinder.csv", "./data/H2CAS.bin");
    } catch (IOException ex) {
      ex.printStackTrace();
    }
  }

}
