package anl.aida;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.StringTokenizer;

import org.apache.uima.collection.metadata.CpeCasProcessor;
import org.apache.uima.collection.metadata.CpeDescription;
import org.apache.uima.collection.metadata.CpeDescriptorException;

import anl.aida.ae.gate.GateAnnotator;
import anl.aida.ae.lucene.LuceneConsumer;
import anl.aida.reader.AIDACompositeReader;
import anl.aida.util.FileUtilities;
import anl.aida.util.Pair;
import anl.aida.util.UIMAParameterSettings;

/**
 * Retreives a configuration and sets the UIMA CPE parameters appropriately.
 * 
 */
public class Configurator {

  static {
    System.setProperty("java.util.logging.config.file", "./config/FileConsoleLogger.properties");
  }

  private static String GAPP_FILE_KEY = "gate.gapp.file";
  private static String GATE_MAPPING_KEY = "gate.mapping.file";
  private static String GATE_HOME_KEY = "gate.home";
  private static String GATE_JAPE_KEY = "gate.jape.directory";
  private static String GATE_GAZETTEER_KEY = "gate.gazetteer.directory";

  private static String LUCENE_INDEX_KEY = "lucene.index";
  private static String LUCENE_MAPPING_KEY = "lucene.mapping";
  private static String OVERWRITE_INDEX_KEY = "lucene.overwrite.index";

  private static String CACHE_DIR_KEY = "cache.dir";
  private static String CACHE_CREATE_KEY = "cache.create";
  private static String CACHE_START_DATE_KEY = "cache.start.date";
  private static String CACHE_LENGTH_KEY = "cache.length";
  private static String CACHE_TS_DAYS_KEY = "cache.timestep.days";
  private static String CACHE_FILTER_KEY="cache.filter";

  private static String READER_NAMES_KEY = "reader.class.names";
  private static String READER_PROP_PREFIX = "reader.property.";

  private static String JAPE_KEY = "jape.directory";
  private static String GAZETTEER_KEY = "gazetteer.directory";

  private static Configurator instance;

  private Properties props;
  private List<String> readerClassNames = new ArrayList<String>();
  private List<Pair<String, String>> readerProperties = new ArrayList<Pair<String, String>>();

  /**
   * Initializes this Configurator with the specified config file.
   * 
   * @param configFile
   *          the file to read config info from
   * 
   * @return the initialized Configurator.
   * @throws IOException
   *           if there is an error rading the config file.
   */
  public static Configurator init(String configFile) throws IOException {
    instance = new Configurator(configFile);
    return instance;
  }

  public static Configurator getConfigurator() {
    if (instance == null)
      throw new RuntimeException("Configurator must be initialized.");
    return instance;
  }

  private Configurator(String configFile) throws IOException {
    props = new Properties();
    File file = new File(configFile);
    props.load(new BufferedReader(new FileReader(file)));
    initFileProps(file.getParentFile());
    initReaderProps();
    replaceVars();
  }

  // replace any variable entries (e.g ${x.x} with their actual values)
  private void replaceVars() throws IOException {
    for (Map.Entry<Object, Object> entry : props.entrySet()) {
      entry.setValue(reifyVariable(entry.getKey().toString(), 
	  entry.getValue().toString(), new HashSet<String>()));
      //System.out.println(entry.getValue());
    }
  }

  private String reifyVariable(String name, String value, Set<String> workingProps) throws IOException {
    workingProps.add(name);
    int startIndex = value.indexOf("${");
    while (startIndex != -1) {
      int endIndex = value.indexOf("}", startIndex);
      if (endIndex == -1)
	throw new IOException("Invalid variable declaration in '" + value + "'");
      String key = value.substring(startIndex + 2, endIndex);
      String variableVal = props.getProperty(key);
      
      if (variableVal == null)
	throw new IOException("Invalid variable: property '" + key + "' not found");
      if (workingProps.contains(key)) throw new IOException("Invalid variable declaration: circular reference '" +
	  key + "'");
      // reify the variables in the referenced variable
      variableVal = reifyVariable(key, variableVal, workingProps);
      props.put(key, variableVal);
      
      if (startIndex > 0) {
	value = value.substring(0, startIndex) + variableVal
	    + value.substring(endIndex + 1, value.length());
      } else {
	value = variableVal + value.substring(endIndex + 1, value.length());
      }
      startIndex = value.indexOf("${", endIndex);
    }
    workingProps.remove(name);

    return value;
  }

  // parse out the reader related properties.
  private void initReaderProps() throws IOException {
    String names = props.getProperty(READER_NAMES_KEY);
    for (StringTokenizer tok = new StringTokenizer(names, ","); tok.hasMoreTokens();) {
      readerClassNames.add(tok.nextToken().trim());
    }

    int start = READER_PROP_PREFIX.length();
    for (Map.Entry<Object, Object> entry : props.entrySet()) {
      String pName = entry.getKey().toString().trim();
      if (pName.startsWith(READER_PROP_PREFIX)) {
	if (pName.length() > start) {
	  Pair<String, String> prop = new Pair<String, String>(pName.substring(start, pName
	      .length()), entry.getValue().toString());
	  readerProperties.add(prop);
	} else
	  throw new IOException("Invalid reader property value '" + pName + "'");
      }
    }
  }

  // convert paths in file properties so that
  // "." means the directory of the config file.
  private void initFileProps(File configDirectory) throws IOException {
    String[] fileProps = { GAPP_FILE_KEY, GATE_MAPPING_KEY, GATE_HOME_KEY, LUCENE_INDEX_KEY,
	LUCENE_MAPPING_KEY, CACHE_DIR_KEY, GATE_JAPE_KEY, GATE_GAZETTEER_KEY, JAPE_KEY,
	GAZETTEER_KEY};
    Set<String> set = new HashSet<String>(Arrays.asList(fileProps));

    // also do any reader properties that looks like file paths
    for (Map.Entry<Object, Object> entry : props.entrySet()) {
      String pName = entry.getKey().toString().trim();
      if (set.contains(pName) || pName.startsWith(READER_PROP_PREFIX)) {
	String newVal = transformPath(configDirectory, entry.getValue().toString());
	entry.setValue(newVal);
      }
    }
  }

  private String transformPath(File configDir, String val) throws IOException {
    if (val.startsWith("./")) {
      val = configDir.getCanonicalPath() + val.substring(1);
    } else if (val.startsWith("../")) {
      val = configDir.getCanonicalPath() + "/" + val;
    }
    return val;
  }

  /**
   * Gets an interable over the class names of the component readers to use to
   * create the cache. These names are pulled from the config file.
   * 
   * @return an interable over the class names of the component readers to use
   *         to create the cache.
   */
  public Iterable<String> readerClassNames() {
    return readerClassNames;
  }

  /**
   * Gets an iterable over the reader configuration properties. A reader
   * configuration properties is an property in the config file prefixed with
   * reader.property.
   * 
   * @return an iterable over the reader configuration properties.
   */
  public Iterable<Pair<String, String>> readerProperties() {
    return readerProperties;
  }

  // copy the gate and jape files to the gate dirs
  private void copyFiles() throws IOException {
    File targetDir = new File(props.getProperty(GATE_JAPE_KEY));
    File srcDir = new File(props.getProperty(JAPE_KEY));
    //System.out.println("srcDir = " + srcDir);
    //System.out.println("targetDir = " + targetDir);
    FileUtilities.copyDirs(srcDir, targetDir, new FileFilter() {
      @Override
      public boolean accept(File pathname) {
	return !pathname.getName().startsWith(".");
      }
    });

    targetDir = new File(props.getProperty(GATE_GAZETTEER_KEY));
    srcDir = new File(props.getProperty(GAZETTEER_KEY));
    FileUtilities.copyDirs(srcDir, targetDir, new FileFilter() {
      
      public boolean accept(File pathname) {
	return !pathname.getName().startsWith(".");
      }
    });
  }

  /**
   * Configures the specified CPE using the specified config file.
   * 
   * @param desc
   *          the CpeDescription to configure
   */
  public void configure(CpeDescription desc) throws IOException, CpeDescriptorException {
    // copy the gate and jape files to the gate dirs
    copyFiles();
    CpeCasProcessor[] procs = desc.getCpeCasProcessors().getAllCpeCasProcessors();
    for (CpeCasProcessor proc : procs) {
      // System.out.println("proc.getName() = " + proc.getName());
      if (proc.getName().equals("GateAnnotator")) {
	UIMAParameterSettings settings = new UIMAParameterSettings();
	settings.addPair(GateAnnotator.GATE_APP_FILE, props.getProperty(GAPP_FILE_KEY));
	settings.addPair(GateAnnotator.MAPPING_FILE, props.getProperty(GATE_MAPPING_KEY));
	settings.addPair(GateAnnotator.GATE_HOME, props.getProperty(GATE_HOME_KEY));
	proc.setConfigurationParameterSettings(settings);
      } else if (proc.getName().equals("LuceneConsumer")) {
	UIMAParameterSettings settings = new UIMAParameterSettings();
	settings.addPair(LuceneConsumer.CLEAR_INDEX, props.getProperty(OVERWRITE_INDEX_KEY));
	settings.addPair(LuceneConsumer.INDEX_DIR, props.getProperty(LUCENE_INDEX_KEY));
	settings.addPair(LuceneConsumer.MAPPING_FILE, props.getProperty(LUCENE_MAPPING_KEY));
    settings.addPair(LuceneConsumer.KEYWORD_FILTER,props.getProperty(CACHE_FILTER_KEY));
	settings.addPair(LuceneConsumer.CREATE_CACHE, props.getProperty(CACHE_CREATE_KEY));
	settings.addPair(LuceneConsumer.CACHE_START_DATE, props.getProperty(CACHE_START_DATE_KEY));
	settings.addPair(LuceneConsumer.CACHE_LENGTH_DAYS, Integer.valueOf(props
	    .getProperty(CACHE_LENGTH_KEY)));
	settings.addPair(LuceneConsumer.CACHE_TS_INTERVAL_DAYS, Integer.valueOf(props
	    .getProperty(CACHE_TS_DAYS_KEY)));
	settings.addPair(LuceneConsumer.CACHE_OUTPUT_DIR, props.getProperty(CACHE_DIR_KEY));
	proc.setConfigurationParameterSettings(settings);
      }
    }

    // change the index files to the cached archives
    UIMAParameterSettings settings = new UIMAParameterSettings();
    settings.addPair(AIDACompositeReader.CACHE_START_DATE, props.getProperty(CACHE_START_DATE_KEY));
    desc.getAllCollectionCollectionReaders()[0].setConfigurationParameterSettings(settings);
  }

}
