package anl.aida.ae.gate;

import org.apache.uima.cas.Feature;
import org.apache.uima.cas.Type;
import org.apache.uima.cas.TypeSystem;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.ext.DefaultHandler2;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.ParserConfigurationException;
import java.io.File;
import java.io.IOException;
import java.util.List;
import java.lang.reflect.Constructor;

/**
 *         Date: Jan 15, 2009 4:38:31 PM
 */
public class BuilderFactory extends DefaultHandler2 {

  private static final String UIMA = "uima";
  private static final String FEATURE = "feature";
  private static final String FEATURE_NAME = "name";
  private static final String GATE_FEATURE_NAME = "gate_feature_name";
  private static final String FEATURE_TYPE = "type";
  private static final String UPDATED = "updated";

  private static final String UIMA_TYPE_NAME = "type_name";
  private static final String GATE_TYPE_NAME = "gate_type_name";
  private static final String FILTER_NAME = "filter";
  private static final String SET_NAME = "annotation_set";
  private static final String CLASS = "class";

  private static final String INPUTS = "inputs";
  private static final String OUTPUTS = "outputs";
  private static final String NEW = "new";
  private static final String UPDATE = "update";


  private AnnotationBuilder builder;
  private List<InputAnnotationBuilder> inputs;
  private List<OutputAnnotationBuilder> outputs;
  private TypeSystem typeSystem;
  private boolean processInputs, outUpdate;

  private File file;

  public BuilderFactory(String mappingFile) {
    file = new File(mappingFile);
  }

  public void build(TypeSystem typeSystem, List<InputAnnotationBuilder> inputs, List<OutputAnnotationBuilder> outputs) throws ParserConfigurationException, SAXException, IOException {
    this.typeSystem = typeSystem;
    builder = null;
    this.inputs = inputs;
    this.outputs = outputs;

    
      SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
      parser.parse(file, this);

    // builders are added when a new builder is found in the xml
    // so we need to last the last one here there would be no new
    // builder to trigger adding it.
    if (builder != null) addBuilder();
  }

  private void createBuilder(Attributes attributes) throws SAXException {
    String typeName = attributes.getValue(UIMA_TYPE_NAME);
    String gateName = attributes.getValue(GATE_TYPE_NAME);
    Type type = typeSystem.getType(typeName);
    if (type == null) {
      throw new SAXException("UIMA type '" + typeName + "' not found.");
    }

    if (processInputs) {
      String updated = attributes.getValue(UPDATED);
      builder = new InputAnnotationBuilder(gateName, type, Boolean.parseBoolean(updated));
    } else {
      builder = new OutputAnnotationBuilder(gateName, type, outUpdate);
      String filterName = attributes.getValue(FILTER_NAME);
      if (filterName != null && filterName.length() > 0) {
	try {
	  Class<?> filterClass = Class.forName(filterName);
	  AnnotationBuilderFilter filter = (AnnotationBuilderFilter) filterClass.newInstance();
	  builder.setFilter(filter);
	} catch (Exception ex) {
	  throw new SAXException(ex);
	}
      }
    }

    String annSet = attributes.getValue(SET_NAME);
    if (annSet != null && annSet.length() > 0)
      builder.setAnnotationSet(annSet);
  }

  private void createFeature(Attributes attributes) throws SAXException {
    String clazz = attributes.getValue(CLASS);
    if (clazz != null && clazz.length() > 0) {
      try {
        Class<?> builderClass = Class.forName(clazz);
        Constructor<?> ctr = builderClass.getConstructor(TypeSystem.class);
        FeatureBuilder featBuilder = (FeatureBuilder)ctr.newInstance(typeSystem);
        if (processInputs) builder.addFeatureBuilder(featBuilder);
        else builder.addFeatureBuilder(featBuilder);

      } catch (Exception ex) {
        throw new SAXException(ex);
      }

    } else {
      String featureName = builder.getUIMAName() + ":" + attributes.getValue(FEATURE_NAME);
      Feature feature = typeSystem.getFeatureByFullName(featureName);
      if (feature == null) {
        throw new SAXException("UIMA feature '" + featureName + "' not found.");
      }
      String gateName = attributes.getValue(GATE_FEATURE_NAME);
      String typeName = attributes.getValue(FEATURE_TYPE);
      FeatureType type = FeatureType.parse(typeName);
      if (type == null) {
        throw new SAXException("Feature type '" + typeName + "' is not supported.");
      }

      if (processInputs) builder.addFeatureBuilder(new InputFeatureBuilder(gateName, feature, type));
    else builder.addFeatureBuilder(new OutputFeatureBuilder(gateName, feature, type));
    }


  }

  private void addBuilder() {
    if (processInputs) inputs.add((InputAnnotationBuilder) builder);
    else outputs.add((OutputAnnotationBuilder) builder);
    builder = null;
  }

  @Override
  public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException {
    if (qName.equals(INPUTS)) processInputs = true;
    else if (qName.equals(OUTPUTS)) {
      if (builder != null) addBuilder();
      processInputs = false;
    } else if (qName.equals(UPDATE)) outUpdate = true;
    else if (qName.equals(NEW)) outUpdate = false;
    else if (qName.equals(UIMA)) {
      // start of new annotation mapping
      if (builder != null) addBuilder();
      createBuilder(attributes);
    } else if (qName.equals(FEATURE)) {
      createFeature(attributes);
    }
  }
}
