package anl.aida.ae.gate;

import gate.Corpus;
import gate.CorpusController;
import gate.Document;
import gate.Factory;
import gate.Gate;
import gate.creole.ResourceInstantiationException;
import gate.util.persistence.PersistenceManager;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.ParserConfigurationException;

import org.apache.log4j.PropertyConfigurator;
import org.apache.uima.UimaContext;
import org.apache.uima.analysis_component.CasAnnotator_ImplBase;
import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.cas.CAS;
import org.apache.uima.cas.FeatureStructure;
import org.apache.uima.cas.TypeSystem;
import org.apache.uima.resource.ResourceInitializationException;
import org.apache.uima.util.Level;
import org.apache.uima.util.Logger;
import org.xml.sax.SAXException;

/**
 * Class provides Gate annotation services.
 */
public class GateAnnotator extends CasAnnotator_ImplBase {

  public final static String MAPPING_FILE = "anl.aida.ae.gate.GateAnnotator.MappingFile";
  public final static String GATE_APP_FILE = "anl.aida.ae.gate.GateAnnotator.GateAppFile";
  public final static String GATE_HOME = "anl.aida.ae.gate.GateAnnotator.GateHome";
  public final static String GATE_LOG_PROPS = "anl.aida.ae.gate.GateAnnotator.GateLogProps";

  private final static String MESSAGE_DIGEST = "anl.aida.ae.gate.GateAnnotator_Messages";

  private List<InputAnnotationBuilder> inputBuilders;
  private List<OutputAnnotationBuilder> outputBuilders;
  // maps a UIMA type name to a map of GATE annotation ids ->
  // corresponding UIMA features structures. This is used to
  // update existing UIMA annotations with new info from GATE
  // annotations.
  private Map<String, Map<Integer, FeatureStructure>> indexMap;
  private Corpus gateCorpus;
  private BuilderFactory builderFactory;
  private CorpusController gateApp;

  @Override
  public void initialize(UimaContext context) throws ResourceInitializationException {
    
    String logProps = (String) context.getConfigParameterValue(GATE_LOG_PROPS);
    PropertyConfigurator.configure(logProps);
    
    indexMap = new HashMap<String, Map<Integer, FeatureStructure>>();
    String mappingFile = (String) context.getConfigParameterValue(MAPPING_FILE);
    String gateFile = (String) context.getConfigParameterValue(GATE_APP_FILE);
    String gHome = (String) context.getConfigParameterValue(GATE_HOME);
    
    Logger logger = context.getLogger();
    logger.log(Level.INFO, String.format("Initializing GATEAnnotator:%n\tGateHome: %s%n\tGateApp: %s%n\tMappingFile: %s" +
	"%n\tLogging Properties File: %s", 
    	gHome, gateFile, mappingFile, logProps));

    try {

      if (Gate.getGateHome() == null) {
        File gateHome = new File(gHome);

        Gate.setGateHome(gateHome);
        Gate.setPluginsHome(new File(gateHome, "plugins"));
        Gate.setUserConfigFile(new File(gateHome, "user-gate.xml"));
        Gate.setUserSessionFile(new File(gateHome, ".gate.session"));
        Gate.init();
      }

      gateApp = (CorpusController)
              PersistenceManager.loadObjectFromUrl(new File(gateFile).toURI().toURL());

      builderFactory = new BuilderFactory(mappingFile);
      gateCorpus = Factory.newCorpus("UIMA corpus");
      gateApp.setCorpus(gateCorpus);

    } catch (Exception ex) {
      throw new ResourceInitializationException(MESSAGE_DIGEST, "gate_init_error",
              new Object[]{}, ex);
    }
  }


  @Override
  public void typeSystemInit(TypeSystem typeSystem) throws AnalysisEngineProcessException {
    super.typeSystemInit(typeSystem);
    try {
      inputBuilders = new ArrayList<InputAnnotationBuilder>();
      outputBuilders = new ArrayList<OutputAnnotationBuilder>();
      builderFactory.build(typeSystem, inputBuilders, outputBuilders);
    } catch (IOException e) {
      throw new AnalysisEngineProcessException(MESSAGE_DIGEST, "mapping_creation_error", new Object[]{},
              e);
    } catch (ParserConfigurationException e) {
      throw new AnalysisEngineProcessException(MESSAGE_DIGEST, "mapping_creation_error", new Object[]{},
              e);
    } catch (SAXException e) {
      throw new AnalysisEngineProcessException(MESSAGE_DIGEST, "mapping_creation_error", new Object[]{},
              e);
    }
  }

  private void createInputAnnotations(CAS cas, Document doc) throws UGMappingException {
    for (InputAnnotationBuilder builder : inputBuilders) {
      Map<Integer, FeatureStructure> map = new HashMap<Integer, FeatureStructure>();
      builder.run(cas, doc, map);
      if (builder.isIndexed()) {
        indexMap.put(builder.getUIMAName(), map);
      }
    }

  }

  private void createOutputAnnotations(CAS cas, Document doc) throws UGMappingException {
    for (OutputAnnotationBuilder builder : outputBuilders) {
      Map<Integer, FeatureStructure> map = builder.isIndexed() ? indexMap.get(builder.getUIMAName()) :
              new HashMap<Integer, FeatureStructure>();

      if (map == null) throw new UGMappingException("Updated UIMA annotation must be tagged as updated input.");
      builder.run(cas, doc, map);
    }
  }


  @Override
  public void process(CAS cas) throws AnalysisEngineProcessException {

    String docText = cas.getDocumentText();
    Document gateDoc = null;
    try {

      gateDoc = Factory.newDocument(docText);
      createInputAnnotations(cas, gateDoc);
      gateCorpus.add(gateDoc);
      gateApp.execute();

      createOutputAnnotations(cas, gateDoc);
    } catch (ResourceInstantiationException ex) {
      throw new AnalysisEngineProcessException(ex);

    } catch (Throwable ex) {
      throw new AnalysisEngineProcessException(ex);
    } finally {
      gateCorpus.remove(gateDoc);
      Factory.deleteResource(gateDoc);
      indexMap.clear();
    }
  }

  /**
   * Free the GATE objects created by this annotator.
   */
  public void destroy() {

    if (gateCorpus != null) {
      Factory.deleteResource(gateCorpus);
      gateCorpus = null;
    }

    if (gateApp != null) {
      Factory.deleteResource(gateApp);
      gateApp = null;
    }
  }
}
