package anl.aida.ae.lucene;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.StringTokenizer;

import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.Term;
import org.apache.lucene.store.FSDirectory;

import anl.aida.core.STCWriter;
import anl.aida.core.ScoredTermsCollection;
import anl.aida.util.DateRange;
import anl.aida.util.DateRangeProducer;
import anl.aida.util.FileUtilities;
import anl.aida.util.Filter;
import anl.aida.util.AIDAException;

/**
 *         Date: Mar 5, 2009 11:38:37 AM
 */
public class CacheCreator {

  private static final String LUCENE_INDEX_NAME = "index";

  private static final class AllFilter implements Filter<Term> {
    public boolean evaluate(Term item) {
      return true;
    }
  }


  private Filter<Term> filter;
  private IndexReader reader;
  private DateFormat format = new SimpleDateFormat("yyyMMdd");
  private DateRangeProducer drProducer;

  public CacheCreator(DateRangeProducer drProducer, IndexReader reader) {
    this(drProducer, new AllFilter(), reader);
  }

  public CacheCreator(DateRangeProducer drProducer, Filter<Term> filter, IndexReader reader) {
    this.drProducer = drProducer;
    this.filter = filter;
    this.reader = reader;
  }

  private void deleteFiles(File dir) {
    for (File file : dir.listFiles()) {
      if (file.isDirectory()) deleteFiles(file);
      file.delete();
    }
  }

  private String makeFileName(File outputDir, DateRange range, String prefix) {
    StringBuilder buf = new StringBuilder(prefix);
    buf.append("_");
    buf.append(format.format(range.getStart()));
    buf.append("_");
    buf.append(format.format(range.getEnd()));
    buf.append(".bin");
    File file = new File(outputDir, buf.toString());
    return file.getAbsolutePath();
  }

  /**
   * Creates the cache in the specified directory. If the output directory exists,
   * it will be deleted and recreated.
   *
   * @param outputDir the directory in which the cache is created
   * @throws IOException   if there is a file error while creating the cache
   * @throws AIDAException if there is a general error while creating the cache
   */
  public void createCache(File outputDir, ReaderProducer readerProducer, Keywords keywords) throws IOException,
          AIDAException, ParseException {

    if (outputDir.exists()) deleteFiles(outputDir);
    outputDir.mkdirs();

    PrintWriter out = new PrintWriter(new BufferedWriter(new FileWriter(new File(outputDir, "cache.xml"))));
    out.printf("<!-- paths are relative to the location of this file -->%n");
    out.printf("<cache lucene_index=\"%s\">%n", LUCENE_INDEX_NAME);
    out.printf("\t<range>%n");
    out.printf("\t\t<start>%d</start>%n\t\t<length>%d</length>%n\t\t<interval>%d</interval>%n",
            drProducer.getStart().getTime(), drProducer.getTotalDays(), drProducer.getRangeLength());
    out.printf("\t</range>%n");

    // create the STC for the entire cache
    STCCreator creator = new STCCreator();
    DateRange fullRange = new DateRange(drProducer.getStart(), drProducer.getEnd());
    ScoredTermsCollection stc = creator.createCollection(reader, filter, fullRange);
    STCWriter writer = new STCWriter();
    String filename = makeFileName(outputDir, fullRange, "stc");
    writer.writeCollection(filename, stc);
    out.printf("\t<cache_items>%n");
    out.printf("\t\t<cache_item>%s</cache_item>%n", new File(filename).getName());

    File tmpDir = new File(new File(System.getProperty("java.io.tmpdir")), "tmpIndex");
    for (DateRange range : drProducer.ranges()) {
      createTmpIndex(tmpDir, range, readerProducer, keywords);
      filename = makeFileName(outputDir, range, "stc");
      createSTC(tmpDir, filename, range);
      out.printf("\t\t<cache_item>%s</cache_item>%n", new File(filename).getName());
    }
    out.printf("\t</cache_items>%n");
    
    keywords.writeXML(out);
    out.printf("</cache>");
    out.close();

    copyIndex(outputDir);
  }

  private void copyIndex(File outputDir) throws IOException {
    File indexDir = new File(outputDir, LUCENE_INDEX_NAME);
    indexDir.mkdir();
    FSDirectory fs = (FSDirectory) reader.directory();
    File dir = fs.getFile();
    for (File file : dir.listFiles()) {
      FileUtilities.copyFile(file, new File(indexDir, file.getName()));
    }
  }

  private void createSTC(File indexDir, String filename, DateRange range) throws IOException, AIDAException {
    IndexReader reader = IndexReader.open(indexDir);
    STCCreator creator = new STCCreator();
    ScoredTermsCollection stc = creator.createCollection(reader, filter, range);
    STCWriter writer = new STCWriter();
    writer.writeCollection(filename, stc);
  }

  private void createTmpIndex(File tmpDir, DateRange range, ReaderProducer rProducer, 
      Keywords keywords) throws IOException,
          ParseException {
    if (tmpDir.exists()) deleteFiles(tmpDir);
    tmpDir.mkdirs();

    IndexWriter writer = new IndexWriter(tmpDir, new StandardAnalyzer(), IndexWriter.MaxFieldLength.UNLIMITED);
    DateRangeFieldFilter filter = new DateRangeFieldFilter(range.getStart(), range.getEnd());
    // get the docs that pass the filter, make a new doc of those, create index
    // create stc.
    for (int i = 0; i < reader.maxDoc(); i++) {
      Document oldDoc = reader.document(i);
      if (filter.evaluate(oldDoc)) {
        Document doc = new Document();
        doc.add(new Field(LuceneConstants.PATH_FIELD, oldDoc.get(LuceneConstants.PATH_FIELD),
                Field.Store.YES, Field.Index.NOT_ANALYZED));

        Field oldField = oldDoc.getField(LuceneConstants.TITLE_FIELD);
        Field field = new Field(LuceneConstants.TITLE_FIELD, oldField.stringValue(),
                Field.Store.YES, Field.Index.ANALYZED);
        field.setBoost(oldField.getBoost());
        doc.add(field);

        doc.add(new Field(LuceneConstants.TIMESTAMP_FIELD, oldDoc.get(LuceneConstants.TIMESTAMP_FIELD),
                Field.Store.YES, Field.Index.ANALYZED));

        // Add the contents of the file to a field named "contents".  Specify a Reader,
        // so that the text of the file is tokenized and indexed, but not stored.

        // THIS ONLY INDEXES THE FIRST 10K TERMS OF THE DOCUMENT
        doc.add(new Field(LuceneConstants.CONTENTS_FIELD, rProducer.getReader(oldDoc)));
        
        TokenStreamBuilder streamBuilder = new TokenStreamBuilder();
        for (String fieldName : keywords.keywordFields()) {
          Field cacheField = oldDoc.getField(fieldName + "_cache");
          if (cacheField != null) {
          streamBuilder.clear();
          String val = cacheField.stringValue();
          StringTokenizer tok = new StringTokenizer(val, ",");
          while (tok.hasMoreTokens()) {
            streamBuilder.addToken(tok.nextToken());
          }
          doc.add(new Field(fieldName, streamBuilder.buildTokenStream()));
          }
        }
        
        writer.addDocument(doc);
      }
    }

    writer.optimize();
    writer.close();
  }
}
