package anl.aida.ae.lucene;

import anl.aida.util.AIDAException;

import org.apache.lucene.document.DateTools;
import org.apache.lucene.document.Document;
import org.apache.lucene.index.Term;

import java.io.IOException;
import java.sql.*;
import java.text.ParseException;
import java.util.HashMap;
import java.util.Map;

/**
 * TermFunction implementation that inserts the term data into a database.
 * The db config info is stored in term_count_db.properties.
 *
 */
public class DBTermFunction implements TermFunction {

  private static final String DOC_FIND = "select id from documents where path = ?";
  private static final String TERM_FIND = "select id from terms where text = ? and field_id = ? and doc_id = ?";

  private static final String INSERT_FIELD = "insert into fields (name) values (?)";
  private static final String INSERT_DOC = "insert into documents (doc_date, path, title) " +
          "values (?, ?, ?)";

  // fid, did, text
  private static final String INSERT_TERM = "insert into terms (field_id, doc_id, text) values (?, ?, ?)";

  // dterm_id, date_stamp, freq, score
  private static final String INSERT_TERM_SCORE = "insert into term_scores (term_id, date_stamp, freq, score)" +
          " values (?, ?, ?, ?)";


  private Connection connection;
  private Map<String, Integer> fieldCache = new HashMap<String, Integer>();
  private Map<Integer, Integer> docCache = new HashMap<Integer, Integer>();
  private PreparedStatement findDoc, findTerm, insertField, insertDoc, insertTerm, insertScore;
  private Date date;
  private Term currentTerm;

  public DBTermFunction(java.util.Date date) {
    // util.Date to sql.Date
    this.date = new Date(date.getTime());
  }

  /**
   * Initializes the DB connection.
   */
  public void init(int docCount) throws AIDAException {
    try {
      TermCountDAO dao = new TermCountDAO();
      connection = dao.createConnection();
      connection.setAutoCommit(false);

      findDoc = connection.prepareStatement(DOC_FIND);
      findTerm = connection.prepareStatement(TERM_FIND);

      insertField = connection.prepareStatement(INSERT_FIELD, Statement.RETURN_GENERATED_KEYS);
      insertDoc = connection.prepareStatement(INSERT_DOC, Statement.RETURN_GENERATED_KEYS);
      insertTerm = connection.prepareStatement(INSERT_TERM, Statement.RETURN_GENERATED_KEYS);
      insertScore = connection.prepareStatement(INSERT_TERM_SCORE, Statement.RETURN_GENERATED_KEYS);

      Statement stmt = connection.createStatement();
      fieldCache.clear();
      ResultSet set = stmt.executeQuery("select id, name from fields");
      while (set.next()) {
        fieldCache.put(set.getString(2), set.getInt(1));
      }
      set.close();
      stmt.close();

      // DEBUG -- clear term scores for today's date
      PreparedStatement delStmt = connection.prepareStatement("delete from term_scores where date_stamp = ?");
      delStmt.setDate(1, date);
      delStmt.executeUpdate();
      connection.commit();
      delStmt.close();
      // end DEBUG

      docCache.clear();

    } catch (ClassNotFoundException e) {
      throw new AIDAException("Database driver not found.", e);
    } catch (IOException e) {
      throw new AIDAException("Database properties file not found.", e);
    } catch (SQLException e) {
      throw new AIDAException("Database error.", e);
    }
  }

  /**
   * Commits any changes to the DB.
   */
  public void done() throws AIDAException {
    try {
      connection.commit();
      connection.close();
    } catch (SQLException e) {
      throw new AIDAException("Error commiting or closing the database.", e);
    }
  }

  /**
   * Called when a new Term is encounters during iterator.
   *
   * @param term the new term.
   */
  public void newTerm(Term term) {
    currentTerm = term;
  }

  /**
   * Apply some implementor specific operation to the term and its data.
   *
   * @param docId     the numeric id of the document. This is only valid between calls to init
   *                  and done.
   * @param doc       the document the term applies to
   * @param frequency the frequency count of the term in the document
   * @param score     the score of this term for the specified document.
   */
  public void apply(int docId, Document doc, int frequency, float score) throws AIDAException {
    try {
      int termId = findTermId(currentTerm, docId, doc);
      insertScore.clearParameters();
      insertScore.setInt(1, termId);
      insertScore.setDate(2, date);
      insertScore.setInt(3, frequency);
      insertScore.setFloat(4, score);

      insertScore.executeUpdate();

    } catch (SQLException e) {
      throw new AIDAException("Database error.", e);
    } catch (ParseException e) {
      throw new AIDAException("Error parsing lucene date", e);
    }
  }

  private int findTermId(Term term, int docId, Document doc) throws SQLException, ParseException {
    int retVal = -1;
    ResultSet results = null;

    try {
      String text = term.text();
      int fieldId = findFieldId(term.field());
      int dbDocId = findDocId(docId, doc);

      findTerm.clearParameters();
      findTerm.setString(1, text);
      findTerm.setInt(2, fieldId);
      findTerm.setInt(3, dbDocId);
      results = findTerm.executeQuery();

      if (results.next()) {
        retVal = results.getInt(1);

      } else {
        results.close();
        // not found so add it
        insertTerm.clearParameters();
        insertTerm.setInt(1, fieldId);
        insertTerm.setInt(2, dbDocId);
        insertTerm.setString(3, text);
        insertTerm.executeUpdate();
        results = insertTerm.getGeneratedKeys();
        results.next();
        retVal = results.getInt(1);
      }
    } finally {
      if (results != null) results.close();
    }

    return retVal;
  }

  private int findDocId(int lucDocId, Document doc) throws SQLException, ParseException {
    int retVal = -1;
    ResultSet results = null;

    try {
      Integer id = docCache.get(lucDocId);
      if (id == null) {
        String path = doc.get(LuceneConstants.PATH_FIELD);
        findDoc.clearParameters();
        findDoc.setString(1, path);
        results = findDoc.executeQuery();
        if (results.next()) {
          retVal = results.getInt(1);
        } else {
          results.close();
          insertDoc.clearParameters();
          insertDoc.setDate(1, new java.sql.Date(DateTools.stringToTime(doc.get(LuceneConstants.TIMESTAMP_FIELD))));
          insertDoc.setString(2, path);
          insertDoc.setString(3, doc.get(LuceneConstants.TITLE_FIELD));
          insertDoc.executeUpdate();

          results = insertDoc.getGeneratedKeys();
          results.next();
          retVal = results.getInt(1);
          docCache.put(lucDocId, retVal);
        }
      } else {
        retVal = id.intValue();
      }
    } finally {
      if (results != null) results.close();
    }

    return retVal;

  }

  private int findFieldId(String field) throws SQLException {
    int retVal = -1;
    ResultSet results = null;
    try {
      Integer val = fieldCache.get(field);
      if (val == null) {
        insertField.clearParameters();
        insertField.setString(1, field);
        insertField.executeUpdate();
        results = insertField.getGeneratedKeys();
        results.next();
        retVal = results.getInt(1);
        fieldCache.put(field, retVal);
        results.close();
      } else {
        retVal = val.intValue();
      }
    } finally {
      if (results != null) results.close();
    }
    return retVal;
  }
}
