package anl.aida.ae.lucene;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.uima.cas.CAS;
import org.apache.uima.cas.Feature;
import org.apache.uima.cas.FeatureStructure;
import org.apache.uima.cas.Type;
import org.apache.uima.cas.text.AnnotationIndex;

import anl.aida.util.AIDAException;

/**
 * 
 * Default implementation of a FieldBuilder.
 *
 */
public class DefaultFieldBuilder implements FieldBuilder {

  private String fieldName;
  private List<Feature> features = new ArrayList<Feature>();
  private Type type;
  private float boost = 1f;
  private String keywordType = null;


  /**
   * Constructs a DefaultFieldBuilder.
   *
   * @param fieldName the name of the field to build
   * @param type      the annotation type that contains the features that
   *                  become the content of the field
   * @param feature   the feature whose values become the content of the field
   */
  public DefaultFieldBuilder(String fieldName, Type type, Feature feature) {
    this(fieldName, type, feature, 1.0f);
  }

  /**
   * Constructs a DefaultFieldBuilder.
   *
   * @param fieldName the name of the field to build
   * @param type      the annotation type that contains the features that
   *                  become the content of the field
   * @param feature   the feature whose values become the content of the field
   * @param boost the lucene boost value for the created field
   */
  public DefaultFieldBuilder(String fieldName, Type type, Feature feature, float boost) {
    this.fieldName = fieldName;
    this.type = type;
    this.boost = boost;
    addFeature(feature);
  }
  
  
  /**
   * If this field holds keywords, then this gets
   * the keyword type (e.g. "BioAgent"), otherwise
   * this returns null.
   * 
   * @return the keyword type, if applicable, otherwise null.
   */
  @Override
  public String getKeywordType() {
    return keywordType;
  }
  
  /**
   * Sets the keyword type for this field builder.
   * 
   * @param type the keyword type
   */
  public void setKeywordType(String type) {
    keywordType = type;
  }

  /* (non-Javadoc)
   * @see anl.aida.ae.lucene.FieldBuilder#getFieldName()
   */
  @Override
  public String getFieldName() {
    return fieldName;
  }

  /**
   * Adds a feature. The values of the added feature will be
   * added to the field's contents.
   *
   * @param feature the feature to add
   */
  public void addFeature(Feature feature) {
    if (feature != null)
      features.add(feature);
  }
  
  boolean isValid() {
    return features.size() > 0 && type != null;
  }

  /**
   * Builds a Field in the specified Document using values from the specified
   * CAS.
   * 
   * @param cas
   *          the CAS to use to build the field
   * @param doc
   *          the doc to add the fields to
   */
  public Field build(CAS cas, Document doc, Keywords keywords) throws AIDAException {

    TokenStreamBuilder builder = new TokenStreamBuilder();
    AnnotationIndex index = cas.getAnnotationIndex(type);
    for (Iterator iter = index.iterator(); iter.hasNext();) {
      FeatureStructure fs = (FeatureStructure) iter.next();
      for (Feature feature : features) {
	String val = fs.getFeatureValueAsString(feature);
	if (val != null) {
	  builder.addToken(val);
	  if (keywordType != null) {
	    keywords.addKeyword(keywordType, val);
	  }
	}
      }
    }

    Field field = null;
    if (builder.tokenCount() > 0) {
      // not stored but is indexed.
      field = new Field(fieldName, builder.buildTokenStream());
      field.setBoost(boost);
      doc.add(field);
    }
    return field;
  }
}
