package anl.aida.ae.lucene;

import java.io.StringReader;
import java.util.Calendar;
import java.util.GregorianCalendar;

import org.apache.lucene.document.DateTools;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.uima.cas.CAS;
import org.apache.uima.cas.Type;
import org.apache.uima.cas.TypeSystem;
import org.apache.uima.jcas.cas.FSArray;

import anl.aida.types.Date;
import anl.aida.types.DocumentMetaData;
import anl.aida.types.Name;

/**
 * Builds a Lucene document from a CAS.
 *
 */
public class DocumentBuilder {

  private static final String META_DATA_TYPE = "anl.aida.types.DocumentMetaData";
  
  private Calendar cal = GregorianCalendar.getInstance();

  private Type mdType;
  private float titleBoost = 1f;

  public DocumentBuilder(TypeSystem system) {
    mdType = system.getType(META_DATA_TYPE);
  }

  /**
   * Gets the title boost.
   *
   * @return the title boost.
   */
  public float getTitleBoost() {
    return titleBoost;
  }

  /**
   * Sets the title boost value.
   *
   * @param titleBoost the new title boost value
   */
  public void setTitleBoost(float titleBoost) {
    this.titleBoost = titleBoost;
  }

  /**
   * Builds a Lucene Document based on the content of the CAS.
   * 
   * @param cas
   *          the CAS to build the doc from.
   * @return the built Document.
   */
  public Document build(CAS cas) {

    // make a new, empty document
    Document doc = new Document();
    DocumentMetaData data = (DocumentMetaData) cas.getAnnotationIndex(mdType).iterator().next();

    // Add the path of the file as a field named "path". Use a field that is
    // indexed (i.e. searchable), but don't tokenize the field into words.
    doc.add(new Field(LuceneConstants.PATH_FIELD, data.getLocation(), Field.Store.YES,
	Field.Index.NOT_ANALYZED));

    Field field = new Field(LuceneConstants.TITLE_FIELD, data.getTitle(), Field.Store.YES,
	Field.Index.ANALYZED);
    field.setBoost(titleBoost);
    doc.add(field);

    FSArray authors = data.getAuthors();
    StringBuilder builder = new StringBuilder();

    for (int i = 0, n = authors.size(); i < n; i++) {
      Name name = (Name) authors.get(i);

      if (i > 0)
	builder.append(", ");
      builder.append(name.getLastName());
      builder.append(" ");
      builder.append(name.getFirstName());
      String initials = name.getInitials();
      if (initials.length() > 0) {
	builder.append(" ");
	builder.append(initials);
	builder.append(".");
      }
    }
    doc.add(new Field(LuceneConstants.AUTHOR_FIELD, builder.toString(), Field.Store.YES,
	Field.Index.ANALYZED));

    Date date = data.getDate();
    // hack to handle bad date values
    // int month = date.getMonth() == -1 ? 12 : date.getMonth();
    // int day = date.getDay() == -1 ? 31 : date.getDay();
    cal.set(date.getYear(), date.getMonth() - 1, date.getDay());
    long time = cal.getTimeInMillis();

    // Add the last modified date of the file a field named "modified". Use
    // a field that is indexed (i.e. searchable), but don't tokenize the field
    // into words.
    doc.add(new Field(LuceneConstants.TIMESTAMP_FIELD, DateTools.timeToString(time,
	DateTools.Resolution.DAY), Field.Store.YES, Field.Index.ANALYZED));

    // Add the contents of the file to a field named "contents". Specify a
    // Reader,
    // so that the text of the file is tokenized and indexed, but not stored.

    // THIS ONLY INDEXES THE FIRST 10K TERMS OF THE DOCUMENT
    doc.add(new Field(LuceneConstants.CONTENTS_FIELD, new StringReader(cas.getDocumentText())));
    
    // adds the source field to the index
    doc.add(new Field(LuceneConstants.SOURCE_FIELD, data.getSource(), Field.Store.YES,
	    Field.Index.NOT_ANALYZED));

    // return the document
    return doc;
  }
}
