package anl.aida.ae.lucene;

import org.apache.uima.UIMAFramework;
import org.apache.uima.cas.TypeSystem;
import org.apache.uima.cas.Type;
import org.apache.uima.cas.Feature;
import org.apache.uima.util.Level;

import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;
import javax.xml.stream.events.EndElement;
import javax.xml.stream.events.Attribute;
import javax.xml.namespace.QName;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.List;

/**
 * Creates FieldBuilders from an xml description. A sample file looks like:
 *
 * <code>
 * <fields>

    <field name="Lookup.majorType" boost="1.5">
        <feature>anl.aida.types.Lookup:majorType</feature>
        ...
    </field>

    <field name="Lookup.minorType">
        <feature>anl.aida.types.Lookup:majorType"</feature>
        ...
    </field>
</fields>
 *
 * </code>
 *
 * The values of the specified annotation features will become the indexed content of that
 * lucene field.
 *
 */
public class FieldBuilderFactory {

  private String mappingFile;
  private String name, keywordType;
  private DefaultFieldBuilder builder;
  private Float boost = null;
  private float titleBoost = 1f;

  /**
   * Creates a FieldBuilderFactory using the specified file. The file
   * describes the
   *
   * @param file the xml field builder description file
   */
  public FieldBuilderFactory(String file) {
    this.mappingFile = file;
  }

  /**
   * Creates a list of FieldBuilders using the specified TypeSystem.
   *
   * @param system the TypeSystem to use to create the fields
   *
   * @return the created list
   * @throws XMLStreamException if there is an error parsing the field builder description file
   * @throws FileNotFoundException if the field builder description file cannot be found
   */
  public List<FieldBuilder> createBuilders(TypeSystem system) throws XMLStreamException, FileNotFoundException {
    List<FieldBuilder> builders = new ArrayList<FieldBuilder>();

    XMLInputFactory factory = XMLInputFactory.newInstance();
    XMLEventReader reader = factory.createXMLEventReader(new BufferedReader(new FileReader(mappingFile)));
    while (reader.hasNext()) {
      XMLEvent evt = reader.nextEvent();
      if (evt.isStartElement()) {

        StartElement elmt = evt.asStartElement();
        String qName = elmt.getName().getLocalPart();

        if (qName.equals("field")) {
          name = elmt.getAttributeByName(new QName("name")).getValue();
          Attribute aBoost = elmt.getAttributeByName(new QName("boost"));
          if (aBoost != null) {
            try {
              boost = Float.valueOf(aBoost.getValue());
            } catch (NumberFormatException ex) {
              throw new XMLStreamException("Error parsing boost value.", ex);
            }
          }
          
          Attribute kType = elmt.getAttributeByName(new QName("keywordSet"));
          if (kType != null) keywordType = kType.getValue();
          

        } else if (qName.equals("feature")) {
          if (builder == null) createBuilder(system, reader.getElementText());
          else builder.addFeature(system.getFeatureByFullName(reader.getElementText()));
        } else if (qName.equals("fields")) {
          Attribute tBoost = elmt.getAttributeByName(new QName("title_boost"));
          if(tBoost != null) {
            try {
              titleBoost = Float.parseFloat(tBoost.getValue());
            } catch (NumberFormatException ex) {
              throw new XMLStreamException("Error parsing title_boost attribute.", ex);
            }
          }
        }


      } else if (evt.isEndElement()) {
        EndElement elmt = evt.asEndElement();
        String qName = elmt.getName().getLocalPart();
        if (qName.equals("field")) {
          if (builder.isValid()) builders.add(builder);
          builder = null;
          name = null;
          boost = null;
        }
      }
    }

    return builders;
  }

  /**
   * Gets the boost value for the title field.
   *
   * @return the boost value for the title field.
   */
  public float getTitleBoost() {
    return titleBoost;
  }

  private void createBuilder(TypeSystem system, String featureName) {
    int index = featureName.indexOf(":");
    String typeStr = featureName.substring(0, index);
    Type type = system.getType(typeStr);
    if (type == null) {
      UIMAFramework.getLogger().log(Level.WARNING, "Type '" + typeStr + "' not found.");
    }
    Feature feature = system.getFeatureByFullName(featureName);
    if (feature == null) {
      UIMAFramework.getLogger().log(Level.WARNING, "Feature '" + featureName + "' not found.");
    }
    float fBoost = boost == null ? 1f : boost.floatValue();
    builder = new DefaultFieldBuilder(name, type, feature, fBoost);
    builder.setKeywordType(keywordType);
  }
}
