package anl.aida.ae.lucene;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Reader;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.List;

import org.apache.lucene.analysis.Token;
import org.apache.lucene.analysis.TokenStream;
import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.index.Term;
import org.apache.uima.UIMAFramework;
import org.apache.uima.UimaContext;
import org.apache.uima.analysis_component.CasAnnotator_ImplBase;
import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.cas.CAS;
import org.apache.uima.cas.TypeSystem;
import org.apache.uima.resource.ResourceInitializationException;
import org.apache.uima.util.Level;
import org.apache.uima.util.Logger;

import anl.aida.util.AIDAUtils;
import anl.aida.util.AndFilter;
import anl.aida.util.DateRangeProducer;
import anl.aida.util.FileUtilities;
import anl.aida.util.Filter;
import anl.aida.util.KeywordFilter;
import anl.aida.util.AIDAException;

/**
 * UIMA CAS Consumer that writes that writes a CAS' document contents as well
 * as mapped annotations to a lucene index. The annotations to map are
 * specified in an XML mapping file.
 *
 */
public class LuceneConsumer extends CasAnnotator_ImplBase {
  
  private static final String MESSAGE_DIGEST = "anl.aida.ae.lucene.LuceneConsumer_Messages";

  public final static String MAPPING_FILE = "anl.aida.ae.lucene.IndexDescriptor";
  public final static String INDEX_DIR = "anl.aida.ae.lucene.IndexDirectory";
  public final static String CLEAR_INDEX = "anl.aida.ae.lucene.ClearIndex";
  public static final String CACHED_CONTENT_DIR = "anl.aida.ae.lucene.CachedContentDir";
  public static final String CREATE_CACHE = "anl.aida.ae.lucene.CreateCache";
  public static final String CACHE_OUTPUT_DIR = "anl.aida.ae.lucene.OutputDir";
  public static final String CACHE_START_DATE = "anl.aida.ae.lucene.CacheStartDate";
  public static final String CACHE_LENGTH_DAYS = "anl.aida.ae.lucene.CacheLengthDays";
  public static final String KEYWORD_FILTER = "anl.aida.ae.lucene.KeywordFilter";
  public static final String CACHE_TS_INTERVAL_DAYS = "anl.aida.ae.lucene.CacheTSIntervalDays";
 
  private static class CacheFieldBuilder implements FieldBuilder {
    
    private File cacheDir;
    
    @Override
    public String getFieldName() {
      return LuceneConstants.CACHED_CONTENT_PATH;
    }

    public CacheFieldBuilder(String dir) {
      cacheDir = new File(dir);
      if (cacheDir.exists()) {
	FileUtilities.delete(cacheDir);
      }
      cacheDir.mkdirs();
    }
    
    @Override
    public String getKeywordType() {
      return null;
    }
    
    @Override
    public Field build(CAS cas, Document doc, Keywords keywords) throws AIDAException {
      try {
	String timestamp = doc.get(LuceneConstants.TIMESTAMP_FIELD);
	File dir = new File(cacheDir, timestamp);
	if (!dir.exists()) {
	  dir.mkdirs();
	}

	File cache = new File(dir, dir.list().length + ".txt");
	BufferedWriter writer = new BufferedWriter(new FileWriter(cache));
	String content = cas.getDocumentText();
	writer.write(content, 0, content.length());
	writer.close();

	doc.add(new Field(LuceneConstants.CACHED_CONTENT_PATH, cache.getAbsolutePath(),
	    Field.Store.YES, Field.Index.NO));
      } catch (IOException ex) {
	throw new AIDAException(ex);
      }
      
      return null;
    }
  }

  private DocumentBuilder docBuilder;
  private IndexWriter writer;
  private List<FieldBuilder> builders;
  private FieldBuilderFactory factory;
  private String cacheDir, indexDir, startDate, outputDir;
  private int cacheLengthDays, cacheIntervalDays;
  private boolean createCache = false;
  private Logger logger;
  private Token token = new Token();
  private Keywords keywords = new Keywords();
  private Boolean filterCache=false;
  

  @Override
  public void typeSystemInit(TypeSystem typeSystem) throws AnalysisEngineProcessException {
    super.typeSystemInit(typeSystem);

    docBuilder = new DocumentBuilder(typeSystem);

    try {
      if (factory != null) {
        builders = factory.createBuilders(typeSystem);
        docBuilder.setTitleBoost(factory.getTitleBoost());
      } else builders = new ArrayList<FieldBuilder>();
      
      builders.add(new CacheFieldBuilder(cacheDir));
      
      for (FieldBuilder builder : builders) {
	if (builder.getKeywordType() != null) {
	  keywords.addKeywordField(builder.getFieldName());
	}
      }

    } catch (Exception ex) {
      throw new AnalysisEngineProcessException(MESSAGE_DIGEST, "field_builder_error",
              new Object[]{}, ex);
    }
  }
  
  private void deleteFile(File dir) {
    for (File file : dir.listFiles()) {
      if (file.isDirectory()) deleteFile(file);
      file.delete();
    }
  }

  @Override
  public void initialize(UimaContext context) throws ResourceInitializationException {
    super.initialize(context);
    keywords.clearKeywords();
    createCache = (Boolean)context.getConfigParameterValue(CREATE_CACHE);
    filterCache = (Boolean)context.getConfigParameterValue(KEYWORD_FILTER);
    cacheDir = (String)context.getConfigParameterValue(CACHED_CONTENT_DIR);
    indexDir = (String) context.getConfigParameterValue(INDEX_DIR);
    outputDir = (String)context.getConfigParameterValue(CACHE_OUTPUT_DIR);
    startDate = (String)context.getConfigParameterValue(CACHE_START_DATE);
    boolean clearIndex = (Boolean) context.getConfigParameterValue(CLEAR_INDEX);
    String file = (String) context.getConfigParameterValue(MAPPING_FILE);
    cacheLengthDays = (Integer)context.getConfigParameterValue(CACHE_LENGTH_DAYS);
    cacheIntervalDays = (Integer)context.getConfigParameterValue(CACHE_TS_INTERVAL_DAYS);
    
    logger = context.getLogger();
    logger.log(Level.INFO, String.format("Initializing LuceneConsumer:%n\tMappingFile: %s%n" +
	"\tIndex Directory: %s%n" +
	"\tClear Index: %b%n" +
	"\tCreate Cache: %b%n" +
	"\tCache Start Date: %s%n" +
	"\tCache Output Dir: %s%n" +
	"\tCache Duration: %d (days)%n" +
	"\tCache Timestep Interval: %d (days)%n" +
	"\tTmpContentDir: %s",
	(file == null ? "" : file), indexDir, clearIndex, createCache, startDate, outputDir, cacheLengthDays,
	cacheIntervalDays, cacheDir));
    
    File indexDir = new File(this.indexDir);
    if (!indexDir.exists()) {
      indexDir.mkdirs();
    } else {
      if (clearIndex) {
	FileUtilities.delete(indexDir);
	File cache = new File(cacheDir);
        if (cache.exists()) FileUtilities.delete(cache);
      }
    }

    try {
      writer = new IndexWriter(indexDir, new StandardAnalyzer(), IndexWriter.MaxFieldLength.UNLIMITED);
    } catch (IOException e) {
      throw new ResourceInitializationException(MESSAGE_DIGEST, "lucene_writer_error",
              new Object[]{}, e);
    }

    
    if (file != null) factory = new FieldBuilderFactory(file);
  }

  public void process(CAS cas) throws AnalysisEngineProcessException {
    Document doc = docBuilder.build(cas);
    try {
      for (FieldBuilder builder : builders) {
	Field field = builder.build(cas, doc, keywords);
	String keywordType = builder.getKeywordType();
	if (keywordType != null) {
	  addKeyword(field, keywordType, doc);
	}
      }
      writer.addDocument(doc);
      
    } catch (IOException e) {
      UIMAFramework.getLogger().log(Level.SEVERE, e.getMessage(), e);
      throw new AnalysisEngineProcessException(MESSAGE_DIGEST, "lucene_writer_error",
	  new Object[] {}, e);
    } catch (AIDAException ex) {
      UIMAFramework.getLogger().log(Level.SEVERE, ex.getMessage(), ex);
      throw new AnalysisEngineProcessException(MESSAGE_DIGEST, "field_builder_error",
	  new Object[] {}, ex);
    }
  }
  
  private void addKeyword(Field field, String keywordType, Document doc) throws IOException {
    if (field != null && createCache) {
      
      
      StringBuilder builder = new StringBuilder();
      boolean addD = false;
      String val = field.stringValue();
      if (val == null) {
	TokenStream stream = field.tokenStreamValue();
	while (stream.next(token) != null) {
	  if (addD) builder.append(",");
	  else addD = true;
	  builder.append(token.term());
	}
	stream.reset();
	
      } else {
	builder.append(val);
      }
      
      doc.add(new Field(field.name() + "_cache", builder.toString(), Field.Store.YES, Field.Index.NO));
    }
  }

  @Override
  public void collectionProcessComplete() throws AnalysisEngineProcessException {
    super.collectionProcessComplete();
    try {
      if (writer != null) {
	logger.log(Level.INFO, "Optimizing and closing lucene index");
	writer.optimize();
	writer.close();
      }
    } catch (IOException e) {
      throw new AnalysisEngineProcessException(MESSAGE_DIGEST, "lucene_writer_error",
	  new Object[] {}, e);
    }
    
    try {
      if (createCache) createCache();
    } catch (Exception e) {
      UIMAFramework.getLogger().log(Level.SEVERE, e.getMessage(), e);
      throw new AnalysisEngineProcessException(MESSAGE_DIGEST, "cache_writer_error",
	  new Object[] {}, e);
    } 
  }
  
  private void createCache() throws IOException, AIDAException, ParseException {
    Date start = AIDAUtils.getDate(startDate);
    
    Calendar cal = GregorianCalendar.getInstance();
    cal.setTime(start);
    UIMAFramework.getLogger(getClass()).log(Level.INFO, String.format("Creating cache from %s into %s starting at %d %d %d",
	// + 1 for month because months start at 0
	indexDir, outputDir, cal.get(Calendar.YEAR), cal.get(Calendar.MONTH) + 1, cal.get(Calendar.DAY_OF_MONTH)));

    IndexReader reader = IndexReader.open(indexDir);
    // 60 days, 5 day intervals
    DateRangeProducer producer = new DateRangeProducer(start, cacheLengthDays, cacheIntervalDays);
    Filter<Term>filter=null;
    if(!filterCache)
    	filter = new AndFilter<Term>(new StopWordFilter(), new WordFilter(2));
    else
    	filter = new KeywordFilter(keywords);
    CacheCreator creator = new CacheCreator(producer, filter, reader);
    creator.createCache(new File(outputDir), new CachedContentReader(), keywords);
  }
}

class CachedContentReader implements ReaderProducer {

  /* (non-Javadoc)
   * @see anl.aida.ae.lucene.ReaderProducer#getReader(org.apache.lucene.document.Document)
   */
  @Override
  public Reader getReader(Document doc) throws IOException {
    return new BufferedReader(new FileReader(doc.get(LuceneConstants.CACHED_CONTENT_PATH)));
  }
}
