package anl.aida.ae.lucene;

import java.io.IOException;
import java.net.MalformedURLException;

import org.apache.lucene.document.Document;
import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.Term;

import anl.aida.core.STCBuilder;
import anl.aida.core.ScoredTermsCollection;
import anl.aida.util.DateRange;
import anl.aida.util.Filter;
import anl.aida.util.AIDAException;
import cern.colt.map.OpenIntIntHashMap;

/**
 * Creates ScoredTermCollections from a LuceneIndex.
 *
 */
public class STCCreator implements TermFunction {

  private STCBuilder builder;

  private Integer termIndex = 0;
  private Filter<Term> filter;
  private DateRange range;

  private boolean processTerm = false;
  // maps between the lucene indices that are passed
  // in and the indices written to the cache file.
  // this allows us to have consecutive indices in the cache file
  // when that may not be the case with the lucene indices
  private OpenIntIntHashMap indexMap = new OpenIntIntHashMap();
  private int index = 0;

  /**
   * Creates a ScoredTermsCollection.
   *
   * @param reader the reader used to read the lucene index
   * @param range  the date range for this ScoredTermsCollection
   * @return the created ScoredTermsCollection.
   * @throws AIDAException       if there is an error creating the collecion
   * @throws java.io.IOException if there is an error reading the index
   */
  public ScoredTermsCollection createCollection(IndexReader reader, DateRange range) throws AIDAException, IOException {
    return createCollection(reader, new Filter<Term>() {
      public boolean evaluate(Term term) {
        return true;
      }
    }, range);
  }

  /**
   * Creates a ScoredTermsCollection.
   *
   * @param reader the reader used to read the index.
   * @param filter the terms ultimately included in the collection must pass this filter.
   * @param range  the date range for this ScoredTermsCollection
   * @return the created ScoredTermsCollection.
   * @throws AIDAException       if there is an error creating the collecion
   * @throws java.io.IOException if there is an error reading the index
   */
  public ScoredTermsCollection createCollection(IndexReader reader, Filter<Term> filter, DateRange range) throws AIDAException,
          IOException {
    this.filter = filter;
    this.range = range;

    TermScorer scorer = new TermScorer();
    scorer.apply(reader, this);
    return builder.createSTC(true);
  }

  /**
   * Called once prior to calling apply for any terms, documents.
   *
   * @param docCount the number of documents in the index.
   * @throws anl.aida.util.AIDAException if there is an error during initialization.
   */
  public void init(int docCount) throws AIDAException {
    this.builder = new STCBuilder(range);
    indexMap.clear();
  }

  /**
   * Called when a new Term is encounters during iterator.
   *
   * @param term the new term.
   */
  public void newTerm(Term term) {
    processTerm = filter.evaluate(term);

    if (processTerm) {
      termIndex = builder.addTerm(term.text());
    }
  }

  /**
   * Apply some implementor specific operation to the term and its data. The parameters
   * of this call apply to the term passed in the prior call to {@link #newTerm(org.apache.lucene.index.Term)} .
   *
   * @param docId     the numeric id of the document. This is only valid between calls to init
   *                  and done.
   * @param doc       the document the term applies to
   * @param frequency the frequency count of the term in the document
   * @param score     the score of this term for the specified document.
   * @throws anl.aida.util.AIDAException if there is an error.
   */
  public void apply(int docId, Document doc, int frequency, float score) throws AIDAException {
    if (processTerm) {
      
      int cacheDocId = 0;
      if (!indexMap.containsKey(docId)) {
	cacheDocId = index;
	index++;
	indexMap.put(docId, cacheDocId);
      } else {
	cacheDocId = indexMap.get(docId);
      }
      
      try {
        if (score != 0) {
          if (!builder.hasDocument(cacheDocId)) {
            String title = doc.getField(LuceneConstants.TITLE_FIELD).stringValue();
            String path = doc.getField(LuceneConstants.PATH_FIELD).stringValue();
            builder.addDocument(cacheDocId, title, path);
          }
          // by default the builder will sum scores for the
          // same term index
          builder.addScore(cacheDocId, termIndex, frequency, score);
        }

      } catch (MalformedURLException e) {
        throw new AIDAException("Error creating TermsCollection.", e);
      }
    }
  }


  /**
   * Called after apply has been called for all the terms.
   *
   * @throws anl.aida.util.AIDAException if there is an error during the "done" phase.
   */
  public void done() throws AIDAException {
  }


}
