package anl.aida.ae.lucene;

import anl.aida.util.AIDAException;

import org.apache.lucene.document.Document;
import org.apache.lucene.index.Term;


/**
 * Interface for classes that manipulate Lucene Term, Document data. This class
 * works in conjunction with the TermScorer. Methods on this interface will be called
 * by the TermScorer as it iterates through the terms and documents. The TermScorer
 * iteration looks like:
 * <pre>
 * {@link #init(int)}
 * for all terms:
 *    {@link #newTerm(org.apache.lucene.index.Term)}
 *    for all documents:
 *      calc score
 *      {@link #apply(int, org.apache.lucene.document.Document, int, float)}
 * {@link #done()}
 * <p/>
 * Note that a Term contains a field and so the actual term text may occur more than once.
 *
 */
public interface TermFunction {

  /**
   * Called once prior to calling apply for any terms, documents.
   *
   * @param docCount the number of documents in the index.
   * @throws AIDAException if there is an error during initialization.
   */
  void init(int docCount) throws AIDAException;

  /**
   * Called when a new Term is encounters during iterator.
   *
   * @param term the new term.
   */
  void newTerm(Term term);

  /**
   * Apply some implementor specific operation to the term and its data. The parameters
   * of this call apply to the term passed in the prior call to {@link #newTerm(org.apache.lucene.index.Term)} .
   *
   * @param docId     the numeric id of the document. This is only valid between calls to init
   *                  and done.
   * @param doc       the document the term applies to
   * @param frequency the frequency count of the term in the document
   * @param score     the score of this term for the specified document.
   * @throws AIDAException if there is an error.
   */
  void apply(int docId, Document doc, int frequency, float score) throws AIDAException;

  /**
   * Called after apply has been called for all the terms.
   *
   * @throws AIDAException if there is an error during the "done" phase.
   */
  void done() throws AIDAException;
}
