package anl.aida.ae.lucene;

import anl.aida.util.AIDAException;

import org.apache.lucene.index.IndexReader;
import org.apache.lucene.index.TermDocs;
import org.apache.lucene.index.TermEnum;
import org.apache.lucene.search.DefaultSimilarity;
import org.apache.lucene.search.IndexSearcher;
import org.apache.lucene.search.Similarity;

import java.io.IOException;
import java.util.HashSet;
import java.util.Set;

/**
 * Iterates through Terms in a Lucene Index
 * an applies a function to the term data.
 *
 */
public class TermScorer {

  private static final int CACHE_SIZE = 32;
  private static final int BUFF_SIZE = 32;

  // buffered doc ids and frequency counts.
  private int[] docs = new int[BUFF_SIZE];
  private int[] freqs = new int[BUFF_SIZE];

  // cache of tf scores for frequency counts up to
  // the CACHE_SIZE.
  private float[] scoreCache = new float[CACHE_SIZE];
  private Similarity similarity = new DefaultSimilarity();

  private static Set<String> unscoredFields = new HashSet<String>();

  static {
    unscoredFields.add(LuceneConstants.PATH_FIELD);
    unscoredFields.add(LuceneConstants.TIMESTAMP_FIELD);
    unscoredFields.add(LuceneConstants.AUTHOR_FIELD);
  }

  /**
   * Creates a new TermIterator.
   */
  public TermScorer() {

    for (int i = 0; i < CACHE_SIZE; i++) {
      // calculate and cache tf scores for frequency counts
      // up to the CACHE_SIZE
      scoreCache[i] = similarity.tf(i);
    }
  }


  /**
   * Iterates over all the Terms in the index and applies the
   * specified function to them. The order of iteration is:
   * <p/>
   * <pre>
   * func.init();
   * for all Terms:
   *     func.newTerm();
   *     for all Documents:
   *        func.apply();
   * <p/>
   * func.done();
   * </pre>
   * <p/>
   * Note that terms are distinguished by text <b>and</b> field. So two terms
   * may have the same text but be indifferent fields.
   *
   * @param reader the reader to read the index with.
   * @param func   the function to apply to the Terms.
   * @throws IOException   if there is an error iterating over the index
   * @throws AIDAException if there is an error with the TermFunction
   */
  public void apply(IndexReader reader, TermFunction func) throws IOException, AIDAException {

    IndexSearcher searcher = null;

    try {
      func.init(reader.numDocs());
      searcher = new IndexSearcher(reader);
      TermEnum terms = reader.terms();

      while (terms.next()) {
        org.apache.lucene.index.Term term = terms.term();
        func.newTerm(term);
        String field = term.field();
        if (!unscoredFields.contains(field)) {
          byte[] norms = reader.norms(field);
          float idf = similarity.idf(term, searcher);
          TermDocs termDocs = reader.termDocs(term);
          int max = termDocs.read(docs, freqs);

          while (max != 0) {
            for (int index = 0; index < max; index++) {
              int doc = docs[index];
              int freq = freqs[index];
              float tf = freq < CACHE_SIZE ? scoreCache[freq] : similarity.tf(freq);
              float fieldNorm = norms == null ? 0f : Similarity.decodeNorm(norms[doc]);
              float score = tf * idf * fieldNorm;

              func.apply(doc, reader.document(doc), freq, score);
            }

            max = termDocs.read(docs, freqs);
          }
          termDocs.close();
        }
      }
      terms.close();

      func.done();
    } finally {
      if (searcher != null) searcher.close();
    }


  }
}
