package anl.aida.core;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import anl.aida.util.DateRange;
import anl.aida.util.IntIterator;
import anl.aida.util.AIDAException;
import anl.aida.util.ObjectArrayIter;
import cern.colt.function.IntObjectProcedure;
import cern.colt.list.IntArrayList;
import cern.colt.map.OpenIntObjectHashMap;

/**
 * Builds an ScoredTermsCollection given MIFSTerms and Documents.
 *
 */
public class STCBuilder {
  private STCollection collection;

  private static class TermImpl implements MIFSTerm {

    private String text;
    private int maxF, minF, docCount;
    private float maxS, minS;

    public int getDocumentCount() {
      return docCount;
    }

    public float getMinScore() {
      return minS;
    }

    public float getMaxScore() {
      return maxS;
    }

    public int getMinFrequency() {
      return minF;
    }

    public int getMaxFrequency() {
      return maxF;
    }

    public String getText() {
      return text;
    }

    void setFreqs(int freq) {
      minF = Math.min(freq, minF);
      maxF = Math.max(freq, maxF);
    }

    void setScores(float score) {
      minS = Math.min(score, minS);
      maxS = Math.max(score, maxS);
    }

    public String toString() {
      return text + ": " + docCount + " (" + minF + ", " + maxF + "), (" + minS + ", " + maxS + ")";
    }
  }

  private static class FreqScore {
    int freq;
    float score;

    private FreqScore(int freq, float score) {
      this.freq = freq;
      this.score = score;
    }
  }

  // struct holding doc and scores
  private static class DocScores {
    MIFSDocument doc;
    // map of term index to float score
    Map<Integer, FreqScore> scores;
  }

  private static class DSArrayIter implements Iterable<MIFSDocument>, Iterator<MIFSDocument> {

    private OpenIntObjectHashMap docs;
    private int idx = 0;

    DSArrayIter(OpenIntObjectHashMap docs) {
      // need to sort 
      this.docs = docs;
    }

    public boolean hasNext() {
      return idx < docs.size();
    }

    public Iterator<MIFSDocument> iterator() {
      return this;
    }

    public MIFSDocument next() {
      MIFSDocument doc = ((DocScores)docs.get(idx)).doc;
      idx++;
      return doc;
    }

    public void remove() {
      throw new UnsupportedOperationException();
    }
  }

  private static class DIDIterator implements IntIterator {

    private Integer termID;
    private OpenIntObjectHashMap docs;
    private IntArrayList docIds;
    private int next = -1;
    private int max;

    private DIDIterator(Integer termID, OpenIntObjectHashMap docs) {
      this.termID = termID;
      this.docs = docs;
      docIds = docs.keys();
      max = docIds.size();
      next = findNext();
    }

    private int findNext() {
      for (int i = next + 1; i < max; i++) {
	int docId = docIds.get(i);
        DocScores scores = (DocScores)docs.get(docId);
        if (scores.scores.containsKey(termID)) {
          next = i;
          return i;
        }
      }

      next = max;
      return next;
    }

    public boolean hasNext() {
      return next != max;
    }

    public int next() {
      int tmp = next;
      next = findNext();
      return docIds.get(tmp);
    }
  }

  private static class STCollection implements ScoredTermsCollection {


    private TermImpl[] terms;
    private OpenIntObjectHashMap docs = new OpenIntObjectHashMap();
    private DateRange dateRange;


    /**
     * Gets an iterable over all the terms in this collection.
     *
     * @return an iterable over all the terms in this collection.
     */
    public Iterable<MIFSTerm> terms() {
      return new ObjectArrayIter<MIFSTerm>(terms);
    }

    /**
     * Returns an iterator over a set of elements of type T.
     *
     * @return an Iterator.
     */
    @Override
    public Iterator<MIFSDocument> iterator() {
      return documents().iterator();
    }

    /**
     * Gets an iterable over all the documents in this collection.
     *
     * @return an iterable over all the documents in this collection.
     */
    public Iterable<MIFSDocument> documents() {
      return new DSArrayIter(docs);
    }

    /**
     * Gets the number of terms in the collection.
     *
     * @return the number of terms in the collection.
     */
    public int getTermCount() {
      return terms.length;
    }

    /**
     * Gets the term at the specified index.
     *
     * @param index the index of the term
     * @return the term at the specified index.
     */
    public MIFSTerm getTerm(int index) {
      return terms[index];
    }

    /**
     * Gets the index for the specified term.
     *
     * @param term the term to get the index for
     * @return the index for the specified term, or -1 if the term
     *         is not found.
     */
    public int getTermIndex(String term) {
      for (int i = 0; i < terms.length; i++) {
        if (terms[i].getText().equals(term)) return i;
      }

      return -1;
    }

    /**
     * Gets the number of documents in the collection.
     *
     * @return the number of documents in the collection.
     */
    public int getDocumentCount() {
      return docs.size();
    }

    /**
     * Gets the document at the specified index.
     *
     * @param index the index of the document
     * @return Gets the document at the specified index.
     */
    public MIFSDocument getDocument(int index) {
      DocScores score = (DocScores)docs.get(index);
      return score == null ? null : score.doc;
    }

    /**
     * Gets the frequency of the specified term in the specified document.
     *
     * @param docIndex  the index of the document
     * @param termIndex the index of the term
     * @return the frequency of the specified term in the specified document.
     */
    public int getFrequency(int docIndex, int termIndex) {
      DocScores ds = (DocScores)docs.get(docIndex); 
      if (ds != null)  {
	FreqScore fs = ds.scores.get(termIndex);
	return fs == null ? 0 : fs.freq;
      }
      return 0;
    }

    /**
     * Gets the score of the specified term in the specified document.
     *
     * @param docIndex  the index of the document
     * @param termIndex the index of the term
     * @return the score of the specified term in the specified document.
     */
    public float getScore(int docIndex, int termIndex) {
      DocScores ds = (DocScores)docs.get(docIndex); 
      if (ds != null)  {
	FreqScore fs = ds.scores.get(termIndex);
	return fs == null ? 0 : fs.score;
      }
      return 0;
    }

    /**
     * Gets an iterator of document indices of docs that
     * contain a Term with the specified index.
     *
     * @param termIndex the term index
     * @return n iterator of document indices of docs that
     *         contain a Term with the specified index.
     */
    public IntIterator getDocumentsFor(int termIndex) {
      return new DIDIterator(termIndex, docs);
    }

    /**
     * Gets the date range (inclusive) covered by this
     * ScoredTermsCollection.
     *
     * @return the date range (inclusive) covered by this
     *         ScoredTermsCollection.
     */
    public DateRange getRange() {
      return dateRange;
    }
    
    class SDCProcedure implements IntObjectProcedure {
      
      private ScoredDocumentCallback callback;
      
      public SDCProcedure(ScoredDocumentCallback callback) {
	this.callback = callback;
      }
      
      public boolean apply(int first, Object second) {
	DocScores ds = (DocScores)second;
	callback.newDocument(first, ds.doc);
        for (int j = 0, k = terms.length; j < k; j++) {
          TermImpl term = terms[j];
          FreqScore val = ds.scores.get(j);
          float score = val == null ? 0f : val.score;
          int freq = val == null ? 0 : val.freq;
          callback.apply(j, term, freq, score);
        }
	return true;
      }
    };
   
    /**
     * Iterates through the collection in document order calling
     * the relevant methods in the callback. In document order,
     * each document is processed in order such that all the term
     * scores for that document are passed to the callback before
     * proceeding to the next document.
     *
     * @param callback methods are called on this obj during iteration
     */
    public void iterateByDocument(ScoredDocumentCallback callback) {
      docs.forEachPair(new SDCProcedure(callback));
    }
    
class STCProcedure implements IntObjectProcedure {
      
       ScoredTermCallback callback;
       int termIndex;
       AIDAException ex;
      
      public STCProcedure(ScoredTermCallback callback) {
	this.callback = callback;
      }
      
      public void newTerm(int index, MIFSTerm term) {
	callback.newTerm(index, term);
	termIndex = index;
      }
      
      public AIDAException getEx() {
	return ex;
      }
      
      public boolean apply(int first, Object second) {
	try {
	DocScores ds = (DocScores)second;
	FreqScore val = ds.scores.get(termIndex);
        float score = val == null ? 0f : val.score;
        int freq = val == null ? 0 : val.freq;
        callback.apply(first, ds.doc, freq, score);
	} catch (AIDAException ex) {
	  this.ex = ex;
	  return false;
	}
	
	return true;
      }
    };

    /**
     * Iterates through the collection in term order calling
     * the relevant methods in the callback. In term order, all
     * the scores for all the documents for a specific term are
     * passed to the callback before starting the next term.
     *
     * @param callback methods are called on this obj during iteration
     */
    public void iterateByTerm(ScoredTermCallback callback) throws AIDAException {
      STCProcedure stcProcedure = new STCProcedure(callback);
      for (int j = 0, k = terms.length; j < k; j++) {
        TermImpl term = terms[j];
        stcProcedure.newTerm(j, term);
        docs.forEachPair(stcProcedure);
        if (stcProcedure.getEx() != null) throw stcProcedure.getEx();
      }
    }
  }

  private List<TermImpl> termList = new ArrayList<TermImpl>();
  private Map<String, Integer> termMap = new HashMap<String, Integer>();

  public STCBuilder(DateRange range) {
    collection = new STCollection();
    collection.dateRange = new DateRange(range);
  }

  public ScoredTermsCollection createSTC(boolean calculateTermStats) throws AIDAException {
    collection.terms = new TermImpl[termList.size()];
    termList.toArray(collection.terms);
    termList.clear();
    termMap.clear();

    if (calculateTermStats) {
      collection.iterateByTerm(new ScoredTermCallback() {

        TermImpl curTerm;

        public void newTerm(int termIndex, MIFSTerm term) {
          curTerm = (TermImpl) term;
        }

        public void apply(int docIndex, MIFSDocument doc, int frequency, float score) {
          if (frequency > 0) curTerm.docCount++;
          curTerm.setScores(score);
          curTerm.setFreqs(frequency);
        }
      });
    }

    STCollection tmp = collection;
    collection = null;
    return tmp;
  }

  /**
   * Adds a score for the specified doc and term. If the term already has a score
   * for that doc, then the results are added together.
   *
   * @param docId
   * @param termIndex
   * @param freq
   * @param score
   */
  public void addScore(int docId, int termIndex, int freq, float score) {
    DocScores ds = (DocScores)collection.docs.get(docId);
    FreqScore fs = ds.scores.get(termIndex);
    if (fs != null) {
      fs.freq += freq;
      fs.score += score;
    } else ds.scores.put(termIndex, new FreqScore(freq, score));
  }

  public boolean hasDocument(int docId) {
    return collection.docs.containsKey(docId);
  }

  public void addDocument(int docId, String title, String path) throws MalformedURLException {
    DocScores ds = (DocScores)collection.docs.get(docId);
    if (ds == null) {
      MIFSDocument doc = new MIFSDocument(docId, title, new URL(path));
      ds = new DocScores();
      ds.doc = doc;
      ds.scores = new HashMap<Integer, FreqScore>();
      collection.docs.put(docId, ds);
    }
  }

  public Integer addTerm(String termText, int docCount, int minFreq, int maxFreq, float minScore, float maxScore) {
    Integer index = termMap.get(termText);
    if (index == null) {
      index = termList.size();
      termMap.put(termText, index);
      TermImpl termImpl = new TermImpl();
      termImpl.minF = minFreq;
      termImpl.minS = minScore;
      termImpl.docCount = docCount;
      termImpl.maxF = maxFreq;
      termImpl.maxS = maxScore;
      termImpl.text = termText;
      termList.add(termImpl);
    }
    return index;
  }

  public Integer addTerm(String termText) {
    return addTerm(termText, 0, Integer.MAX_VALUE, 0, Float.POSITIVE_INFINITY, 0);

  }
}
