package anl.aida.core;


import anl.aida.util.DateRange;
import anl.aida.util.AIDAException;

import org.apache.lucene.store.IndexOutput;
import org.apache.lucene.store.MMapDirectory;

import java.io.IOException;

/**
 * Writes a ScoredTermsCollection out to a binary file.
 */
public class STCWriter implements ScoredTermCallback {

  private IndexOutput out;
  private int termIndex;

  /**
   * Called when a new term is encountered during
   * iteration.
   *
   * @param termIndex the index of the term in the collection
   * @param term      the term itself
   */
  public void newTerm(int termIndex, MIFSTerm term) {
    this.termIndex = termIndex;
  }


  /**
   * Notifies the implementor of a document score for the
   * term specified in the last call to newTerm.
   *
   * @param docIndex  the index of the document for which this is the term score
   * @param doc       the document itself
   * @param frequency the frequency count of the term in the document
   * @param score     the term score for the document
   */
  public void apply(int docIndex, MIFSDocument doc, int frequency, float score) throws AIDAException {
    try {
      if (frequency > 0) {
        out.writeInt(termIndex);
        out.writeInt(docIndex);
        out.writeInt(frequency);
        out.writeInt(Float.floatToIntBits(score));
      }
    } catch (IOException ex) {
      throw new AIDAException(ex);
    }
  }

  /**
   * Writes the specified collection to the specified file with the specified timestamp.
   *
   * @param file  the path of the file to write to
   * @param terms the scored terms collection to write
   * @throws IOException if there is an error writing the file.
   */
  public void writeCollection(String file, ScoredTermsCollection terms) throws IOException {
    MMapDirectory mmap = null;
    try {
      mmap = new MMapDirectory();
      out = mmap.createOutput(file);
      DateRange range = terms.getRange();
      out.writeLong(range.getStart().getTime());
      out.writeLong(range.getEnd().getTime());
      int docCount = terms.getDocumentCount();
      out.writeInt(docCount);
      int termCount = terms.getTermCount();
      out.writeInt(termCount);

      for (MIFSDocument doc : terms.documents()) {
        out.writeString(doc.getTitle());
        out.writeString(doc.getPath());
      }

      for (int i = 0; i < termCount; i++) {
        MIFSTerm term = terms.getTerm(i);
        out.writeString(term.getText());
        out.writeInt(term.getDocumentCount());
        out.writeInt(term.getMaxFrequency());
        out.writeInt(term.getMinFrequency());
        out.writeInt(Float.floatToIntBits(term.getMaxScore()));
        out.writeInt(Float.floatToIntBits(term.getMinScore()));
      }

      terms.iterateByTerm(this);
    } catch (AIDAException ex) {
      throw (IOException) ex.getCause();

    } finally {
      if (out != null) out.close();
      if (mmap != null) mmap.close();
    }
  }
}
