package anl.aida.data;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

/**
 * Encapsulates the axes / the dimensions of some data.
 * 
 * @version $Revision$ $Date$
 */
public class Axes<T extends CoordAxis> {

  public static final int TIME_STEP_NOT_FOUND = -1;

  private T xAxis, yAxis, timeAxis, layerAxis;
  private List<T> axes = new ArrayList<T>();

  /**
   * Creates an empty set of axes.
   */
  public Axes() {
  }
  

  public Axes(List<T> axes) {
    this.axes = new ArrayList<T>(axes);

    for (T axis : axes) {
      if (axis.getAxisType() == AxisType.X_AXIS)
	this.xAxis = axis;
      else if (axis.getAxisType() == AxisType.Y_AXIS)
	this.yAxis = axis;
      else if (axis.getAxisType() == AxisType.TIME)
	this.timeAxis = axis;
      else if (axis.getAxisType() == AxisType.LAYER)
	this.layerAxis = axis;
    }
  }
  
  public int size() {
    return axes.size();
  }
  
  /**
   * Assumes only two axes and gets the
   * non-time axis.
   * 
   * @return the non-time axis.
   */
  public T getNonTimeAxis() {
    for (T axis : axes) {
      if (axis.getAxisType() != AxisType.TIME) return axis;
    }
    return null;
  }

  /**
   * Returns the Date for the specified timestep. If the we are unable to parse
   * the time step axis sufficient to create a date then this returns null.
   * 
   * @param timestep
   *          the time step we want the date for.
   * @return the Date for the specified timestep, or null if there is no time
   *         axis, or null if unable to parse the time axis units.
   */
  public Date getDate(int timestep) {
    if (timeAxis != null)
      return ((TimeCoordAxis) timeAxis).getDate(timestep);
    return null;
  }

  /**
   * Gets the starting date of the time axis.
   * 
   * @return the starting date of the time axis.
   */
  public Date getStartDate() {
    if (timeAxis != null)
      return getDate(timeAxis.getRange().getOrigin());
    return null;
  }

  /**
   * Gets the ending date of the time axis.
   * 
   * @return the ending date of the time axis.
   */
  public Date getEndDate() {
    if (timeAxis != null) {
      Range range = timeAxis.getRange();
      return getDate(range.getOrigin() + (int) range.getExtent() - 1);
    }
    return null;
  }

  /**
   * Gets the timestep that corresponds to the specified date. If no such
   * timestep is found then return TIME_STEP_NOT_FOUND.
   * 
   * @param date
   *          the date whose time step we want
   * @return the timestep that corresponds to the specified date. If no such
   *         timestep is found then return TIME_STEP_NOT_FOUND.
   */
  public int getTimeStep(Date date) {
    if (timeAxis != null)
      return ((TimeCoordAxis) timeAxis).getTimeStep(date);
    return TIME_STEP_NOT_FOUND;
  }

  /**
   * Gets all the axis in this Axes object.
   * 
   * @return all the axis in this Axes object.
   */
  public List<T> getAxes() {
    return axes;
  }
  
  /**
   * Gets the first axis of the specified type.
   * 
   * @param type the type of axis to get
   * 
   * @return the first axis of the specified type or null if no
   * axis of that type is found
   */
  public T getAxis(AxisType type) {
    for (T axis : axes) {
      if (axis.getAxisType() == type) return axis;
    }
    
    return null;
  }
  
  /**
   * Gets the named axis.
   * 
   * @param name the name of the axis to get
   * @return the named axis or null if the axis is not found.
   */
  public T getAxis(String name) {
    for (T axis : axes) {
      if (axis.getName().equals(name)) return axis;
    }
    return null;
  }

  /**
   * Gets meta data for the z or layer axis.
   * 
   * @return the metadata for the z axis or null if it doesn't exist.
   */
  public T getZAxis() {
    return layerAxis;
  }

  /**
   * Gets meta data for the time axis.
   * 
   * @return the metadata for the time axis or null if it doesn't exist.
   */
  public T getTimeAxis() {
    return timeAxis;
  }

  /**
   * Gets meta data for the x-axis.
   * 
   * @return the metadata for the x-axis or null if it doesn't exist.
   */
  public T getXAxis() {
    return xAxis;
  }

  /**
   * Gets meta data for the y-axis.
   * 
   * @return the metadata for the y-axis or null if it doesn't exist.
   */
  public T getYAxis() {
    return yAxis;
  }
}
