/**
 * 
 */
package anl.aida.data;

import java.util.ArrayList;
import java.util.List;

import javax.measure.unit.Unit;

import ucar.ma2.Array;
import ucar.ma2.Index;
import anl.aida.formula.IllegalFormulaException;

/**
 * Combines two or more DataFrames into a single DataFrame such that the
 * variables of each DataFrame become a category axis in the new DataFrame. The
 * DataFrames to be combined must have single Axes of the Time type and they
 * must be compatible across all the Data Frames.
 * 
 */
public class Combiner {

  private List<DataFrame> frames = new ArrayList<DataFrame>();

  public void addDataFrame(DataFrame frame) {
    frames.add(frame);
  }

  private void checkAxes() throws IllegalFormulaException {
    DataFrameAxis timeAxis = null;
    Unit<?> unit = null;
    for (DataFrame frame : frames) {
      if (frame.getAxes().size() > 1)
	throw new IllegalFormulaException("Variable '" + frame.getVariable()
	    + " has more than one dimension.");
      if (frame.getAxes().getTimeAxis() == null) {
	throw new IllegalFormulaException("Variable '" + frame.getVariable()
	    + " is missing time dimension.");
      }

      if (timeAxis == null) {
	timeAxis = frame.getAxes().getTimeAxis();
      } else {
	if (!timeAxis.isCompatible(frame.getAxes().getTimeAxis())) {
	  throw new IllegalFormulaException("Time dimensions are incompatible.");
	}
      }

      if (unit == null) {
	unit = frame.getVariable().getUnit();
      } else {
	if (!unit.isCompatible(frame.getVariable().getUnit())) {
	  throw new IllegalFormulaException("Variable units are incompatible.");
	}
      }
    }
  }

  /**
   * Combine the added DataFrames into a single DataFrame with a Category axis
   * composed of the added DataFrame's variables.
   * 
   * @return the created DataFrame.
   * @throws IllegalFormulaException
   *           if the added DataFrames are incompatible.
   */
  public DataFrame combine() throws IllegalFormulaException {
    checkAxes();
    DataFrameBuilder builder = new DataFrameBuilder();
    DataFrameAxis timeAxis = frames.get(0).getAxes().getTimeAxis();
    builder.addAxis(timeAxis);

    String[] categories = new String[frames.size()];
    StringBuilder buf = new StringBuilder();
    int i = 0;
    for (DataFrame frame : frames) {
      categories[i] = frame.getVariable().getName();
      if (i > 0) buf.append(", ");
      buf.append(categories[i]);
      i++;
    }
    
    Variable var = new DefaultVariable(buf.toString(), buf.toString(), frames.get(0).getVariable().getUnit(), null);
    builder.setVariable(var);
    
    AxisData axisData = new AxisData();
    axisData.setUnit(frames.get(0).getVariable().getUnit());
    axisData.setDescription("Category");
    axisData.setName("Category");
    axisData.setType(AxisType.CATEGORY);
    axisData.setRange(new Range(0, categories.length));
    DefaultStringCoordAxis axis = new DefaultStringCoordAxis(categories, axisData);
    builder.addAxis(DataFrameAxis.createDataFrameAxis(axis, 1));

    Array array = ArrayFactory.createDoubleArray(new int[] { timeAxis.getExtent(),
	categories.length });
    
    Index arrayIndex = array.getIndex();
    for (int j = 0; j < frames.size(); j++) {
      DataFrame frame = frames.get(j);
      DataFrameIndex index = frame.getIndex();
      for (int time = 0; time < timeAxis.getExtent(); time++) {
	index.setTime(time);
	arrayIndex.set(time, j);
	array.setDouble(arrayIndex, frame.getDouble(index));
      }
    }
    builder.setArray(array);
    return builder.createDataFrame();
  }
}
