package anl.aida.data;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

/**
 * Manages the loading of data and the creation of data sets.
 * 
 * @version $Revision$ $Date$
 */
public class DataManager {

  private static final Logger logger = Logger.getLogger(DataManager.class);

  // alias, dataset
  private Map<String, Dataset> datasets = new HashMap<String, Dataset>();
  private AliasGenerator aliasGenerator = new AliasGenerator();
  
  /**
   * Creates a Dataset from the data at the specified URL. This DataManager
   * iterate through its collection of DataLoaders and use the first one that
   * can handle the data at the specified URL to create the Dataset. It will
   * also assign the Dataset an alias.
   * 
   * @param url
   *          the url of the data
   * @return a Dataset created from the data at the specified URL. If no Dataset
   *         can be created the null object {@link #DataManager#NULL_DATASET
   *         DataManager.NULL_DATASET} will be returned.
   *
  public List<Dataset> createDatasets(URL url) {
    for (DataLoader loader : dataLoaders) {
      if (loader.canHandle(url)) {
        try {
          // System.out.println("before creating dataset.");
          List<Dataset> data = loader.createDatasets(url);
          // System.out.println("after creating dataset. Dataset list size: " +
          // data.size());
          for (Dataset dataset : data) {
            setLoaderMap.put(dataset, loader);
            String alias = aliasGenerator.getNextAlias();
            dataset.setAlias(alias);
            datasets.put(alias, dataset);
            // System.out.println("alias: " + alias);
          }
          return data;
        } catch (IOException e) {
          // e.printStackTrace();
          msgCenter.warn("Error while creating a Dataset", e);
        }
      }
    }
  
    return NULL_DATASETS;
  }
   */
  
  /**
   * Creates a DataSet from a classname of a class
   * that implements Dataset.
   * 
   * @param className the className of the dataset.
   * @return the created dataset or null if the dataset could not
   * be created.
   */
  public Dataset createDataset(String className) {
    Dataset dataset = null;
    try {
      Class<?> klass = Class.forName(className);
      dataset = (Dataset)klass.newInstance();
      dataset.setAlias(aliasGenerator.getNextAlias());
      this.datasets.put(dataset.getAlias(), dataset);
    } catch (ClassNotFoundException ex) {
      logger.error("Error while creating dataset from className '" + className + "'", ex);
    } catch (InstantiationException ex) {
      logger.error("Error while creating dataset from className '" + className + "'", ex);
    } catch (IllegalAccessException ex) {
      logger.error("Error while creating dataset from className '" + className + "'", ex);
    }
    return dataset;
  }
  
  /**
   * Adds the specified dataset to the list managed by this
   * data manager.
   * 
   * @param dataset the dataset to add
   * @return the managed dataset.
   */
  public Dataset addDataset(Dataset dataset) {
    dataset.setAlias(aliasGenerator.getNextAlias());
    datasets.put(dataset.getAlias(), dataset);
    return dataset;
  }

  /**
   * Gets the datasets managed by this data manager.
   * 
   * @return the datasets managed by this data manager.
   */
  public List<Dataset> getDatasets() {
    List<Dataset> datasets = new ArrayList<Dataset>(this.datasets.values());
    return datasets;
  }

  /**
   * Replaces the specified datasets current alias with the new one. This does
   * not effect formulas in any way.
   * 
   * @param aliasMap
   *          key is the dataset and value is the new alias.
   */
  public void replaceAliases(Map<Dataset, String> aliasMap) {
    for (Dataset dataset : aliasMap.keySet()) {
      datasets.remove(dataset.getAlias());
    }

    for (Dataset dataset : aliasMap.keySet()) {
      String alias = aliasMap.get(dataset);
      dataset.setAlias(alias);
      aliasGenerator.markAliasUsed(alias);
      datasets.put(alias, dataset);
    }
  }

  /**
   * Gets the datareader appropriate for reading the specified dataset.
   * 
   * @param set
   *          the set that we want to read.
   * @return the datareader appropriate for reading the specified dataset.
   *
  public DataReader getDataReader(Dataset set) {
    return setLoaderMap.get(set).createReader(set);
  }
  */

  /**
   * Gets the number of datasets currently loaded.
   * 
   * @return the number of datasets currently loaded.
   */
  public int getDatasetCount() {
    return datasets.size();
  }

  /**
   * Gets the dataset with the specified alias.
   * 
   * @param alias
   *          the alias of the dataset to get.
   * @return the dataset with the specified alias.
   */
  public Dataset getDataset(String alias) {
    return datasets.get(alias);
  }

  /**
   * Splits a formula variable name into the dataset alias and variable name.
   * 
   * @param variableName
   *          the name to split
   * @return a string array whose first element is the alias and whose second
   *         element is the unaliased variable name.
   */
  public String[] splitVarName(String variableName) {
    return aliasGenerator.splitAlias(variableName);
  }

  /**
   * Closes the specified dataset and removes it from the datasets managed by
   * this manager.
   * 
   * @param alias
   *          the alias of the dataset to close
   * @throws IOException
   *           if there is an error closing the dataset.
   *
  public void closeDataset(String alias) throws IOException {
    Dataset set = datasets.remove(alias);
    if (set != null)
      set.close();
  }
  */

  /**
   * Closes all the datasets managed by this manager and clears the list of
   * datasets managed by the DatasetManager.
   * 
   * @throws IOException
   *           if there is an error closing a dataset.
   *
  public void closeAllDatasets() throws IOException {
    for (Dataset set : datasets.values()) {
      set.close();
    }
    datasets.clear();
  }
  */
}
