package anl.aida.data;

import java.util.HashMap;
import java.util.Map;
import java.util.StringTokenizer;

import javax.measure.unit.BaseUnit;
import javax.measure.unit.Unit;
import static javax.measure.unit.SI.*;
import static javax.measure.unit.NonSI.*;

/**
 * @version $Revision$ $Date$
 */
public class Units {

  private static Units instance = new Units();

  private Map<String, Unit<?>> unitMap = new HashMap<String, Unit<?>>();
  public static final Unit<?> MISSING_UNIT = new BaseUnit("_MISSING_UNIT_");

  private Units() {
    // areas
    unitMap.put("m2", METER.times(METER));
    unitMap.put("km2", KILO(METER).times(KILO(METER)));
    unitMap.put("ha", HECTARE);
    unitMap.put("hectare", HECTARE);
    unitMap.put("acre", HECTARE.times(.40469));
    unitMap.put("sq mi", MILE.times(MILE));

    // mass
    unitMap.put("kg", KILOGRAM);
    unitMap.put("g", GRAM);
    unitMap.put("lb", POUND);
    unitMap.put("lbs", POUND);
    unitMap.put("oz", OUNCE);
    unitMap.put("mg", MILLI(GRAM));
    unitMap.put("\u00b5g", MICRO(GRAM));
    unitMap.put("ug", MICRO(GRAM));
    unitMap.put("ng", NANO(GRAM));

    // others
    unitMap.put("deg K", KELVIN);
    unitMap.put("deg k", KELVIN);
    unitMap.put("K", KELVIN);
    unitMap.put("k", KELVIN);
    unitMap.put("m3", METER.times(METER).times(METER));
    unitMap.put("in3", CUBIC_INCH);

    // parts per million
    unitMap.put("PPM", Unit.ONE.divide(1000000));
    unitMap.put("ppm", Unit.ONE.divide(1000000));

    // length
    unitMap.put("cm", CENTI(METER));
    unitMap.put("m", METER);
    unitMap.put("in", INCH);
    unitMap.put("ft", FOOT);
    unitMap.put("yd", YARD);
  }

  /**
   * Parses the unit string into a Unit object. When possible a known unit (e.g.
   * kg, kg / m2, etc.) will be returned otherwise a unit unique to the string
   * will be returned.
   * 
   * @param unitString
   *          the string to parse
   * @return the created Unit.
   */
  public static Unit createUnit(String unitString) {
    return instance.makeUnit(unitString);
  }

  public static String getFormattedName(Unit unit) {
    return instance.getName(unit);
  }

  private Unit makeUnit(String unitString) {
    unitString = unitString.trim();
    // convert superscripts
    unitString = unitString.replace('\u00b2', '2');
    unitString = unitString.replace('\u00b3', '3');

    if (unitString.contains("/")) {
      return getDividedUnit(unitString);
    }

    return getUnit(unitString);
  }

  private Unit getDividedUnit(String unitString) {
    StringTokenizer tok = new StringTokenizer(unitString, "/");
    if (tok.countTokens() != 2)
      return null;
    String strRhs = tok.nextToken().trim();
    String strLhs = tok.nextToken().trim();
    if (!(unitMap.containsKey(strRhs) && unitMap.containsKey(strLhs))) {
      return new BaseUnit(unitString);
    }
    Unit<?> rhs = getUnit(strRhs);
    Unit<?> lhs = getUnit(strLhs);
    try {
      return rhs.divide(lhs);
    } catch (Exception ex) {
      return new BaseUnit(unitString);
    }
  }

  private Unit getUnit(String unitString) {
    Unit unit = unitMap.get(unitString);
    if (unit == null) {
      try {
	// see if we can make an SI type unit out of it
	unit = Unit.valueOf(unitString);
      } catch (IllegalArgumentException ex) {
	unit = new BaseUnit(unitString);
      }
    }
    return unit;
  }

  private String getName(Unit unit) {
    if (unit.equals(unitMap.get("PPM")))
      return "PPM";
    else
      return unit.toString();
  }

  private UnsupportedUnitException getException(String unitString) {
    return new UnsupportedUnitException("Unsupported unit format: '" + unitString + "'");
  }

}
