package anl.aida.formula;

import java.util.ArrayList;
import java.util.List;

import javax.measure.unit.Unit;

import ucar.ma2.Array;
import ucar.ma2.IndexIterator;
import anl.aida.data.Axes;
import anl.aida.data.AxisRange;
import anl.aida.data.DataFrame;
import anl.aida.data.DataFrameAxis;
import anl.aida.data.DataFrameBuilder;
import anl.aida.data.DataManager;
import anl.aida.data.DataTransformer;
import anl.aida.data.DefaultVariable;
import anl.aida.data.Units;
import anl.aida.data.Variable;
import anl.aida.parser.ASTTreeInfo;

/**
 * Formula implementation that produces data suitable for a tile plot. That is,
 * the data will include every value for each x,y cell in the domain.
 * 
 * @version $Revision$ $Date$
 */
public class DefaultFormula implements Formula {

  private ASTTreeInfo treeInfo;

  class VarFramePair {
    FormulaVariable var;
    DataFrame frame;
    IndexIterator iter;
    Unit<?> unit;

    public VarFramePair(DataFrame frame, FormulaVariable var) {
      this.frame = frame;
      this.var = var;
      unit = var.getUnit();
      iter = frame.getArray().getIndexIterator();
    }

    public void setFrame(DataFrame frame) {
      this.frame = frame;
      iter = frame.getArray().getIndexIterator();
    }
  }

  private FormulaParser parser;
  private DataTransformer transformer;
  private Unit<?> commonUnit;
  private List<FormulaVariable> variables = new ArrayList<FormulaVariable>();

  public DefaultFormula(FormulaParser parser, DataTransformer transformer, Unit<?> commonUnit) {
    this.parser = parser;
    this.transformer = transformer;
    this.commonUnit = commonUnit;
  }

  /*
   * (non-Javadoc)
   * 
   * @see anl.verdi.formula.Formula#getFormulaAsString()
   */
  @Override
  public String getFormulaAsString() {
    return parser.getFormulaAsString();
  }

  private List<VarFramePair> transformData(List<VarFramePair> data) {
    // transform the values in each array according to some strategy
    for (VarFramePair pair : data) {
      // transform the data (e.g. take the average over all values for each
      // time step.
      pair.setFrame(transformer.transform(pair.frame));
    }

    return data;
  }

  // make sure all the variables in the formula can
  // be converted into the common unit
  private void createFormulaVars(DataManager manager) throws IllegalFormulaException {
    String name = "";
    try {
      for (String varName : treeInfo.getVariableNames()) {
	name = varName;
	FormulaVariable fVar = FormulaVariable.createVariable(varName, manager);
	variables.add(fVar);
      }
    } catch (StringIndexOutOfBoundsException ex) {
      throw new IllegalFormulaException("Invalid dataset alias on variable '" + name + "'");
    }
  }

  private List<VarFramePair> readData(List<AxisRange> ranges) throws IllegalFormulaException {
    List<VarFramePair> results = new ArrayList<VarFramePair>();
    for (FormulaVariable var : variables) {
      DataFrame frame = var.evaluate();
      // do the unit conversion
      /*
       * if (convertUnits && !var.getUnit().equals(commonUnit)) { Array array =
       * frame.getArray(); Unit unit = var.getUnit(); for (IndexIterator iter =
       * array.getIndexIteratorFast(); iter.hasNext();) { double val =
       * iter.getDoubleNext(); iter.setDoubleCurrent(Measure.valueOf(val,
       * unit).doubleValue(commonUnit)); } }
       */
      VarFramePair varFramePair = new VarFramePair(frame, var);
      results.add(varFramePair);
    }
    return results;
  }

  /**
   * Validates the formula. This checks that the units, layers domains and
   * timestep are coherent and within range.
   * 
   * @param manager
   *          the data manager used to validate the formula
   * @param ranges
   *          the ranges that will filter or constrain the evaluation.
   * @return a ValidationResult containing the results of the validation.
   */
  public ValidationResult validate(DataManager manager, List<AxisRange> ranges) {
    try {
      treeInfo = parser.parse();
      createFormulaVars(manager);
      if (commonUnit == null && variables.size() > 0) {
	FormulaVariable var = variables.get(0);
	commonUnit = var.getUnit();
      } else if (variables.size() == 0) {
	// create a unit as a placeholder
	// this should be fine because if there are
	// no variables then the units don't matter.
	commonUnit = Units.createUnit("PPM");
      }
      FormulaValidator validator = new FormulaValidator(variables);
      return validator.validate(commonUnit, ranges);
    } catch (IllegalFormulaException e) {
      return ValidationResult.fail(e);
    }
  }

  public DataFrame evaluate(DataManager manager, List<AxisRange> ranges)
      throws IllegalFormulaException {
    if (treeInfo == null)
      throw new UnsupportedOperationException("Validate() must be called before evaluate");
    List<VarFramePair> results = readData(ranges);
    transformData(results);
    DataFrame frame = results.get(0).frame;

    Axes<DataFrameAxis> axes = frame.getAxes();
    int timeIndex = axes.getTimeAxis() == null ? -1 : axes.getTimeAxis().getArrayIndex();
    int otherIndex = axes.getNonTimeAxis() == null ? -1 : axes.getNonTimeAxis().getArrayIndex();
    /*
     * int layerIndex = axes.getZAxis() == null ? -1 :
     * axes.getZAxis().getArrayIndex(); int xIndex = axes.getXAxis() == null ?
     * -1 : axes.getXAxis().getArrayIndex(); int yIndex = axes.getYAxis() ==
     * null ? -1 : axes.getYAxis().getArrayIndex();
     */
    parser.setCoodinateIndices(timeIndex, otherIndex, -1, -1);

    DataFrameBuilder builder = new DataFrameBuilder();
    for (VarFramePair pair : results) {
      builder.addDataset(pair.frame.getDataset());
    }
    // the data frames produced by the readers should all have the same shape.
    // so we can take the axes from the first one.
    for (DataFrameAxis axis : axes.getAxes()) {
      builder.addAxis(axis);
    }

    builder.setArray(evaluateFormula(results));
    builder.setVariable(createVariable());
    DataFrame dataFrame = builder.createDataFrame();
    return transformer.transform(dataFrame);
  }

  private Array evaluateFormula(List<VarFramePair> results) throws IllegalFormulaException {
    for (VarFramePair pair : results) {
      String name = pair.var.getAliasedName();
      parser.setVariable(name, pair.frame);
    }
    FormulaArray array = parser.evaluate();
    return array.getArray();
  }

  private Variable createVariable() {
    String formula = parser.getFormulaAsString();
    return new DefaultVariable(formula, formula, commonUnit, null);
  }
}
