package anl.aida.reader;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.apache.uima.cas.CAS;
import org.apache.uima.collection.CollectionException;
import org.apache.uima.collection.CollectionReader_ImplBase;
import org.apache.uima.resource.ResourceInitializationException;
import org.apache.uima.util.Level;
import org.apache.uima.util.Logger;
import org.apache.uima.util.Progress;
import org.apache.uima.util.ProgressImpl;

import anl.aida.Configurator;
import anl.aida.util.AIDAUtils;
import anl.aida.util.Pair;

/**
 * UIMA CollectionReader reader that delegates the actual reading of documents
 * etc. to AIDAComponentReader implementations. The implementations are
 * instantiated during initialization from a from a list in the
 * ComponentReaderFile.
 */
public class AIDACompositeReader extends CollectionReader_ImplBase {

  //public static final String COMPONENTS = "anl.aida.reader.ComponentsFile";
  public static final String CACHE_START_DATE = "anl.aida.reader.CacheStartDate";
  private final static String MESSAGE_DIGEST = "anl.aida.reader.Reader_Messages";
  
  public static final String READERS_KEY = "readers";

  private List<AIDAComponentReader> readers = new ArrayList<AIDAComponentReader>();
  private int index = 0, count = 0;
  private AIDAComponentReader reader = null;

  @Override
  public void initialize() throws ResourceInitializationException {
    String startDate = (String) getConfigParameterValue(CACHE_START_DATE);
   
    Logger logger = getLogger();
    logger.log(Level.INFO, String.format(
	"Initializing AIDACompositeReader - for cache start Date: %s.", startDate));
    String nameForEx = "";
    
    try {
      Configurator config = Configurator.getConfigurator();
      for (String name : config.readerClassNames()) {
	logger.log(Level.INFO, String.format("AIDACompositeReader creating component reader: %s",
	    name));
	nameForEx = name;
	Class<?> clazz = Class.forName(name);
	readers.add((AIDAComponentReader) clazz.newInstance());
      }
      
      for (Pair<String, String> prop : config.readerProperties()) {
	  setConfigParameterValue(prop.getFirst(), prop.getSecond());
      }

      Date cacheStartDate = AIDAUtils.getDate(startDate);
      for (AIDAComponentReader reader : readers) {
	reader.initialize(this, cacheStartDate);
      }

      if (readers.size() == 0) {
	throw new ResourceInitializationException(MESSAGE_DIGEST, "no_components_error",
	    new Object[0]);
      }

      if (readers.size() > 0) {
	reader = readers.get(index);
	next();
      }

    } catch (IOException ex) {
      throw new ResourceInitializationException(MESSAGE_DIGEST, "io_error",
	  new Object[] {}, ex);
    } catch (ClassNotFoundException ex) {
      throw new ResourceInitializationException(MESSAGE_DIGEST, "component_create_error",
	  new Object[] { nameForEx }, ex);
    } catch (InstantiationException ex) {
      throw new ResourceInitializationException(MESSAGE_DIGEST, "component_create_error",
	  new Object[] { nameForEx }, ex);
    } catch (IllegalAccessException ex) {
      throw new ResourceInitializationException(MESSAGE_DIGEST, "component_create_error",
	  new Object[] { nameForEx }, ex);
    } catch (CollectionException ex) {
      throw new ResourceInitializationException(MESSAGE_DIGEST, "component_create_error",
	  new Object[] { nameForEx }, ex);
    }
  }

  private void next() throws CollectionException, IOException {
    if (!reader.hasNext()) {
      reader.close();
      index++;
      if (index < readers.size()) {
	reader = readers.get(index);
      } else {
	reader = null;
      }
    }
  }

  public void getNext(CAS cas) throws IOException, CollectionException {
    try {
      count++;
      reader.getNext(cas);
    } finally {
      next();
    }
  }

  public boolean hasNext() throws IOException, CollectionException {
    return reader != null;
  }

  public Progress[] getProgress() {
    return new Progress[] { new ProgressImpl(count, -1, Progress.ENTITIES, true) };
  }

  public void close() throws IOException {
    if (reader != null)
      reader.close();
  }
}
