/**
 * 
 */
package anl.aida.reader;

import java.io.File;
import java.io.IOException;
import java.util.Date;

import org.apache.uima.UIMAFramework;
import org.apache.uima.resource.ConfigurableResource;
import org.apache.uima.resource.ResourceInitializationException;
import org.apache.uima.util.Level;

import anl.aida.util.StandardIndexLineParser;

/**
 * Abstract class that will builds an AIDAComponentReader around a ContentReader
 * implementation. Subclasses new only implement {@link getContentSource} and 
 * {@link AbstractAIDAComponentReader.getIndexFileKey}.
 * 
 */
public abstract class ContentComponentReader extends AbstractAIDAComponentReader {
  
  protected ContentReader reader;

  /* (non-Javadoc)
   * @see anl.aida.reader.AbstractAIDAComponentReader#getNextResult()
   */
  @Override
  protected ReaderResult getNextResult() throws IOException {
    lineItems = indexIter.nextParsedLine(lineItems);
    location = lineItems[StandardIndexLineParser.URL];
    Date date = new Date(Long.parseLong(lineItems[StandardIndexLineParser.TIMESTAMP]));
    checkDate(date);
    
    String author = lineItems[StandardIndexLineParser.AUTHOR];
    int index = author.indexOf(";");
    if (index > -1) author = author.substring(0, index);
    
    return reader.read(location,  lineItems[StandardIndexLineParser.TITLE], date, author);
  }
  
  /**
   * Gets the source of the content.
   * 
   * @return the source of the content.
   */
  protected abstract String getContentSource();
  
  /**
   * Gets the directory to cache the downloaded content in.
   * 
   * @return the directory to cache the downloaded content in.
   */
  protected abstract String getCacheDirectoryKey();
  
  /**
   * Gets the name of the index file to create when caching the content.
   * 
   * @return the name of the index file to create when caching the content.
   */
  protected abstract String getCacheIndexKey();

  
  /* (non-Javadoc)
   * @see anl.aida.reader.AIDAComponentReader#initialize(org.apache.uima.resource.ConfigurableResource)
   */
  @Override
  public void initialize(ConfigurableResource resource, Date cacheStartDate) throws ResourceInitializationException {
    super.initialize(resource, cacheStartDate);
    String[] line = indexIter.nextParsedLine(new String[4]);
    indexIter.rewind();
    String url = line[StandardIndexLineParser.URL];
    if (!url.startsWith("http://")) {
      String indexFile = (String)resource.getConfigParameterValue(getIndexFileKey());
      File file = new File(indexFile);
      reader = new CachedContentReader(getContentSource(), file.getParent());
    }
    
    String cacheDir = (String) resource.getConfigParameterValue(getCacheDirectoryKey());
    String cacheIndex = (String) resource.getConfigParameterValue(getCacheIndexKey());
    if (cacheDir != null && cacheIndex != null) {
      UIMAFramework.getLogger(getClass()).log(Level.INFO, String.format("Caching Retrieved Content in %s, %n\tindex: %s",
	  cacheDir, cacheIndex));
      try {
	processors.add(new ResultCacher(cacheDir, cacheIndex));
      } catch (IOException e) {
	UIMAFramework.getLogger(getClass()).log(Level.SEVERE, "Unable to create result cacher", e);
      }
    }
  }

}
