package anl.aida.reader;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Date;

import org.htmlparser.Tag;
import org.htmlparser.beans.StringBean;
import org.htmlparser.tags.LinkTag;

/**
 * Scrapes article content from NYTimes pages. It searches the passed in URL for
 * the print link and then follows that link to get the printable page. The
 * printable page is much cleaner and this then scrapes that.
 * 
 */
public class NYTimesHTMLReader implements ContentReader {

  private static final String DOC_START = "___DOC_START___";
  private static final String DOC_END = "___DOC_END___";

  /**
   * Looks for the article body tag and places start and end tags in the output
   * to mark the start and end of the text.
   */
  private class NYTStringBean extends StringBean {

    boolean started = false;
    boolean done = false;

    @Override
    public void visitTag(Tag tag) {
      super.visitTag(tag);
      if (!done && tag.getTagName().equals("DIV")) {
	if (isStartTag(tag)) {
	  mBuffer.append("\n");
	  mBuffer.append(DOC_START);
	  started = true;
	}
      }
    }

    private boolean isStartTag(Tag tag) {
      String attrib = tag.getAttribute("id");
      if (attrib != null && attrib.equals("articleBody"))
	return true;

      attrib = tag.getAttribute("class");
      if (attrib != null && attrib.equals("articleBody"))
	return true;
      return false;
    }

    @Override
    public void visitEndTag(Tag tag) {
      super.visitEndTag(tag);

      if (started && tag.getTagName().equals("DIV")) {
	started = false;
	mBuffer.append("\n");
	mBuffer.append(DOC_END);
	done = true;
      }
    }
  }

  // looks for the "pagewanted=print" link and
  // stores it.
  private class NYTPrintLinkBean extends StringBean {

    boolean done = false;

    @Override
    public void visitTag(Tag tag) {
      super.visitTag(tag);
      if (!done && tag instanceof LinkTag) {
	//System.out.println("link = " + tag);
	String link = ((LinkTag) tag).getLink();
	if (link.contains("pagewanted=print")) {
	  done = true;
	  printLink = link;
	} else if (link.contains("pagewanted=all")) {
	  singlePageLink = link;
	}
      }
    }
  }

  // a do-nothing processor whose done() always returns true.
  private static class NullProcessor implements DocumentProcessor {

    @Override
    public boolean done() {
      return true;
    }

    @Override
    public void processLine(String line) {
    }

    @Override
    public void reset() {
    }

  }

  // processes lines looking for the start and end tags applied
  // by the NYTStringBean (above). Anything between those tags
  // is the doc's content.
  private class Processor implements DocumentProcessor {

    private boolean start = false;
    private boolean done = false;

    public void processLine(String line) {
      // System.out.println("line = " + line);
      if (start) {
	if (line.equals(DOC_END)) {
	  start = false;
	  done = true;
	} else {
	  // System.out.println("line = " + line);
	  content.append(line);
	  content.append(" ");
	}
      }

      if (!start && line.equals(DOC_START))
	start = true;
    }

    public boolean done() {
      return done;
    }

    public void reset() {
      start = false;
      done = false;
    }
  }

  private StringExtractor extractor, printLinkExtractor;
  private DocumentProcessor processor = new Processor();
  private DocumentProcessor nullProc = new NullProcessor();
  private StringBuilder content;
  private ReaderResult result;
  
  private String printLink = "", singlePageLink = "";
  private NYTStringBean bean = new NYTStringBean();
  private NYTPrintLinkBean printBean = new NYTPrintLinkBean();

  public NYTimesHTMLReader() {
    extractor = new StringExtractor(bean);
    printLinkExtractor = new StringExtractor(printBean);
  }

  /**
   * Reads the text from the specified URL. The title, date and author are
   * expected to be provided from another source, e.g. an RSS feed entry or an
   * archive "index."
   * 
   * @param url
   *          the url to read from
   * @param title
   *          the article title
   * @param date
   *          the date of the article
   * @return the result of reading and parsing the text at the url
   * @throws IOException
   *           if there is an error while reading
   */
  public ReaderResult read(String url, String title, Date date, String author) throws IOException {
    result = new ReaderResult();
    result.setTitle(title);
    result.setDate(new ParsedDate(date));
    
    // get a good URL if the url is a feed redirect.
    //url = ReaderUtilities.getRedirectedURL(url).toExternalForm();
   // System.out.println(url);
    
    result.setLocation(url);
    result.setSource(ReaderConstants.NYTIMES_SOURCE);
    result.setLink("");

    ParsedAuthors authors = new ParsedAuthors();
    authors.addName("", author, "");
    result.setAuthors(authors);
    
    printBean.done = false;
    printLink = singlePageLink = "";
    try {
      printLinkExtractor.run(url, nullProc);
    } catch (IOException ex) {
      // try to run it on manually received text
      printBean.done = false;
      printLink = singlePageLink = "";
      printLinkExtractor.runWithManualFetch(url, nullProc);
      //UIMAFramework.getLogger(getClass()).log(Level.WARNING, "Error reading url: '" + url + "'", ex);
    }

    String link = url;
    if (printLink.length() > 0) {
      link = printLink;
      link = fixLink(link, url);
    } else if (singlePageLink.length() > 0) {
      link = singlePageLink;
      link = fixLink(link, url);
    }

    bean.done = false;
    bean.started = false;
    content = new StringBuilder();
    processor.reset();
    link = link.replace("amp;", "");
    extractor.run(link, url, processor);
    result.setContents(content.toString());

    return result;
  }
  
  private String fixLink(String link, String url) throws MalformedURLException {
    if (link.startsWith("/")) {
	URL aURL = new URL(url);
	link = aURL.getProtocol() + "://" + aURL.getHost() + link;
    }
    return link;
  }
}
