package anl.aida.reader;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;

/**
 * Preprocesses promed mail alerts, dividing them up
 * if more than one "document" is in the mail text. Each
 * individual document is given the mail title.
 * 
 */
public class ProMedPreProcessor {
  
  private static final String SECTION_BREAK = "****";
  private static final String FOOTER = "ProMED-mail makes every effort to";
  
  private interface Streamer {
    
    Reader next() throws IOException;
    void read(BufferedReader in)  throws IOException;
  }
  
  private class SingleStreamer implements Streamer {
    
    private boolean done = false;

    @Override
    public Reader next() throws IOException {
      if (done) return null;
      done = true;
      return new StringReader(buf.toString());
    }
    
    public void read(BufferedReader in) throws IOException {
      String line = "";
      // read in the rest of the lines
      while ((line = in.readLine()) != null) {
        buf.append(line);
        buf.append("\n");
      }
      
    }
  }
  
  private class MultiStreamer implements Streamer {

    private boolean done = false;
    private String title = "";
    private BufferedReader in;

    @Override
    public Reader next() throws IOException {
      readChunk();
      if (done) return null;
      return new StringReader(buf.toString());
    }

    public void read(BufferedReader in) throws IOException {
      // find the title in the current buf.
      String line = "";
      StringBuilder titleBuf = new StringBuilder();
      BufferedReader reader = new BufferedReader(new StringReader(buf.toString()));
      while ((line = reader.readLine()) != null) {
	titleBuf.append(line);
	titleBuf.append("\n");
	if (line.startsWith("***************************")) break;
      }
      title = titleBuf.toString();
      
      this.in = in;
      int count = 0;
      // read past the first "******"
      while ((line = in.readLine()) != null) {
	// go beyond title line
	if (line.startsWith(SECTION_BREAK)) break;
	count++;
      }
    }
    
    private void readChunk() throws IOException {
      String line;
      buf = new StringBuilder();
      buf.append(title);
      int titleSize = buf.length();
      // read lines in until we get to the section break
      // or to the end
      while ((line = in.readLine()) != null) {
	if (line.startsWith(SECTION_BREAK) || line.startsWith(FOOTER)) break;
	buf.append(line);
	buf.append("\n");
      }
      
      done = titleSize == buf.length();
    }
  }
  
  private Streamer streamer;
  private StringBuilder buf = new StringBuilder();
  private BufferedReader reader;
  
  /**
   * Creates a ProMedPreProcessor that will read from the specified input stream.
   * 
   * @param stream the stream to read from
   * @throws IOException if there is an exception reading the stream.
   */
  public ProMedPreProcessor(Reader stream) throws IOException {
    reader = new BufferedReader(stream);
    String line = "";
    while ((line = reader.readLine()) != null) {
      buf.append(line);
      buf.append("\n");
      if (line.startsWith("In this update") || (line.contains("[") && line.contains("Date") && line.contains("]"))) {
	streamer = new MultiStreamer();
	break;
	
      } else if (line.startsWith("Date") || line.startsWith("Sent")) {
	streamer = new SingleStreamer();
	break;
      }
    }
    
    streamer.read(reader);
  }
  
 
  /**
   * Gets the reader for the next document to read, or null if there is no
   * more documents.
   * 
   * @return the reader for the next document to read, or null if there is no
   * more documents.
   * @throws IOException if there is an error getting the next reader to use.
   */
  public Reader next() throws IOException {
    return streamer.next();
  }
  
  /**
   * Closes this preprocessor.
   * 
   * @throws IOException
   */
  public void close() throws IOException {
    reader.close();
  }

}
