package anl.aida.reader;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.Reader;

/**
 * Reads Pro Med email txt for Date, Title content etc.
 * This uses a chain of LineParsers to do the parsing.
 * 
 */
public class ProMedTxtReader {

  private class NullParser implements LineParser {

    @Override
    public LineParser next() {
      return this;
    }

    @Override
    public boolean parse(String line, StringBuilder buf) {
      return false;
    }
  }
  
  private class TitleParser implements LineParser {
    
    private StringBuilder title = new StringBuilder();

    @Override
    public LineParser next() {
      return new ContentParser();
    }

    @Override
    // accumulates lines until its finds *************
    public boolean parse(String line, StringBuilder buf) {
      if (line.startsWith("****")) {
	result.setTitle(title.toString().trim());
	return true;
      }
      title.append(line);
      title.append(" ");
      return false;
    }
  }
  
  

  private class ContentParser implements LineParser {

    @Override
    public LineParser next() {
      return new NullParser();
    }

    @Override
    public boolean parse(String line, StringBuilder buf) {
      if (isEndTag(line))
	return true;
      buf.append(line);
      // append in case line break is breaking words
      buf.append(" ");
      return false;
    }

    protected boolean isEndTag(String line) {
      return (line.equals("--"));
    }
  }

  /*
  private class DateParser implements LineParser {

    private DateFormat format = new SimpleDateFormat("dd MMM yyyy");
    private DateFormat format2 = new SimpleDateFormat("MMMM dd, yyyy");
    @Override
    public LineParser next() {
      return new AuthorParser();
    }

    @Override
    public boolean parse(String line, StringBuilder buf) {
      if (line.startsWith("Date") || line.startsWith("Sent")) {
	// find the index where first number is -- this ignores the day name
	// which may or may not be present and may have a different format
	int index = 0;
	// some dates look like 22-24 jan, so skip past the -
	int start = line.indexOf("-");
	start = start == -1 ? 0 : start;
	for (int i = start, n = line.length(); i < n; i++) {
	  if (Character.isDigit(line.charAt(i))) {
	    index = i;
	    break;
	  }
	}
	
	
	Date date = format.parse(line, new ParsePosition(index));
	if (date == null) {
	  line = line.substring(line.indexOf(" ")).trim();
	  date = format2.parse(line, new ParsePosition(0));
	}
	result.setDate(new ParsedDate(date));
	return true;
      }

      return false;
    }
  }
  

  private class AuthorParser implements LineParser {
    
    LineParser next;

    @Override
    public LineParser next() {
      return next;
    }

    @Override
    // pro med has a source line which id's the alerts source
    // we use that as the author here
    public boolean parse(String line, StringBuilder buf) {
      // fonte is source in Portuguese!!!, Sourve is mispelled source:?
      if (line.startsWith("Source:") || line.startsWith("Fonte:") || line.startsWith("Sourve:")) {
	ParsedAuthors authors = new ParsedAuthors();
	String source = line.substring(8);
	authors.addName("", source, "");
	next = new LinkParser();
	result.setAuthors(authors);
	return true;
      } else if (line.startsWith("From:")) {
	int index = line .indexOf("<<a href");
	String source = "";
	if (index == -1) {
	   source = line.substring(6);
	} else {
	  source = line.substring(6, index);
	}
	next = new ContentParser();
	ParsedAuthors authors = new ParsedAuthors();
	authors.addName("", source.trim(), "");
	result.setAuthors(authors);
	return true;
      }

      return false;
    }
  }
  */

  private class LinkParser implements LineParser {

    @Override
    public LineParser next() {
      return new ContentParser();
    }

    @Override
    public boolean parse(String line, StringBuilder buf) {
      if (line.startsWith("<<http")) {
	int index = line.indexOf(">");
	String link = line.substring(2, index);
	result.setLink(link);
	return true;
      }

      return false;
    }
  }
  
  private ReaderResult result;

  /**
   * Reads the pro med text from the specified reader and returns a result.
   * 
   * @param in the input
   * @return the ReaderResult
   * @throws IOException if there is an error during the reading
   */
  public ReaderResult run(String location, Reader in) throws IOException {
    result = new ReaderResult();
    result.setSource(ReaderConstants.PRO_MED_SOURCE);
    ParsedAuthors authors = new ParsedAuthors();
    authors.addName("", "Pro Med", "");
    result.setAuthors(authors);
    BufferedReader reader = new BufferedReader(in);
    StringBuilder buffer = new StringBuilder();
    String line;
    LineParser parser = new TitleParser();
    while ((line = reader.readLine()) != null) {
      line = line.trim();
      boolean switchParser = parser.parse(line, buffer);
      if (switchParser)
	parser = parser.next();
    }

    result.setLocation(location);
    result.setContents(buffer.toString());
    return result;
  }
}
