package anl.aida.sv;

import anl.aida.core.DocumentID;
import anl.aida.core.DocumentIDMatcher;

import org.apache.lucene.store.IndexInput;
import org.apache.lucene.store.MMapDirectory;
import pitt.search.semanticvectors.ObjectVector;

import java.io.IOException;
import java.util.Iterator;

/**
 * Reads Semantic Vector vectors that have been written
 * using a DocumentVectorStoreWriter.
 *
 */
public class DocumentVectorStoreReader {

  private IndexInput indexInput;
  private boolean hasHeader;

  public DocumentVectorStoreReader(String vectorFile) throws IOException {
    MMapDirectory dir = new MMapDirectory();
    this.indexInput = dir.openInput(vectorFile);

    /* Read number of dimensions from header information. */
    String test = indexInput.readString();
    /* Include "-" character to avoid unlikely case that first term is "dimensions"! */
    if ((test.equalsIgnoreCase("-dimensions"))) {
      ObjectVector.vecLength = indexInput.readInt();
      this.hasHeader = true;
    } else {
      System.err.println("No file header for file " + vectorFile +
              "\nAttempting to process with default vector length: " +
              ObjectVector.vecLength +
              "\nIf this fails, consider rebuilding indexes - existing " +
              "ones were probably created with old version of software.");
      this.hasHeader = false;
    }
  }

  public Iterable<ObjectVector> getAllVectors() {
    try {
      indexInput.seek(0);
      if (hasHeader) {
        indexInput.readString();
        indexInput.readInt();
      }
    }
    catch (IOException e) {
      e.printStackTrace();
    }
    return new VectorIterable(indexInput);
  }

  /**
   * Given an object, get its corresponding vector. <br>
   *
   * @param matcher the DocumentIDMatcher that will match the
   *                desired vector
   * @return vector from the VectorStore, or null if not found.
   * @throws IOException if there is an error reading the vectors
   */
  public float[] getVector(DocumentIDMatcher matcher) throws IOException {

    indexInput.seek(0);
    if (hasHeader) {
      indexInput.readString();
      indexInput.readInt();
    }
    while (indexInput.getFilePointer() < indexInput.length() - 1) {
      int id = indexInput.readInt();
      String path = indexInput.readString();
      if (matcher.match(id, path)) {
        float[] vector = new float[ObjectVector.vecLength];
        for (int i = 0; i < ObjectVector.vecLength; ++i) {
          vector[i] = Float.intBitsToFloat(indexInput.readInt());
        }
        return vector;
      } else {
        indexInput.seek(indexInput.getFilePointer() + 4 * ObjectVector.vecLength);
      }
    }

    return null;

  }

  /**
   * Trivial (costly) implementation of getNumVectors that iterates and counts vectors.
   *
   * @return the numbe of stored vectors
   */
  public int getNumVectors() {
    Iterator allVectors = this.getAllVectors().iterator();
    int i = 0;
    while (allVectors.hasNext()) {
      allVectors.next();
      ++i;
    }
    return i;
  }

  /**
   * Implements the hasMoreElements() and nextElement() methods
   * to give Enumeration interface from store on disk.
   */
  public class VectorIterable implements Iterable<ObjectVector>, Iterator<ObjectVector> {
    IndexInput indexInput;

    public VectorIterable(IndexInput indexInput) {
      this.indexInput = indexInput;
    }

    public Iterator<ObjectVector> iterator() {
      return this;
    }

    public ObjectVector next() {
      DocumentID id = null;
      float[] vector = new float[ObjectVector.vecLength];
      try {
        int stcid = indexInput.readInt();
        String path = indexInput.readString();
        id = new DocumentID(stcid, path);
        for (int i = 0; i < ObjectVector.vecLength; ++i) {
          vector[i] = Float.intBitsToFloat(indexInput.readInt());
        }
      }
      catch (IOException e) {
        e.printStackTrace();
      }
      return new ObjectVector(id, vector);
    }

    public void remove() {
      throw new UnsupportedOperationException();
    }

    public boolean hasNext() {
      return (indexInput.getFilePointer() < indexInput.length());
    }
  }
}
