package anl.aida.sv;

import anl.aida.core.KeywordGenerator;
import anl.aida.util.HighestTermScores;
import anl.aida.util.AIDAException;
import pitt.search.semanticvectors.ObjectVector;
import pitt.search.semanticvectors.VectorUtils;
import pitt.search.semanticvectors.ZeroVectorException;

import java.io.IOException;

/**
 * Generates keywords using the semantic vector package.
 *
 */
public class SVKeywordGenerator implements KeywordGenerator {

  private SemanticVectorStore vectorStore;

  /**
   * Creates a KeywordGenerator using the specifed SemanticVectorStore.
   *
   * @param vectorStore the wrapper to use for SV access
   * @throws IOException if there is an error reading the stores.
   */
  public SVKeywordGenerator(SemanticVectorStore vectorStore) throws IOException {
    this.vectorStore = vectorStore;
  }


  /**
   * Generates and returns the specified number of keywords for the document
   * identified by the specified path.
   *
   * @param docPath the path of the document to generate words for
   * @param count   the number of keywords to generate
   * @return the collection of HighestTermScores
   * @throws AIDAException if there is an error generating the keywords
   */
  public HighestTermScores generate(String docPath, int count) throws AIDAException {
    try {
      Searcher searcher = new Searcher(docPath);
      return searcher.getNearestNeighbors(count);
    } catch (ZeroVectorException ex) {
      throw new AIDAException("Vector error", ex);
    } catch (IOException ex) {
      throw new AIDAException("Error reading vectors", ex);
    }
  }

  /**
   * Generates the specified number of keywords that are most similar
   * to the specified vector.
   *
   * @param vector the vector to search against
   * @param count  the number of terms to return
   * @return the found keywords.
   */
  public HighestTermScores generate(float[] vector, int count) {
    Searcher searcher = new Searcher(vector);
    return searcher.getNearestNeighbors(count);
  }

  private class Searcher {

    float[] queryVector;

    /**
     * @param docPath the path to the doc to get keywords for
     * @throws ZeroVectorException if there is an error with the vector
     * @throws IOException         if there is an error reading the vector data
     */
    Searcher(String docPath) throws ZeroVectorException, IOException {
      this.queryVector = vectorStore.getDocumentVector(docPath);
      if (queryVector == null || VectorUtils.isZeroVector(this.queryVector)) {
        throw new ZeroVectorException("Query vector is zero ... no results.");
      }
    }

    Searcher(float[] vector) {
      queryVector = vector;
    }

    public float getScore(float[] testVector) {
      return VectorUtils.scalarProduct(this.queryVector, testVector);
    }

    public HighestTermScores getNearestNeighbors(int numResults) {
      HighestTermScores results = new HighestTermScores(numResults);

      float score;

      for (ObjectVector testElement : vectorStore.getTermVectors()) {
        score = getScore(testElement.getVector());
        results.add(testElement.getObject().toString(), score);

      }

      return results;
    }

  }
}
