package anl.aida.sv;

import anl.aida.core.DocumentIDMatcher;
import anl.aida.util.EnumIter;
import pitt.search.semanticvectors.ObjectVector;
import pitt.search.semanticvectors.VectorStoreReader;

import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

/**
 * Wraps access to the semantic vectors data.
 *
 */
public class SemanticVectorStore {
  
  private static final float[] EMPTY_VECTOR = new float[0];

  private static class PathMatcher implements DocumentIDMatcher {

    String path;

    public boolean match(int id, String path) {
      return this.path.equals(path);
    }
  }

  private static class IdMatcher implements DocumentIDMatcher {

    int id;

    public boolean match(int id, String path) {
      return this.id == id;
    }
  }


  private DocumentVectorStoreReader docReader;
  private VectorStoreReader termReader;
  private PathMatcher pathMatcher = new PathMatcher();
  private IdMatcher idMatcher = new IdMatcher();

  private Map<Integer, float[]> docCache = new HashMap<Integer, float[]>();
  private Map<String, float[]> termCache = new HashMap<String, float[]>();

  /**
   * Creates an SVWrapper using the specified term and doc stores.
   *
   * @param docStorePath  the path to the document terms store
   * @param termStorePath the path to the terms store
   * @throws java.io.IOException if there is an error reading the stores.
   */
  public SemanticVectorStore(String docStorePath, String termStorePath) throws IOException {
    docReader = new DocumentVectorStoreReader(docStorePath);
    termReader = new VectorStoreReader(termStorePath);
  }

  public synchronized float[] getDocumentVector(String docPath) throws IOException {
    pathMatcher.path = docPath;
    return docReader.getVector(pathMatcher);
  }

  /**
   * Gets the document vector for the document with the specified stc Index.
   *
   * @param stcIndex the index
   * @return the document vector
   * @throws IOException if there is an error finding the document vector
   */
  public synchronized float[] getDocumentVector(int stcIndex) throws IOException {
    float[] vector = docCache.get(stcIndex);
    if (vector == null) {
      idMatcher.id = stcIndex;
      vector = docReader.getVector(idMatcher);
      docCache.put(stcIndex, vector);
      //System.out.println("docCache.size() = " + docCache.size());
    }
    return vector;
  }

  public synchronized float[] getTermVector(String term) {
    float[] vector = termCache.get(term);
    if (vector == null) {
      vector = termReader.getVector(term);
      if (vector == null ) termCache.put(term, EMPTY_VECTOR);
      else termCache.put(term, vector);
      
      // if vector is the empty we return null
      // the empty vector just short circuits the look
      // up for a term that doesn't exist
    } else if (vector == EMPTY_VECTOR) return null;

    return vector;
  }

  public synchronized Iterable<ObjectVector> getTermVectors() {
    return new EnumIter<ObjectVector>(termReader.getAllVectors());
  }

  public Iterable<ObjectVector> getDocumentVectors() {
    return docReader.getAllVectors();
  }
}
