package anl.aida.ae.lucene;

import anl.aida.ae.lucene.DefaultFieldBuilder;
import anl.aida.ae.lucene.FieldBuilder;
import anl.aida.ae.lucene.FieldBuilderFactory;
import anl.aida.ae.lucene.Keywords;
import anl.aida.ae.lucene.LuceneConstants;
import anl.aida.util.AIDAException;
import static junit.framework.Assert.assertEquals;
import org.apache.lucene.analysis.Token;
import org.apache.lucene.analysis.TokenStream;
import org.apache.lucene.document.Document;
import org.apache.lucene.index.IndexReader;
import org.apache.uima.UIMAFramework;
import org.apache.uima.analysis_engine.AnalysisEngine;
import org.apache.uima.analysis_engine.AnalysisEngineDescription;
import org.apache.uima.analysis_engine.AnalysisEngineProcessException;
import org.apache.uima.cas.CAS;
import org.apache.uima.cas.Feature;
import org.apache.uima.cas.Type;
import org.apache.uima.cas.TypeSystem;
import org.apache.uima.collection.CollectionProcessingEngine;
import org.apache.uima.collection.EntityProcessStatus;
import org.apache.uima.collection.StatusCallbackListener;
import org.apache.uima.collection.metadata.CpeDescription;
import org.apache.uima.resource.ResourceInitializationException;
import org.apache.uima.resource.ResourceSpecifier;
import org.apache.uima.util.InvalidXMLException;
import org.apache.uima.util.XMLInputSource;
import static org.junit.Assert.assertNotNull;
import org.junit.Test;

import javax.xml.stream.XMLStreamException;
import java.io.IOException;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Date;
import java.util.List;

/**
 * 
 *         Date: Jan 22, 2009 9:12:42 AM
 */
public class LucenceConsumerTests {


  private CollectionProcessingEngine engine;
  private StatusCallbackListenerImpl listener;

  public void setUp(String cpePath) throws IOException, InvalidXMLException, ResourceInitializationException {
    XMLInputSource xml = new XMLInputSource(cpePath);
    CpeDescription desc = UIMAFramework.getXMLParser().parseCpeDescription(xml);
    engine = UIMAFramework.produceCollectionProcessingEngine(desc);
    listener = new StatusCallbackListenerImpl();
    engine.addStatusCallbackListener(listener);

  }

  private void run() throws ResourceInitializationException {
    engine.process();
    try {
      // engine.process occurs in different thread
      // so we wait here for it to finish. The StatusCallbackListener
      // does a notifyAll() when the collection processing is complete
      // and that stops the wait.
      synchronized (listener) {
        listener.wait();
      }
    } catch (InterruptedException e) {
      e.printStackTrace();
    }
  }


  /*
  @Test
  public void testScorer() throws IOException, MIFSException, ClassNotFoundException, SQLException {
    TermCountDAO dao = new TermCountDAO();
    Connection connection = dao.createConnection();
    Statement stmt = connection.createStatement();
    stmt.execute("delete from term_scores");
    stmt.execute("delete from terms");
    stmt.execute("delete from fields");
    stmt.execute("delete from documents");
    stmt.close();
    connection.close();

    TermScorer scorer = new TermScorer();
    IndexReader reader = IndexReader.open("./test_data/pub_med_5");
    scorer.apply(reader, new DBTermFunction(new Date()));

    connection = dao.createConnection();
    stmt = connection.createStatement();
    ResultSet set = stmt.executeQuery("select count(*) from documents");
    set.next();
    assertEquals(reader.numDocs(), set.getInt(1));
    set.close();

    set = stmt.executeQuery("select count(*) from fields");
    set.next();
    // 4 are contents, title
    assertEquals(2, set.getInt(1));
    set.close();

    // scores calculated using the Luke (Lucene Index Toolbox)
    // name is the field here
    set = stmt.executeQuery("select score, freq from terms_full where text = 'hantavirus' and name = 'title'");
    set.next();
    assertEquals(.59884f, set.getFloat(1), .0001);
    assertEquals(1, set.getInt(2));
    set.close();

    set = stmt.executeQuery("select score, freq from terms_full where text = 'hantavirus' and name = 'contents'");
    set.next();
    assertEquals(0.1047f, set.getFloat(1), .0001);
    assertEquals(1, set.getInt(2));
    set.close();


    // note that hev is an acronym but luence lowercases the terms so hev rather than HEV.
    set = stmt.executeQuery("select score, freq from terms_full where text = 'hev' and name = 'title'");
    set.next();
    assertEquals(.6775f, set.getFloat(1), .0001);
    assertEquals(2, set.getInt(2));
    set.close();


    set = stmt.executeQuery("select score, freq from terms_full where text = 'hev' and name = 'contents'");
    set.next();
    assertEquals(.4734f, set.getFloat(1), .0001);
    assertEquals(10, set.getInt(2));
    set.close();

    stmt.close();
    connection.close();
  }
  */

  @Test
  public void testSimpleConsumer() throws ResourceInitializationException, IOException, InvalidXMLException {
    setUp("./test/anl/mifs/sample_cpe.xml");
    run();

    // read the index
    IndexReader reader = IndexReader.open("./indices/uima");
    assertEquals(5, reader.numDocs());
    for (int i = 0; i < reader.numDocs(); i++) {
      // make sure the fields got stored.
      // Note that contents is indexed but not stored.
      assertNotNull(reader.document(i).get(LuceneConstants.AUTHOR_FIELD));
      assertNotNull(reader.document(i).get(LuceneConstants.PATH_FIELD));
      assertNotNull(reader.document(i).get(LuceneConstants.TITLE_FIELD));
      assertNotNull(reader.document(i).get(LuceneConstants.TIMESTAMP_FIELD));
    }

  }

  @Test
  public void testMappedConsumer() throws ResourceInitializationException, IOException, InvalidXMLException {
    setUp("./test/anl/mifs/gate_gaz_cpe.xml");
    run();

    // read the index
    IndexReader reader = IndexReader.open("./indices/uima");
    assertEquals(5, reader.numDocs());
    for (int i = 0; i < reader.numDocs(); i++) {
      // make sure the fields got stored.
      assertNotNull(reader.document(i).get(LuceneConstants.AUTHOR_FIELD));
      assertNotNull(reader.document(i).get(LuceneConstants.PATH_FIELD));
      assertNotNull(reader.document(i).get(LuceneConstants.TITLE_FIELD));
      assertNotNull(reader.document(i).get(LuceneConstants.TIMESTAMP_FIELD));

      // contents was indexed -- figure out how to check for these when
      // they are not stored.
      //assertNotNull(reader.document(i).getField("Lookup.combined"));
      //assertNotNull(reader.document(i).getField("Lookup.minorType"));
    }
  }

  @Test
  public void testBuilder() throws IOException, InvalidXMLException, ResourceInitializationException, 
  AnalysisEngineProcessException, AIDAException {
    AnalysisEngine ae = createAE();
    CAS cas = ae.newCAS();

    TypeSystem system = cas.getTypeSystem();

    Type type = system.getType("anl.mifs.types.Lookup");
    Feature feature = system.getFeatureByFullName("anl.mifs.types.Lookup:majorType");

    cas.setDocumentText("In November, Mr. Smith went to France.");
    ae.process(cas);

    Document doc = new Document();
    DefaultFieldBuilder builder = new DefaultFieldBuilder("Lookup.majorType", type, feature);
    builder.addFeature(system.getFeatureByFullName("anl.mifs.types.Lookup:minorType"));
    builder.build(cas, doc, new Keywords());

    TokenStream stream = doc.getField("Lookup.majorType").tokenStreamValue();
    Token token = new Token();
    String[] vals = {"date", "month", "title", "male", "location", "country"};
    int i = 0;
    while (token != null) {
      token = stream.next(token);
      if (token != null) {
        assertEquals(vals[i++], token.term());
      }
    }
    assertEquals(vals.length, i);
  }

  @Test
  public void testXMLBuilder() throws IOException, InvalidXMLException, ResourceInitializationException,
          AnalysisEngineProcessException, XMLStreamException, AIDAException {
    AnalysisEngine ae = createAE();
    CAS cas = ae.newCAS();
    cas.setDocumentText("In November, Mr. Smith went to France.");
    ae.process(cas);

    TypeSystem system = cas.getTypeSystem();
    FieldBuilderFactory fac = new FieldBuilderFactory("./test/anl/mifs/field_mapping.xml");
    List<FieldBuilder> builders = fac.createBuilders(system);
    Document doc = new Document();
    for (FieldBuilder builder : builders) {
      builder.build(cas, doc, new Keywords());
    }

    TokenStream stream = doc.getField("Lookup.combined").tokenStreamValue();
    Token token = new Token();
    String[] vals = {"date", "month", "title", "male", "location", "country"};
    int i = 0;
    while (token != null) {
      token = stream.next(token);
      if (token != null) {
        assertEquals(vals[i++], token.term());
      }
    }
    assertEquals(vals.length, i);

    stream = doc.getField("Lookup.minorType").tokenStreamValue();
    token = new Token();
    vals = new String[]{"month", "male", "country"};
    i = 0;
    while (token != null) {
      token = stream.next(token);
      if (token != null) {
        assertEquals(vals[i++], token.term());
      }
    }
    assertEquals(vals.length, i);
  }

  private AnalysisEngine createAE() throws IOException, InvalidXMLException, ResourceInitializationException {
    String MAPPING_FILE = "anl.mifs.ae.gate.GateAnnotator.MappingFile";
    String GATE_APP_FILE = "anl.mifs.ae.gate.GateAnnotator.GateAppFile";

    XMLInputSource in = new XMLInputSource("./desc/GateDescriptor.xml");
    ResourceSpecifier spec = UIMAFramework.getXMLParser().parseResourceSpecifier(in);
    ((AnalysisEngineDescription) spec).getMetaData().getConfigurationParameterSettings().
            setParameterValue(MAPPING_FILE, "./gate_app/gazetteer_map.xml");
    ((AnalysisEngineDescription) spec).getMetaData().getConfigurationParameterSettings().
            setParameterValue(GATE_APP_FILE, "./gate_app/gazetteer.gapp");

    AnalysisEngine ae = UIMAFramework.produceAnalysisEngine(spec);
    return ae;
  }

  class StatusCallbackListenerImpl implements StatusCallbackListener {

    boolean wait = true;

    public void entityProcessComplete(CAS cas, EntityProcessStatus entityProcessStatus) {
    }

    public void initializationComplete() {
    }

    public void batchProcessComplete() {
    }

    public void collectionProcessComplete() {
      synchronized (this) {
        notifyAll();
      }
    }

    public void paused() {
    }

    public void resumed() {
    }

    public void aborted() {
    }
  }

}
