package anl.aida.ae.lucene;

import static junit.framework.Assert.assertFalse;
import static junit.framework.Assert.assertTrue;
import static org.junit.Assert.assertEquals;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.ParseException;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

import org.apache.lucene.document.DateTools;
import org.apache.lucene.document.Document;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.IndexReader;
import org.junit.Test;

import anl.aida.ae.lucene.CacheCreator;
import anl.aida.ae.lucene.DateRangeFieldFilter;
import anl.aida.ae.lucene.Keywords;
import anl.aida.ae.lucene.LuceneConstants;
import anl.aida.ae.lucene.ReaderProducer;
import anl.aida.ae.lucene.WordFilter;
import anl.aida.util.DateRange;
import anl.aida.util.DateRangeProducer;
import anl.aida.util.AIDAException;

/**
 * Lucene related tests.
 *
 */
public class LuceneTests {

  static class TestReaderProducer implements ReaderProducer {
    public Reader getReader(Document doc) {
      String path = doc.get(LuceneConstants.PATH_FIELD);
      XMLEventReader reader = null;
      String absText = "";
      try {

        XMLInputFactory factory = XMLInputFactory.newInstance();
        reader = factory.createXMLEventReader(new BufferedInputStream(new URL(path).openStream()));

        while (reader.hasNext()) {
          XMLEvent evt = reader.nextEvent();
          if (evt.isStartElement()) {
            StartElement elmt = evt.asStartElement();
            String qName = elmt.getName().getLocalPart();

            if (qName.equals("AbstractText")) {
              absText = reader.getElementText();

            }
          }
        }
      } catch (FileNotFoundException e) {
        e.printStackTrace();
      } catch (XMLStreamException e) {
        e.printStackTrace();
      } catch (MalformedURLException e) {
        e.printStackTrace();
      } catch (IOException e) {
        e.printStackTrace();
      } finally {
        try {
          if (reader != null) reader.close();
        } catch (XMLStreamException e) {
          e.printStackTrace();
        }
      }
      return new StringReader(absText);
    }
  }

  Calendar cal = GregorianCalendar.getInstance();

  private Date getDate(int year, int month, int day) {
    cal.set(year, month, day, 0, 0, 0);
    cal.set(Calendar.MILLISECOND, 0);
    return cal.getTime();
  }

  private Date add(Date date, int delta) {
    cal.setTime(date);
    cal.add(Calendar.DAY_OF_MONTH, delta);
    return cal.getTime();
  }

  @Test
  public void testDateRangeProducer() {
    Date date = getDate(2008, 9, 1);
    int rangeLength = 5;
    int index = 0;
    DateRangeProducer producer = new DateRangeProducer(date, 52, rangeLength);
    for (DateRange range : producer.ranges()) {
      Date start = add(date, index);
      assertEquals(start, range.getStart());
      Date end = add(start, rangeLength);
      assertEquals(end, range.getEnd());
      //System.out.println(range.getStart() + " - " + range.getEnd());
      index++;
    }

    // -1 because index is incremented at the end
    assertEquals(52 - rangeLength, index - 1);
  }

  @Test
  public void testCacheCreator() throws IOException, AIDAException, ParseException {
    Calendar cal = GregorianCalendar.getInstance();
    cal.set(2008, 9, 1, 0, 0);
    Date start = cal.getTime();
    cal.add(Calendar.DAY_OF_MONTH, 60);
    Date end = cal.getTime();

    IndexReader reader = IndexReader.open("./indices/cache");
    DateRangeProducer producer = new DateRangeProducer(start, 60, 5);
    CacheCreator creator = new CacheCreator(producer, new WordFilter(2), reader);
    creator.createCache(new File("/tmp/cache"), new TestReaderProducer(), new Keywords());
  }
  
  @Test
  public void testDateFilter() throws ParseException {
    Date start = getDate(2009, 0, 1);
    Date end = getDate(2009, 0, 10);

    DateRangeFieldFilter filter = new DateRangeFieldFilter(start, end);
    assertTrue(filter.evaluate(DateTools.round(start.getTime(), DateTools.Resolution.DAY)));
    assertTrue(filter.evaluate(DateTools.round(end.getTime(), DateTools.Resolution.DAY)));
    assertTrue(filter.evaluate(DateTools.round(add(start, 3).getTime(), DateTools.Resolution.DAY)));

    assertFalse(filter.evaluate(DateTools.round(add(start, -1).getTime(), DateTools.Resolution.DAY)));
    assertFalse(filter.evaluate(DateTools.round(add(end, 1).getTime(), DateTools.Resolution.DAY)));

    Document doc = new Document();
    Date date = add(start, -1);
    doc.add(new Field(LuceneConstants.TIMESTAMP_FIELD, DateTools.dateToString(date, DateTools.Resolution.DAY),
            Field.Store.YES, Field.Index.ANALYZED));
    assertFalse(filter.evaluate(doc));

    doc = new Document();
    date = start;
    doc.add(new Field(LuceneConstants.TIMESTAMP_FIELD, DateTools.dateToString(date, DateTools.Resolution.DAY),
            Field.Store.YES, Field.Index.ANALYZED));
    assertTrue(filter.evaluate(doc));

    doc = new Document();
    date = add(start, 3);
    doc.add(new Field(LuceneConstants.TIMESTAMP_FIELD, DateTools.dateToString(date, DateTools.Resolution.DAY),
            Field.Store.YES, Field.Index.ANALYZED));
    assertTrue(filter.evaluate(doc));

    doc = new Document();
    date = end;
    doc.add(new Field(LuceneConstants.TIMESTAMP_FIELD, DateTools.dateToString(date, DateTools.Resolution.DAY),
            Field.Store.YES, Field.Index.ANALYZED));
    assertTrue(filter.evaluate(doc));

    doc = new Document();
    date = add(end, 1);
    doc.add(new Field(LuceneConstants.TIMESTAMP_FIELD, DateTools.dateToString(date, DateTools.Resolution.DAY),
            Field.Store.YES, Field.Index.ANALYZED));
    assertFalse(filter.evaluate(doc));
  }
}
