package anl.aida.reader;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.IOException;
import java.util.Date;

import org.junit.Test;

import anl.aida.reader.CachedContentReader;
import anl.aida.reader.GuardianHTMLReader;
import anl.aida.reader.LATimesHTMLReader;
import anl.aida.reader.NYTimesHTMLReader;
import anl.aida.reader.ReaderConstants;
import anl.aida.reader.ReaderResult;
import anl.aida.util.IndexIterator;
import anl.aida.util.StandardIndexLineParser;

public class ReaderTests {
  
  @Test
  public void cacheContentTest() throws IOException {
    CachedContentReader reader = new CachedContentReader("BBC Source", "./test_data/bbc");
    IndexIterator iter = new IndexIterator("./test_data/bbc/bbc_cached_index.txt");
    String[] lineItems = new String[4];
    
    // iterate through them all making sure data is good
    while (iter.hasNext()) {
      lineItems = iter.nextParsedLine(lineItems);
      String location = lineItems[StandardIndexLineParser.URL];
      Date date = new Date(Long.parseLong(lineItems[StandardIndexLineParser.TIMESTAMP]));
     
      
      String author = lineItems[StandardIndexLineParser.AUTHOR];
      int index = author.indexOf(";");
      if (index > -1) author = author.substring(0, index);
      
      ReaderResult result = reader.read(location,  lineItems[StandardIndexLineParser.TITLE], date, author);
      ReaderResult.Error error = result.isValid();
      assertTrue(error.toString(), error == ReaderResult.Error.NONE);
    }
    
    // test specific content
    ReaderResult result = reader.read("./bbc_content.zip!content_100.txt", "Foo", new Date(), "Bill Smith");
    assertTrue(result.getContents().startsWith("Obama Mid-East plans in jeopardy By Roger Hardy BBC Middle East analyst One"));
    assertTrue(result.getContents().trim().endsWith("many would conclude that he had bitten off more than he can chew."));
    assertEquals(3512, result.getContents().length());
    assertEquals("http://news.bbc.co.uk/2/hi/middle_east/8242387.stm", result.getLocation());
    
    
  }

  @Test
  public void nytReaderTest() throws IOException {
    // note this url may expire at some point
    //String url = "http://feeds.nytimes.com/click.phdo?i=e340f49dca4a8f7280c3d8aacb513c2a";
    //String url = "http://www.nytimes.com/2009/09/02/world/asia/02afghan.html";
    String url = "http://www.nytimes.com/2009/09/01/world/americas/01webbrazil.html?partner=rss&emc=rss";
    String title = "Changes Proposed to Benefit States Oil Giant in Brazil";
    String author = "Alexei Barrionuevo";

    NYTimesHTMLReader reader = new NYTimesHTMLReader();
    Date date = new Date();
    ReaderResult result = reader.read(url, title, date, author);
    ReaderResult.Error error = result.isValid();
    assertTrue(error.toString(), error == ReaderResult.Error.NONE);
    assertEquals(date, result.getDate().getJavaDate());
    assertEquals(title, result.getTitle());
    assertEquals(ReaderConstants.NYTIMES_SOURCE, result.getSource());
    String contents = result.getContents();
    String start = "RIO DE JANEIRO ";
    assertEquals(start, contents.substring(0, start.length()));
    String end = "development of the new fields.  ";
    assertEquals(end, contents.substring(contents.length() - end.length(), contents.length()));
  }
  
  @Test
  public void latReaderTest() throws IOException {
    // note this url may expire at some point
    //String url = "http://feeds.nytimes.com/click.phdo?i=e340f49dca4a8f7280c3d8aacb513c2a";
    //String url = "http://www.nytimes.com/2009/09/02/world/asia/02afghan.html";
    String url = "http://articles.latimes.com/2006/jan/01/nation/na-teen1";
    String title = "After Flight Back From Iraq, Teen to Be Grounded By Jamie Malernee, Kevin Smith And Karla Shores";
    String author="";
    LATimesHTMLReader reader = new LATimesHTMLReader();
    Date date = new Date();
    ReaderResult result = reader.read(url, title, date, author);
    ReaderResult.Error error = result.isValid();
    assertTrue(error.toString(), error == ReaderResult.Error.NONE);
    assertEquals(date, result.getDate().getJavaDate());
    assertEquals(title, result.getTitle());
    assertEquals(ReaderConstants.LATIMES_SOURCE, result.getSource());
    String contents = result.getContents();
    String start = "After Flight Back From Iraq";
    assertEquals(start, contents.substring(0, start.length()));
    assertTrue(contents.contains("prestigious prep school"));
  }
  
  @Test
  public void guardiaReaderTest() throws IOException {
    // note this url may expire at some point
    //String url = "http://feeds.nytimes.com/click.phdo?i=e340f49dca4a8f7280c3d8aacb513c2a";
    //String url = "http://www.nytimes.com/2009/09/02/world/asia/02afghan.html";
    String url = "http://www.guardian.co.uk/news/blog/2008/jan/01/michaelwhitespoliticalblog56";
    String title = "Michael White's political blog: January 1";
    String author="";
    GuardianHTMLReader reader = new GuardianHTMLReader();
    Date date = new Date();
    ReaderResult result = reader.read(url, title, date, author);
    ReaderResult.Error error = result.isValid();
    assertTrue(error.toString(), error == ReaderResult.Error.NONE);
    assertEquals(date, result.getDate().getJavaDate());
    assertEquals(title, result.getTitle());
    assertEquals(ReaderConstants.GUARDIAN_SOURCE, result.getSource());
    String contents = result.getContents();
    String start = "Please activate cookies in order to turn autoplay off " +
    		"guardian.co.uk News Web Michael White's political blog: January 1 Mid morning, " +
    		"West London: And a Happy New Year to you too, though I fear the coming year";
    assertEquals(start, contents.substring(0, start.length()));
    assertTrue(contents.contains("Are we approaching a major turning point " +
    		"in the global pecking order"));
  }
}
